#include "Ancestry.h"
#include "read_control.h"
#include "write_trees.h"
#include "Simple_IS.h"
#include "Li_and_Stephens.h"
#include "Griffiths_Marjoram.h"

//With trees:
//1) Add in code to check for no mut/coal events and choose/upweight those
//that can.
//2) Add in code to look for monomorphic sites - set wattersons theta
//according to only SEGREGATING sites

//void Ancestry::Input_params(string con_Fname)
//{
//	/*Read_Parameters(
//	    theta           , rho              , SMC              , importance_sample, sim_data        ,
//	    using_rec_map   , all_nodes        , DEBUG            , sampler          , no_runs         ,
//	    change_rho      , bridge_resolution, change_theta     , output_MLE       , choose_hap_first,
//	    coal_fast       , threshold        , choose_nonrec    , first_nonrec     , use_logs        ,
//	    cheap_coal_upwt , display_events   , display_haps     , display_weights  , set_seed        ,
//	    using_wattersons, greedy_mle       , write_likelihoods, write_trees      , write_full_data ,
//	    collect_times   , record_every_time, collect_densities, which_set        ,  file_tag       ,
//	    con             , con_Fname
//	);*/
//  return;
//}

void Ancestry::Input_data(string data_Fname, const int& which_set)
{
	Read_Data(smpl_size, all_sites, no_segsts,  using_rec_map, print_data, site_posns, which_set, con,
	    data_Fname, data);
//	Read_Data(smpl_size, all_sites, no_segsts,  using_rec_map, print_data, site_posns, which_set, data_Fname, data);
}

void Ancestry::getHapmapDat(
    const string& data_Fname,
    const ChapmapParams& hapmapParams,
    map<double, snp>& hapMapDat,
    vector<double>& genmap,
    vector<double>& physmap,
    int& n_gen,
    int& n_hap
  )

{
   read_genetic_map(
         hapmapParams ,
         genmap       , // genetic map for region
         physmap        // physical map for region
       );
   read_haplotype_data(
         hapmapParams ,
         hapMapDat    ,
         genmap       , // genetic map for region
         physmap      , // physical map for region
         n_gen        ,
         n_hap
       );
   cout << "Read in haplotypes, legend and genetic map.\n";
}

void Ancestry::initialiseHapList(const int& run)
{
  // This hap_list refers to the structure keeping track of all the ancestral states of the data. It also records which sites are ancestral.
  hap_list.clear();
  hap_list.resize(smpl_size);
  for (int index=0; index<smpl_size; index++)
  {
    if (!using_data)
      hap_list.at(index).alleles.resize(rec_sites,-99); // 0 denotes non ancestral, 1 denotes ancestral TYPE (or arbitrary type), 2 denotes dervied TYPE (or or arbitrary)
    else
    {
      if(data.size()==0)
        cout << "  Warning! data is of size 0 in Initialise. Problem reading in data?\n";
      hap_list.at(index).alleles = data.at(index);
      if(hap_list.at(index).alleles.size()==0) {
        cout << "  Warning! no SNPs?\n";
      } else {
        if(DEBUG>1 && index==0 && run==-1) {
          cout << "  hap_list.at(0).alleles.at(0) = " << hap_list.at(0).alleles.at(0) << "\n";
        }        
      }
      
      if(rec_sites != (int)hap_list.at(0).alleles.size()){
        cout << "Warning! expecting " << rec_sites
             << " sites, but there are " << hap_list.at(0).alleles.size()
             << " in the data." << "\n";
        getchar();
      }
    }
    hap_list.at(index).nodes.resize(rec_sites, index);
    hap_list.at(index).left_rec_edge  = 0;
    hap_list.at(index).right_rec_edge = rec_sites-1;
    hap_list.at(index).rec_dist = rec_map.at(hap_list.at(index).right_rec_edge) - rec_map.at(0) ;   // When not using a real rec_map recombination is still done using that variable at present.
    hap_list.at(index).reached_MRCA = 0;
  }
  if(DEBUG>0 && hap_list.size()==0) {
    cout << "  Warning, hap_list has size zero! (in initialiseHapList)\n";
    getchar();
  }  
  if(DEBUG>2 && run==-1)
    cout << "  DEBUG = " << DEBUG << "\n";
  if(DEBUG>2 && run==-1) { // This double checks against those printed with the data itself
     cout << "  \n";
     cout << "  Recombination positions: \n";
     cout << "  0=" << rec_map.at(0) << ", ";
     if(rec_map.size()>4) {
       cout << "  1=" << rec_map.at(1) << ", ";
       cout << "  2=" << rec_map.at(2) << ", ";
       cout << "  (end-1)=" << rec_map.at(rec_map.size()-2) << ", ";
       cout << "  (end)=" << rec_map.at(rec_map.size()-1) << ", ";
       cout << "  \n";
     } else {
       cout << "  length of rec_map = " << rec_map.size() << "\n";
    }
  }
  if(display_haps) {
    cout << "  The haplotype list at initialisation (having copied from 'data') is:\n";
    Display_Haplotypes(false);
  }
  return;
}

void Ancestry::initialiseRhoAndTheta(const int& run)
{
  int i;
  if(using_wattersons)
  {
    double watter1=0;
    double wattersons;

    // watterson's estimate of theta calculation:
    for (i=1; i<smpl_size; i++)
      watter1 += 1/((double)i);
    wattersons = (double)no_segsts/watter1;
    if(DEBUG > 1)
      cout << "Assuming there really are " <<  no_segsts <<
              " segregating sites (careful here), \n`watterson's estimate of theta is: "
              << wattersons << endl;
    theta = wattersons;
    con.theta = theta;
  }

  if(change_rho)
    Change_Rho(rho);
  if(1) cout << "  rho = " << rho << "\n";

  // This records the cumulative recombination distance reached by each site. It should be zero at site 0 and rho(*totalPhysicalDist?) at the last site.
  rec_map.clear();
  //typically this is false, so the following lines are run.
  // In reality the variable recombination rate is dealt with by having 'site_posns' as genetic distances
  if(!using_rec_map)
  {
    rec_map.resize(rec_sites);
    //for (i=0; i<rec_sites; i++)
    //  rec_map.at(i) = (rho/(double)(rec_sites-1)) * (double)i;

    if(!((int)site_posns.size()==rec_sites) ||
       !((int)rec_map.size()   ==rec_sites) ) {
      cout << "  Warning, unexpected sizes for site_posns or rec_map.\n";
      cout << "  This is usually caused by a problem in reading the parameters or the data.\n";
      cout << "  site_posns.size() = " << site_posns.size() << ", and \n" <<
              "  rec_map.size()    = " <<  rec_map.size()   << "\n";
      cout << "The value of 'rec_sites' = " << rec_sites << ", this is what they should be sized to.\n";
      cout << "The filename of the control file is: " << con_Fname.c_str() << "\n";
      cout << "Press any key to continue: " << flush;
      getchar();
    }
    for (i=0; i<rec_sites; i++)
      rec_map.at(i) = rho * site_posns.at(i); // This is the total recombination fraction between the first site and this site.

  }
  rho_per_site  = rho/(double)(rec_sites-1);
  theta_per_site  = theta/(double)(no_segsts);
  return;
}
// Initialises the marginal genealogy storage matrix (generally called before each run):
void Ancestry::initialiseMargGens()
{
  // This is a matrix of NODES. For each locus there are 2n-1 nodes keeping track of the leaves and coalescence events for the tree at that locus.
  int i,j;
  marg_gens.clear();
  marg_gens.resize( (2*smpl_size)-1 );
  for (i=0; i<(2*smpl_size)-1; i++)
    marg_gens.at(i).resize(rec_sites);

  // Initialise first n NODES at each site.
  for (i=0; i<smpl_size; i++)
  {
    for (j=0; j<rec_sites; j++)
    {
      marg_gens.at(i).at(j).n1          = NULL;
      marg_gens.at(i).at(j).n2          = NULL;
      marg_gens.at(i).at(j).c_time      = 0;
      marg_gens.at(i).at(j).time_below  = 0;
      marg_gens.at(i).at(j).active      = 1;
      marg_gens.at(i).at(j).index_site  = j;
      marg_gens.at(i).at(j).index_epoch = i;
    }
  }
  for (i=smpl_size; i<(2*smpl_size)-1; i++)
    for (j=0; j<rec_sites; j++)
      marg_gens.at(i).at(j).active    = 0;
}

//Initialises the data structures such as the matrices for the data, the recombination map etc.
void Ancestry::Initialise(const int& run)
{
	int i;
	
  if(data.size()==0) {
    cout << "  Warning! data is of size 0 at beginning of Initialise. Problem reading in data?\n";
    cout << "  On run " << run << "\n";
  }
	if(run==-1)
	{
		display_tree		= 0;
		know_derived		= 0;
    con.display_tree=display_tree;
		con.know_derived=know_derived;
		if(no_runs >= 11)
		{
			display_tree		= 0; con.display_tree=display_tree;
			display_haps		= 0; con.display_haps=display_haps;
			display_events	= 0; con.display_events=display_events;
			if(DEBUG>1.9)
				cout << "Toning down DEBUG to 1.9 as there are too many runs for that level of output.\n";
			DEBUG				= 1.9; con.DEBUG = DEBUG;
		}
    too_many_runs = 1e6+1;
		if(no_runs > (too_many_runs-1)/2) {
			record_every_time	= 0; con.record_every_time = record_every_time;
	  if(DEBUG>0)
	      cout << "  Warning, with this many runs times won't be tracked.\n";
		}
    if(no_runs >= too_many_runs && DEBUG>0) {
      cout << "  Warning, with this many runs even less may/will be tracked.\n";
    }
		total_no_recs = 0; //This has been moved from the main body - but I don't understand how it could've worked there, maybe it didn't

		// Initialise rho and theta (so also rho_per_site and rec_map etc.):
    // set up rec_sites:
		if (sim_data)   rec_sites = all_sites; 
    else            rec_sites = no_segsts;
    con.rec_sites = rec_sites;
    if(DEBUG>=1)
      cout << "rec_sites = " << rec_sites << " (in Initialise) \n";
		initialiseRhoAndTheta(run);
	}

	//count_type_1.clear();
	//count_type_1.resize(no_segsts);
	still_segregating.clear();
	still_segregating.resize(no_segsts,1);

	initialiseHapList(run);
	// To avoid too much memory allocation and deallocation I keep track of which lineages are currently active in the simulation. (The memory needed can only increase.)
	live_lineages.clear();
	live_lineages.resize(smpl_size,1);

	initialiseMargGens();
	
	node_MRCA = (2*smpl_size)-1; //So when we've activated node no.(MRCA_node-1) we've reached the MRCA at that site. (because things are indexed from 0)

	mut_times.clear();
	// At each segregating site (SNP position) there should be exactly 1 mutation, this records at what time that mutation happens.		
	if (!sim_data)
		mut_times.resize(no_segsts);	
	
	// This is a vector of counters to record which node each tree has currently reached. 
	which_node.clear();
	which_node.resize(rec_sites,smpl_size-1); // The initial value is set at this as the first n nodes are dealt with before any simulation.

	event_densities.clear();
	prior_weights.clear();
	choice_quotients.clear();

	MRCA_times.clear();
	if(record_every_time)
	{
		MRCA_times.resize(no_runs);
		for(i=0; i<no_runs; i++)
			MRCA_times.at(i).resize(no_segsts);
	}
	else
	{
		//MRCA_times.resize(1);
		//for(i=0; i<1; i++)
		//	MRCA_times.at(i).resize(no_segsts);
	}

		
	lbridge_weights.clear(); //** This matrix never changes size, resizing on every run is therefore inefficient.**//
		lbridge_weights.resize(max(1,change_rho));
	for(i=0; i<max(1,change_rho); i++)
		lbridge_weights.at(i).resize(bridge_resolution,0);
	rec_records.clear();

	// This records the number of currently live/active lineages 
	actv_lngs = hap_list.size();
	dead_lngs = 0;
	MRCA_lngs = 0;

	//Initialising rates
	rec_rate	= (rec_map.at(rec_sites-1)-rec_map.at(0)) * (double)smpl_size / 2.0; // Note the use of 'rec_sites' here.
	coal_rate	= (double)smpl_size * (double)(smpl_size-1)/2.0;
	mut_rate	= (double)smpl_size * double(no_segsts) * (theta_per_site/2.0);							 // Note the use of 'rec_sites' here.
	evnt_rate	= rec_rate + coal_rate + mut_rate;

	crnt_height = 0; // Stores the height of each epoch int the tree

	evnt_nmbr	= 0; // Mainly to locate problems. Counts number of events that have occurred.	

	//total_no_recs = 0; //This has been MOVED to only happening whenever run==0


	// Surely the latter cannot work as Q_file will go out of scope?
	if(collect_times && !record_every_time)
	{
		bool Q_info = 1;
		string Q_filename;
		if(!Q_file.is_open())
		{
			Q_filename = Create_Times_Filename(Q_info);
			Q_file.open(Q_filename.c_str(),ios::app);
			if(Q_file.fail())
			{
				cout << "Warning, opening the time file failed. \n";
				getchar();
			}
			if(!(Q_file.good()))
			{
				cout << "Warning, There has been at least one error with the times file. \n";
				getchar();
			}
		}	
	}
	if(collect_densities)
	{
		string densities_Filename;		
		if(!densities_file.is_open())
		{
			densities_Filename = Create_Densities_Filename();
			densities_file.open(densities_Filename.c_str(),ios::app);
			if(densities_file.fail())
			{
				cout << "Warning, opening the densities file failed. \n";
				getchar();
			}
			if(!(densities_file.good()))
			{
				cout << "Warning, There has been at least one error with the densities file. \n";
				getchar();
			}				
		}
	}
	/////////#################/////////////
    ////////   Initialise con /////////////
	/////////#################/////////////
  con.rec_rate    = rec_rate;
	con.coal_rate   = coal_rate;
	con.mut_rate    = mut_rate;
	con.evnt_rate   = evnt_rate;
	con.crnt_height = crnt_height;
	con.evnt_nmbr   = evnt_nmbr;
	con.actv_lngs   = actv_lngs;
	con.dead_lngs   = dead_lngs;
	con.MRCA_lngs   = MRCA_lngs;
	con.rec_sites   = rec_sites;
	con.node_MRCA   = node_MRCA;
}




// This function can be used to run lots of simulations, plot a likelihood surface and report MLE estimates.
void Ancestry::Controller()
{
	int i, run;
    
    laverage_likelihood = -1E308;
	lsum_bridge_weights.resize(max(1,change_rho));
	for(i=0; i<max(1,change_rho); i++)
		lsum_bridge_weights.at(i).resize(bridge_resolution,-1E308);
	lsum_bridge_weights_squared.resize(max(1,change_rho));
	for(i=0; i<max(1,change_rho); i++)
		lsum_bridge_weights_squared.at(i).resize(bridge_resolution,-1E308);
	//for(i=0; i<lsum_bridge_weights.size(); i++)  
		//lsum_bridge_weights.at(i) = -1E308;
	//for(i=0; i<lsum_bridge_weights_squared.size(); i++)  
		//lsum_bridge_weights_squared.at(i) = -1E308;
	Define_Bridge_Rhos();
	Make_Rho_Vector();
     
	// Console outputs:
	if( importance_sample && con.verbose>=1)
	  Output_Sampler();
	if(!importance_sample && con.verbose>=1)
	    cout << "Not importance sampling, just generating genealogies from the prior.\n";
	
	time_t Beg_time, Current_time, Last_time, End_time;
	Beg_time = time(NULL); End_time = time(NULL);
  if(con.verbose>=1) 
  	cout << "  Time at start = " << Beg_time << "\n";
	if(!SMC && con.verbose>=1) 
	  cout << "  Using the full Coalescent Model.\n";
	if( SMC && con.verbose>=1) 
	  cout << "  Using the SMC Model.\n";
	if(importance_sample && con.verbose>=1)
	{
		if(choose_hap_first) 
		  cout << "  Choosing Haplotype 1st, before calculating event weights.\n";
		else 
		  cout << "  Choosing Haplotype properly, calculating weights on all haps for each event.\n";
	}
	if(coal_fast && con.verbose>=1)
		cout << "  Using a slightly hacky event choosing algorithm: FastCoal, \nThreshold = " << threshold << "\n";
	if(con.greedy_mle && con.verbose>=1)
		cout << "  Using a NON-STOCHASTIC (and greedy) MLE finding algorithm. \n";
	if((choose_nonrec || first_nonrec)  && con.verbose>=1)
		cout << "  Always using non recombination events wherever possible for speed/max likelihood.\n";
	if(first_nonrec && con.verbose>=1)
		cout << "  Always using the first Coalescence/Mutation found for speed.\n";
	if(!use_logs && (con.verbose>=1 || DEBUG>=1))
		cout << "  NOT Using logs when dynamic programming.\n";
  if(write_trees && con.verbose>=1)
    cout << "  Writing trees to file.\n";
  if((write_trees && con.verbose>=1) || verbose>1)
    cout << "  Directory for output, from argument outputDir = " << outputDir << ".\n";
  if(!write_trees && con.verbose>=2)
    cout << "  Not writing trees to file.\n";
		
	//Creating the output filenames
	string Likelihoods_Filename;		
	Likelihoods_Filename = Create_Likelihoods_Filename();
				
	//*********// 
	int rho_steps = 1;
	if(change_rho) rho_steps = change_rho;
  if(data.size()==0)
    cout << "  Warning! data is of size 0 (in Controller). Problem reading in data?\n";
  int approxNumRuns = ceil(pow(10, round(log10(no_runs))));
  for(rho_step = 0; rho_step<rho_steps; rho_step++)
	{
		Initialise(-1);
		if(rho_step==0)	{			cout << "  Theta per sequence = " << theta << "\n";}
		lsum_IS_weights = -1E308;  lsum_squared_IS_weights = -1E308;
		if(write_full_data || no_runs < too_many_runs)
		{
			lIS_Weights.clear(); 
			lIS_Weights.resize(no_runs);    
			lPr_Probs.clear(); 
			lPr_Probs.resize(no_runs);    
			rec_positions.clear(); 
			rec_positions.resize(no_runs);		
		}
		string Full_Data_Filename;
		Full_Data_Filename = Create_Full_Data_Filename();
		string times_Filename;
		times_Filename	= Create_Times_Filename(0); //The parameter says what type of times info: 1 indicates for Gil's Q*

		non_Markovian_runs = 0; ends_the_same = 0; ends_same_rec = 0; had_rec = 0;
		average_recs = 0;			
		if(con.greedy_mle && no_runs>1) {
		  cout << "  Warning, greedy_mle=true, but no_runs=" << no_runs << "\n";
		  no_runs = 1;
    }
		for(run=0; run<no_runs; run++)
		{
			Display_Run(run, rho_steps, approxNumRuns);
			Initialise(run);		event_vector.clear();
			//*********************************************************************************//
			Simulator();	
			//*********************************************************************************//
			Collate_ARG_Info(run);
			if(write_full_data)
				Write_Full_Data(Full_Data_Filename, run, no_runs, rec_positions.at(run).size());
      if(collect_times)
			{
				Update_Gil_Markov();
				if(collect_times && record_every_time && run == no_runs-1)
					Write_Times(times_Filename, run, Beg_time, End_time);
			}
			if(collect_densities)
				Write_Densities(run, Beg_time, End_time);
		}
		Summarise_Simulations();
	
		End_time = time(NULL);		
		cout << "Time Now = " << End_time << " so time since beginning = " <<  End_time - Beg_time<< "\n";
		cout << "\n\n";					

    if(DEBUG>=1) {
      cout << "  write_likelihoods = "    << write_likelihoods    << "\n";
      cout << "  Likelihoods_Filename = " << Likelihoods_Filename << "\n";
    }
		if(write_likelihoods)
			Write_Likelihoods(Likelihoods_Filename, laverage_IS_weight, no_runs, lsum_squared_IS_weights, Beg_time, Current_time, Last_time, End_time);
	} 
	//End of Rho_Step Loop
		
	// After all runs are finished:
	if(change_rho && write_likelihoods)
		Write_Likelihoods(Likelihoods_Filename, laverage_IS_weight, no_runs, lsum_squared_IS_weights, Beg_time, Current_time, Last_time, End_time);
	if(DEBUG>=2)
	  getchar();
}

void Ancestry::Update_Gil_Markov()
{
	if(rec_records.size()>0)
		had_rec ++;
	if(marg_gens.at(2*smpl_size-2).at(0).c_time == marg_gens.at(2*smpl_size-2).at(rec_sites-1).c_time)
	{
		ends_the_same++;
		if(rec_records.size()>0)
		{
			ends_same_rec++;
			int check_sites;
			for(check_sites=1; check_sites<rec_sites; check_sites++)
			{
				if(marg_gens.at(2*smpl_size-2).at(check_sites-1).c_time != marg_gens.at(2*smpl_size-2).at(check_sites).c_time)
				{
					non_Markovian_runs++;
					//check_sites = rec_sites; 
					break;
				}
			}
		}
	}
}

// This function simulates a genealogy. 
void Ancestry::Simulator()
{
	// These variables are mainly for debugging/display purposes:
	int  break1;
  bool usefulEventNumber = false;
  bool displayThisEvent  = false;
	
  // Required variables to record and manage the genealogy as it's created:
	int daughter1, daughter2; // These are used only in the case of coalescence as the pair which potentially coalesce influences which event finally occurs.
	int rec_lineage;
	int rec_site;
	int mut_lineage;
	int mut_site;
	int new_type;
	double levent_weight;
	
	double	wtng_time;
	int	event; // 1 = Coal, 2 = Mut, 3 = Rec
	
	Imp_Sampler* pointer = 0;
	if(importance_sample)
	{
		switch(sampler)
		{
		case 1:
			pointer = new CSimple_IS(con, rho, rho_per_site, smpl_size, all_sites, no_segsts, rec_sites, DEBUG, SMC, cheap_coal_upwt, using_rec_map, importance_sample, know_derived, site_posns, still_segregating, hap_list, rec_map, ran, coal_fast, greedy_mle);		
			break;
		case 2:
			pointer = new CL_and_S(con, rho, rho_per_site, smpl_size, all_sites, no_segsts, rec_sites, DEBUG, SMC, cheap_coal_upwt, using_rec_map, importance_sample, know_derived, still_segregating, hap_list, rec_map, ran, site_posns, index_live, coal_fast, greedy_mle);
			break;
		case 3:
			pointer = new CC_and_McV(con, rho, rho_per_site, smpl_size, all_sites, no_segsts, rec_sites, DEBUG, cheap_coal_upwt, SMC, using_rec_map, importance_sample, know_derived, still_segregating, hap_list, rec_map, ran, site_posns, index_live, coal_fast, greedy_mle);
			break;
		case 4:
			pointer = new CG_and_M(con, rho, rho_per_site, smpl_size, all_sites, no_segsts, rec_sites, DEBUG, cheap_coal_upwt, SMC, using_rec_map, importance_sample, know_derived, still_segregating, hap_list, rec_map, ran, site_posns, index_live, coal_fast, greedy_mle);
			break;
		default:
			cout << "Error, unable to identify which importance sampler to use, using Li and Stephens style one.\n";
			getchar();
			break;
		}	
	}	
	Imp_Sampler& This_Sampler = *pointer;	


	Calc_index_live();
	//while((!SMC && actv_lngs != 1) || (SMC && !stop_SMC()))
	while(!Stop_Simulation())
	{
		wtng_time	= ran->exponential(1/evnt_rate); // The input parameter of exponential(double) seems to be its mean.
		crnt_height	+= wtng_time;	

		evnt_nmbr ++;
		//temp
		if(display_haps)
			Display_Haplotypes();			
		if(importance_sample)
		{
			This_Sampler.Update_Imp_Sampler(actv_lngs, dead_lngs, live_lineages, index_live, rec_rate, coal_rate, mut_rate, evnt_rate, evnt_nmbr, hap_list);			
			/////////#################/////////////
			////////    Update con    ///////////// //Eventually 'Update Imp Sampler' Will just update con directly, but am keeping it this way for debugging.
			/////////#################/////////////
			con.rec_rate    = rec_rate;
			con.coal_rate   = coal_rate;
			con.mut_rate    = mut_rate;
			con.evnt_rate   = evnt_rate;
			con.crnt_height = crnt_height;
			con.evnt_nmbr   = evnt_nmbr;
			con.actv_lngs   = actv_lngs;
			con.dead_lngs   = dead_lngs;
			con.MRCA_lngs   = MRCA_lngs;
		
			if(choose_hap_first || sampler == 1)
			{
				event = This_Sampler.Choose_Event(daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type, levent_weight, rec_rate, coal_rate, mut_rate, evnt_rate, threshold);
				Update_Event_Vector(event, daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type, levent_weight);
			}
			else
			{
				event = This_Sampler.Choose_from_all_events(daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type, levent_weight);
				if(evnt_nmbr == 1 && event != 1 && 0)
				{
					cout << "Have not coalesced at first event!\n";
					getchar();
				}
				Update_Event_Vector(
				    event        ,
				    daughter1    , daughter2    ,
				    rec_lineage  , rec_site     ,
				    mut_lineage  , mut_site     ,
				    new_type     , levent_weight
				  );
			}   
			//if (1 && display_events) cout << "The event weight this time is: " << exp(levent_weight) << "\n";
		}
		else
		{
			event = Prior_Choose_Event(daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type, levent_weight);
			Update_Event_Vector(event, daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type, levent_weight);				
		}
		if(importance_sample)
		{
			//calc_choice_quotients(levent_weight, choice_quotients, event, wtng_time, rec_site);
			calc_choice_quotients(levent_weight, event, wtng_time, rec_site);
			Bridge_Sample(event, rec_site, levent_weight);
		}
		else
		{
			calc_densities(event, wtng_time, rec_site);
		}
    if(  evnt_nmbr==1   || evnt_nmbr==5     || evnt_nmbr==30 || 
         evnt_nmbr==125 || evnt_nmbr==20000 || evnt_nmbr==50000
         || ((evnt_nmbr % 500 )==0 && evnt_nmbr< 3000 ) 
         || ((evnt_nmbr % 2000)==0 && evnt_nmbr<=10000))
      usefulEventNumber = true;
    else 
      usefulEventNumber = false;
    if(display_events || (verbose>1.5 && usefulEventNumber && no_runs<10))
      displayThisEvent = true;
    else 
      displayThisEvent = false;
		if(displayThisEvent) cout << "Event " << evnt_nmbr << ": ";				
		switch(event)
		{
  		case 1: 
			{//Coalescence
				//if(DEBUG>1 || 1) cout << "Coalescing\n";
				if(displayThisEvent) 
				{
					cout << "Coalescencing lineages " << daughter1 << " and " << daughter2;
					//cout << "Index_Live: ";
					//for (tempI = 0; tempI < index_live.size(); tempI++) cout << " " << index_live.at(tempI);
					//cout << "\n";				
				}
				Coalesce(wtng_time, daughter1, daughter2); // includes update of the various rates
				break;
			}
  		case 2: 
			{//Mutation
				if(displayThisEvent)  cout << "Mutating lineage " << mut_lineage 
				                           << " at site " << mut_site << " to type " << new_type << "";
				Mutate(mut_lineage, mut_site, new_type);
				break;
			}
  		case 3: 
			{//Recombination
				if(displayThisEvent)  cout << "Recombining lineage " << rec_lineage << " at gap " << rec_site << "";
				Recombine(wtng_time, rec_lineage, rec_site); //Recombines on 'daughter1'
				break;
			}
  		default:
			{
				cout << "Unrecognised event type to be simulated in history!" << "\n";
				getchar();
			}
		}	
	  if(displayThisEvent) 
	    cout << " (" << actv_lngs << " left)" << endl; // using endl to flush the buffer and hence check for loops
	  break1 = evnt_nmbr;
	}
	delete pointer;
	
	if(display_haps)
	{
		Display_Hap_info(-1,-1,-1);
		if(actv_lngs > 1){
			if(!SMC) cout<<"Warning! Have stopped before all lineages have coalesced." << "\n";
			//else getchar();
		}
	}
		
	if(display_tree)
	{
		Display_Tree(all_nodes);
	}
	if(write_trees)
	{
		Write_Tree(marg_gens, hapmapDat, rec_records, event_vector, choice_quotients,
		    rec_sites, rho, smpl_size, SMC, all_nodes, site_posns, con, outputDir, writeType);
	}
	int i;
	if(importance_sample)
	{
		lprior_seq_events = 0; limportance_weight = 0;
		signed int minInf = 2147483646;
		for(i=0; i<evnt_nmbr; i++)
		{
			lprior_seq_events     += log(prior_weights.at(i)); //This isn't a density - it's the probability of timeless event sequences
			limportance_weight    += log(choice_quotients.at(i));		
			if(abs(limportance_weight)==numeric_limits<double>::infinity() && DEBUG>0 && abs(minInf)>2147483645)
			  minInf = i;
		}
		if(abs(minInf)<2147483646 && DEBUG>0) {
      cout << "  Warning! limportance_weight=Inf" << "\n";
      cout << "  Index of first problem = " << minInf << " (of " << choice_quotients.size() << ")\n";
      cout << "  choice_quotients.at(" << minInf << ")=" << log(choice_quotients.at(minInf)) << "\n";
		}
	}
	else
	{
		lARG_density = 0;
		for(i=0; i<evnt_nmbr; i++)
			lARG_density += log(event_densities.at(i));
	}
}

void Ancestry::Coalesce(double wtng_time, int daughter1, int daughter2)
{
	int j;
	double old_rec_dist1,old_rec_dist2;
	int new_lineage;
	double new_rec_dist;
	node* crnt_node;
	int daughter_height1; // For debugging, means I can look at these variables.
	int daughter_height2;
	//node* daughter_n1; // These are the daughter NODES, ie tracking the tree, not the types
	//node* daughter_n2;
	int node_index;

	old_rec_dist1	= hap_list.at(daughter1).rec_dist;
	old_rec_dist2	= hap_list.at(daughter2).rec_dist;
	new_lineage		= daughter1;
	
	if(DEBUG>=3)
	{	
			cout << "********************************************************************" << "\n";
			cout << "The daughter lineages are: " << daughter1 << " and " << daughter2 << " and their haplotypes are: " << "\n";
			Display_Hap_info(daughter1,daughter2,-1);
	}

	// Creates the new lineage out of the two coalescing ones. The original 'seq1' is destroyed in the process and seq2 is set to inactive.
	// It also updates the node structure (tree)	
	for (j=0; j<rec_sites; j++)
	{
		// I need to update the list below as currently seq1 or seq2 might not be live lineages.
		if(hap_list.at(new_lineage).alleles.at(j) != hap_list.at(daughter2).alleles.at(j))
		{
			//getchar(); // This cannot happen when rho = 0;
			if (hap_list.at(new_lineage).alleles.at(j) == 0) 
			{
				hap_list.at(new_lineage).alleles.at(j)	= hap_list.at(daughter2).alleles.at(j);
				hap_list.at(new_lineage).nodes.at(j)	= hap_list.at(daughter2).nodes.at(j);
			}
	
			else 
			{
				if (hap_list.at(daughter2).alleles.at(j) == 0) {} // We only keep daughter1 so no need for:  // hap_list.at(daughter2).at(alleles).at(j) = hap_list.at(daughter1).at(alleles).at(j);
				else Coal_Error(new_lineage, daughter2);
			}
		}
		else //Intialise a new node at all sites of overlapping material, deactivate the daughter nodes:
		{
			// Only update the haplotypes and nodes at overlapping ancestral sites (so when they're not 0) and when the MRCA  has not been reached at this site (node != node_MRCA).
			if(Live_Site(hap_list.at(daughter1).alleles.at(j)) && which_node.at(j) < node_MRCA-1)
			{
				which_node.at(j) ++;	  if(which_node.at(j) >= (int)marg_gens.size()){
					cout << "Warning! Node index has indexed a non-existant node!" << "\n";
					getchar();
					exit(1);
				}
				node_index				= which_node.at(j);
				crnt_node				= &marg_gens.at(node_index).at(j);	//For short vairable names
				daughter_height1		= hap_list.at(daughter1).nodes.at(j); // NOTE! daughter1 is both the OLD haplotype and the index of the one being created!! Requires care.
				daughter_height2		= hap_list.at(daughter2).nodes.at(j);

				crnt_node->n1			= &marg_gens.at(daughter_height1).at(j);
				crnt_node->n2			= &marg_gens.at(daughter_height2).at(j);
				crnt_node->c_time		= crnt_height;
				crnt_node->time_below	= crnt_node->n1->time_below + crnt_node->n2->time_below + 2.0*crnt_height - crnt_node->n1->c_time - crnt_node->n2->c_time;
				crnt_node->active		= 1;
				crnt_node->index_site	= j;
				crnt_node->index_epoch	= which_node.at(j);				
				crnt_node->n1->active	= 0;
				crnt_node->n2->active	= 0;
			
				hap_list.at(new_lineage).nodes.at(j) = node_index;
				//cout << "n1: site= " << crnt_node->n1->index_site << "    leaf/epoch = " << crnt_node->n1->index_epoch << "\t";
				//cout << "n2: site= " << crnt_node->n2->index_site << "    leaf/epoch = " << crnt_node->n2->index_epoch << "\n";
				if((int)which_node.at(j) == (int)(marg_gens.size()-1))
					hap_list.at(new_lineage).alleles.at(j) *= -1;
			}
		}
	}

	// Updating anc_mat boundaries as far as rec rate is concerned, note that the 'else' statements have instructinos as well.
	if(hap_list.at(daughter1).left_rec_edge < hap_list.at(daughter2).left_rec_edge)	 
		hap_list.at(new_lineage).left_rec_edge = Find_Left_Edge(hap_list.at(daughter1), hap_list.at(new_lineage));
	else
		hap_list.at(new_lineage).left_rec_edge  = Find_Left_Edge(hap_list.at(daughter2), hap_list.at(new_lineage));
	
	//cout << "Daughter1 edges: " << hap_list.at(daughter1).left_rec_edge << " " << hap_list.at(daughter1).right_rec_edge << "\n";
	if(hap_list.at(daughter1).right_rec_edge >hap_list.at(daughter2).right_rec_edge) 
		hap_list.at(new_lineage).right_rec_edge = Find_Right_Edge(hap_list.at(daughter1), hap_list.at(new_lineage));
	else																			 
		hap_list.at(new_lineage).right_rec_edge = Find_Right_Edge(hap_list.at(daughter2), hap_list.at(new_lineage));

	if(hap_list.at(new_lineage).right_rec_edge == no_segsts + 1 || hap_list.at(new_lineage).left_rec_edge == no_segsts + 1 ) // This should happen only when sequence has reached its MRCA // Can CHECK this using NODES?
	{
		if(hap_list.at(new_lineage).left_rec_edge != no_segsts + 1 || hap_list.at(new_lineage).right_rec_edge != no_segsts + 1 )
		{
			cout << "Warning, inconsistent information on whether hap has reached MRCA everywhere \n(Info from left_anc_edge calculators).\n";
			getchar();
		}
		hap_list.at(new_lineage).reached_MRCA = 1;
		hap_list.at(new_lineage).rec_dist = 0;
		live_lineages.at(new_lineage) = -1; // Need to update checks on live_lineages to REJECT things that are -1.
		MRCA_lngs++; actv_lngs --;
		new_rec_dist = 0;
	}

	else
	{
		new_rec_dist = rec_map.at(hap_list.at(new_lineage).right_rec_edge) - rec_map.at(hap_list.at(new_lineage).left_rec_edge);				
		hap_list.at(new_lineage).rec_dist = new_rec_dist;
	}
	
	//House keeping:
	live_lineages.at(daughter2)	= 0; dead_lngs++; actv_lngs --; // This temporarily tells the program not to read a certain bit of allocated memory. It does not affect nodes.
	Calc_index_live();			
	evnt_rate = Update_Rates(1, new_lineage, -1, old_rec_dist1, old_rec_dist2, new_rec_dist); // 1st 3 params are: event (1=coal), lin1, lin2,	
}

void Ancestry::Calc_index_live()
{
	unsigned int i;
	index_live.resize(0);
	//Compile a list of indexes of active lineages.
	for(i=0; i<hap_list.size(); i++)
	{
		if(Lineage_Alive(i)) // This specific check avoids counting lineages that have reached their MRCA. Perhaps the if statement should be done with a function
			index_live.push_back(i);
	}
}

// This function checks for shared ancestral sites on ALREADY (Live) INDEXED lineages.
int Ancestry::Check_Coal(int daughter1, int daughter2)
{
	if(!(hap_list.at(daughter1).right_rec_edge < hap_list.at(daughter2).left_rec_edge || hap_list.at(daughter2).right_rec_edge < hap_list.at(daughter1).left_rec_edge))
		return 1; //There is overlapping material, so they can coalesce
	else
		return 0; //There is no overlapping material
}

// This function checks to see if 'live_index' is '1' at the index concerned. If it is, the lineage is still live. 
// Otherwise it has coalesced away (so doesn't really exist) or has reached its MRCA (need no longer be considered.)
inline bool Ancestry::Lineage_Alive(const int& lineage_index)
{
	if(live_lineages.at(lineage_index) == 1)
		return 1;
	else 
		return 0;
}

// This function checks to see if 'live_index' is '0' at the index concerned. If it is, the lineage is complete junk. 
// This lineage has coalesced away (so doesn't really exist, and the data inside is complete junk)
// Live lineages and MRCA lineages are still stored.
inline bool Ancestry::Junk_Lineage(const int& lineage_index)
{
	if(live_lineages.at(lineage_index) == 0)
		return 1;
	else 
		return 0;
}

// If all of the root nodes are active then stop the simulation. Returning 0 means that we don't stop.
int Ancestry::stop_SMC()
{
	int i,j,k;
	int node_stop;
	int haps_stop;
		
	// Work out whether to stop according to active root nodes
	node_stop = 1;
	for(i=0;i<rec_sites;i++)
	{
		if(!marg_gens.at(node_MRCA-1).at(i).active)
		{
			//return 0;
			node_stop = 0;
			i=rec_sites;
		}
	}
	
	// Work out whether to stop according to whether or not any further coalescence is possible.
	haps_stop = 1;
	for(i=0;i<actv_lngs;i++)
	{
		for(j=0;j<i;j++)
		{
			for(k=0; k< rec_sites; k++)
			{
				if(Live_Site(hap_list.at(index_live.at(i)).alleles.at(k)) && Live_Site(hap_list.at(index_live.at(j)).alleles.at(k)))
				{
					haps_stop = 0;
					j=actv_lngs;i=actv_lngs;k=rec_sites; //ie. STOP
				}
			}
		}
	}
	if(haps_stop != node_stop)
	{
		if(node_stop) 
			cout << "Warning: All root nodes active but some haplotypes share an ancestral site!" << "\n"
				 << "So more coalescence events are possible here." << "\n";
		else
			cout << "Warning: No haplotypes share an ancestral site and yet not all terminal nodes are yet active.." << "\n"
				 << "This is an ERROR. All marginal trees should reach MRCA under all both Coal and SMC." << "\n";
		
		Display_Hap_info(-1,-1,-1);
		getchar();
	}
	return node_stop;	
}

inline bool Ancestry::Stop_Simulation()
{
	if(actv_lngs)
		return 0;
	else
		return 1;
}

//This function calculates the PRIOR coalescence rate (under the SMC), this has nothing to do with TYPES or 'data'.
int Ancestry::Count_Coal_Pairs()
{
	int i,j;
	int no_coal_pairs;
	
	no_coal_pairs = 0;
	if(!SMC)
	 cout << "Warning, have called an SMC function when the SMC flag is off!\n";
	for(i=0; i<actv_lngs; i++)
	{
		for(j=0; j<i; j++)
		{
			//no_coal_pairs += Simple_IS.Check_Coal(index_live.at(i),index_live.at(j));
			no_coal_pairs += Check_Coal(index_live.at(i),index_live.at(j));
		}
	}
	return no_coal_pairs;
}

//This function calculates the PRIOR mutation rate, this has nothing to do with TYPES or 'data'.
double Ancestry::Calc_Mut_Rate()
{
	int i,j;
	int no_mutable_sites = 0;
	//double theta_per_site = theta/(double)no_segsts;

	for(i=0; i<actv_lngs; i++)
	{
		for(j=0; j<no_segsts; j++)
		{
			if(Live_Site(hap_list.at(index_live.at(i)).alleles.at(j)))
				no_mutable_sites++;
		}
	}
	mut_rate = (theta_per_site/2.0) * no_mutable_sites;
	return mut_rate;
}

double Ancestry::Update_Rates(int event, int new_lineage1, int new_lineage2, double old_rec_dist1, double old_rec_dist2, double new_rec_dist)
{
	//double evnt_rate;
	double old_rec_rate = rec_rate;
	
	// Update event rates more efficiently using knowledge of the previous event:
	switch(event)
		{
		case 1: 
			{//Coalescence

				rec_rate  = old_rec_rate - old_rec_dist1/2.0 -  old_rec_dist2/2.0 + new_rec_dist/2.0;
				
				if(SMC)
					coal_rate = Count_Coal_Pairs();
					//coal_rate = (double)actv_lngs * ((double)actv_lngs - 1)/2.0; // This is only correct when some coalescence events are rejected and then another TIME is is resmapled.
				else
					coal_rate = (double)actv_lngs * ((double)actv_lngs - 1)/2.0;
				
				//mut_rate = (double)actv_lngs * (double)no_segsts * theta_per_site/2.0; // This is wrong! Needs to be updated according to number of overlapping sites in coalescence event.
				mut_rate = Calc_Mut_Rate(); // This is slow! Would be faster to update according to number of overlapping sites in coalescence events.
				break;

			}
		case 2: 
			{//Recombination
				//mut_rate is unchanged.
				
				if(SMC)
				{
					// Does the REC_rate ever change after a rec under the SMC?					
					coal_rate = Count_Coal_Pairs();
					//*** Fix rec_rate ***//
					rec_rate  = old_rec_rate - rec_mat_lost/2.0; // rec_mat_lost also accounts for the rebraking up of trapped non-ancestral material.					
				}
				else
				{
					rec_rate  = old_rec_rate - rec_mat_lost/2.0; // rec_mat_lost also accounts for the rebraking up of trapped non-ancestral material.
					coal_rate = (double)actv_lngs * ((double)actv_lngs - 1)/ 2.0;
				}
				break;
			}
		case 3: 
			{//Mutation
				// If the prior rate of mutation is positive and the IS functions can find singletons then mutations are possible.
				break;
			}
		}

	evnt_rate = rec_rate + coal_rate + mut_rate;		
	return evnt_rate;
}

//Does recombination in any way affect the node structure? I don't think so.
void Ancestry::Recombine(double wtng_time, int& rec_lineage, int& rec_site)
{
	int new_lineage;
	recomb_event this_recomb;
	
	if(importance_sample)
	{
		//Simple_IS.Choose_rec_details(rec_lineage, rec_site, rec_to);		
	}
	else
	{
		rec_lineage = Prior_Choose_Rec_Lineage();
		Prior_Choose_position(rec_lineage, rec_site);	
	}
	
	new_lineage = Create_new_Lineage(rec_lineage);
	Update_hap_list(rec_lineage, new_lineage, rec_site);
	Calc_index_live();				
	Update_Rates(2, rec_lineage, new_lineage, -1, -1, -1);

	this_recomb.lineage = rec_lineage;
	this_recomb.epoch	= evnt_nmbr;
	this_recomb.r_time	= crnt_height;
	rec_records.push_back(this_recomb);	
	total_no_recs++; //This really is the TOTAL for this rho - summed over all runs
}

int Ancestry::Create_new_Lineage(int rec_lineage)
{
	int i;
	int new_lineage = std::numeric_limits<int>::quiet_NaN();

//checks:
	if(!((int)hap_list.size() == dead_lngs + actv_lngs + MRCA_lngs) ||
			dead_lngs < 0                                           ||
			actv_lngs > (int)hap_list.size()                        ||
			hap_list.size() != live_lineages.size()                  )
	{
		cout << "Warning: inconsistency in the number of live lineages" << "\n"; 
		getchar();
	}
	
	if(dead_lngs)
	{
		// It's probably more efficient to count these backwards as later lineages are more likely to be set to zero.
		for(i=0; i<(int)hap_list.size(); i++)
		{
			if(Junk_Lineage(i)) // This will overwrite 'dead' lineages, but not those that have reached their MRCA.
			{
				new_lineage = i;
				live_lineages.at(i) = 1;
				i=hap_list.size();//ie. break out of loop.
			}
		}
		dead_lngs--; 
	}
	else
	{
		haplotype blank_hap;	
		blank_hap.alleles.resize(rec_sites,0);	
		blank_hap.nodes.resize(rec_sites);	
		hap_list.push_back(blank_hap);
		new_lineage = hap_list.size()-1;
		live_lineages.push_back(1);
	}
	actv_lngs ++;

	//Error checking. Check them in this function to keep 'Recombine' clear of debris.
	if(rec_lineage<0 || new_lineage<0 || rec_lineage>(int)live_lineages.size()|| new_lineage>(int)live_lineages.size())
	{cout << "Warning! Have failed to choose a valid daughter lineages" << endl;getchar();exit(1);}
	return new_lineage;
}

// Recombination Function: This removes ancestral material from rec_ineage and initialises new_lineage
void Ancestry::Update_hap_list(int rec_lineage, int new_lineage, int& rec_site)
{
	int i,k;
	int stopL, stopR;
	int l_anc_edge, r_anc_edge; // the l's and r's here are confusing. These refer to the left and right of the RECOMBINATION. They then become the R and L edges respec.

	//Note: if rec_site = 0 the the rec happens between the zeroth and 1st sites. That way rec_locus takes all values in {0,1,2,...} so is indexd as C++ likes
	//Calculate left and right hand boundaries for the new lineages, this affects the new recombination rate as well.
	if(DEBUG>=3)
	{
		cout << "\n";
		cout << "Event no.: " << evnt_nmbr << "(recombination).\nThe recombining haplotype reads: " << "\n";
		for (int k=0; k<rec_sites; k++) {cout << hap_list.at(rec_lineage).alleles.at(k) << " ";}
		cout << "\n";
		cout << "and the recombination distance on this haplotype is: " << hap_list.at(rec_lineage).rec_dist << "\n";
		cout << "Locus (gap between sites) of recombination: " << rec_site << "\n";		
	}

	stopL = 0; 
	l_anc_edge = rec_site;
	while(!stopL)
	{
		if (Live_Site(hap_list.at(rec_lineage).alleles.at(l_anc_edge)))
			stopL = 1;
		else
			l_anc_edge --;
		if (l_anc_edge<0){
			cout << "Warning! Have been unable to identify the first ancestral site to the left of the recombination break_point" << "\n";
			cout << "The relevant haplotype reads: " << "\n";
			for (int k=0; k<rec_sites; k++) {cout << hap_list.at(rec_lineage).alleles.at(k) << " ";}
			cout << "\n";
			cout << "Rec distance on this haplotype is: " << hap_list.at(rec_lineage).rec_dist << "\n";
			getchar();exit(1);
		}	
	}
	hap_list.at(new_lineage).right_rec_edge = hap_list.at(rec_lineage).right_rec_edge; // Have to set daughter 2 here before daughter 1 gets changed.
	hap_list.at(rec_lineage).right_rec_edge = l_anc_edge; // The left of the rec event is the RIGHT of the left hand haplotype.
	hap_list.at(rec_lineage).rec_dist = rec_map.at(hap_list.at(rec_lineage).right_rec_edge) - rec_map.at(hap_list.at(rec_lineage).left_rec_edge);				

	stopR = 0; r_anc_edge = rec_site+1;
	while(!stopR)
	{
		//if (hap_list.at(daughter1).alleles.at(rec_locus+1))
		if (Live_Site(hap_list.at(rec_lineage).alleles.at(r_anc_edge)))
			stopR = 1;
		else
			r_anc_edge ++;
		if (r_anc_edge>=rec_sites){
			cout << "Warning! Have been unable to identify the first ancestral site to the right of the recombination break_point" << "\n";
			cout << "The relevant haplotype reads: " << "\n";
			for (int k=0; k<rec_sites; k++) {cout << hap_list.at(rec_lineage).alleles.at(k) << " ";}
			cout << "\n" << "Rec_Dist on this haplotype is: " << hap_list.at(rec_lineage).rec_dist << "\n";
			cout << "Rho is: " << rho << " and left_rec_edge = " << hap_list.at(rec_lineage).left_rec_edge << " and right_rec_edge = " << hap_list.at(rec_lineage).right_rec_edge << "\n";
			cout << "Gap attempted to recombine at: " << rec_site << "\n";
			getchar();exit(1);
		}

	}
	hap_list.at(new_lineage).left_rec_edge = r_anc_edge;	// The right of the rec event is the LEFT of the left hand haplotype.
	//cout << "The edges for daughter 2 are: " << hap_list.at(daughter2).right_rec_edge << " and: " << hap_list.at(daughter2).left_rec_edge << "\n";
	hap_list.at(new_lineage).rec_dist = rec_map.at(hap_list.at(new_lineage).right_rec_edge) - rec_map.at(hap_list.at(new_lineage).left_rec_edge);				
	
	rec_mat_lost = rec_map.at(r_anc_edge) - rec_map.at(l_anc_edge);

	//Daughter1 takes the material to the left of the recombination point and daughter2 takes the amterial to the right.
	for (i=0; i< rec_site+1; i++) // rec_locus: the rec occurred in the rec_locus th gap. Rec_sites is the total number of sites simulation is happening to.
	{
		hap_list.at(new_lineage).alleles.at(i) = 0;
	}		
	for (i=rec_site+1; i< rec_sites; i++) // rec_locus: the rec occurred in the rec_locus th gap. Rec_sites is the total number of sites simulation is happening to.
	{
		hap_list.at(new_lineage).alleles.at(i) = hap_list.at(rec_lineage).alleles.at(i);
		hap_list.at(rec_lineage).alleles.at(i) = 0;
	}	

	//Update node indexes on daughter2: //Note that this leaves lots of spurious node indices on d1 and d2 but that these should never be read as the relevant alleles will be 0.
	// This scheme of not updating node indices at allele=0 sites is faster to run but could lead to bugs.
	for (i=hap_list.at(new_lineage).left_rec_edge; i<hap_list.at(new_lineage).right_rec_edge+1; i++)// Remember there are more sites than rec break points!
	{
		hap_list.at(new_lineage).nodes.at(i) = hap_list.at(rec_lineage).nodes.at(i);
	}
	if(DEBUG>=3)
	{
		cout << "The resulting haplotypes read: " << "\n";
		for (k=0; k<rec_sites; k++) {cout << hap_list.at(rec_lineage).alleles.at(k) << " ";}
		cout << "\n";
		cout << "Left edge on rec_lineage is: " << hap_list.at(rec_lineage).left_rec_edge << " \t and right edge is: " << hap_list.at(rec_lineage).right_rec_edge << "\n";	
		cout << "Rec_Dist on rec_lineage is: " << hap_list.at(rec_lineage).rec_dist << "\n" << "\n";		
		for (k=0; k<rec_sites; k++) {cout << hap_list.at(new_lineage).alleles.at(k) << " ";}
		cout << "\n";
		cout << "Left edge on new_lineage is: " << hap_list.at(new_lineage).left_rec_edge << " \t and right edge is: " << hap_list.at(new_lineage).right_rec_edge << "\n";	
		cout << "Rec_Dist on new_lineage is: " << hap_list.at(new_lineage).rec_dist << "\n" << "\n";		
	}	
}

// This mutates the singleton to the type of the rest of the sequence.
void Ancestry::Mutate(const int& mut_lineage, const int& mut_site, const int& new_type)
{
	hap_list.at(mut_lineage).alleles.at(mut_site) = new_type;
//	int i;
	if(still_segregating.at(mut_site))
		still_segregating.at(mut_site) = 0;
	else
	{
		cout << "Warning, non segregating sites should not mutate.\n";
		getchar();
	}
	//for(i=0; i<no_segsts; i++)
		//cout << still_segregating.at(i) << " ";
	//cout << "\n";
	
}

void Ancestry::Display_Haplotypes(bool useIndexLive)
{
	int i,j;
	int no_to_display;
	int max_lngs = 15;
	int length_to_display;
	int max_length = 25;
	int display_indices = 0;
	
	if(actv_lngs>max_lngs)
	{
		cout << "Warning, there are a lot of lineages to display (" << actv_lngs << "), displaying first " << max_lngs << "\n";
		no_to_display = max_lngs;
	}
	else no_to_display = actv_lngs;

	if(no_segsts>max_length)
	{
		cout << "Warning, there are a lot of sites to display (" << no_segsts << "), displaying first " << max_length << "\n";
		length_to_display = max_length;
	}
	else length_to_display = no_segsts;

	cout << "The haplotypes in epoch " << evnt_nmbr << " are: " << "\n";
	for (i=0; i< no_to_display; i++)
	{
		for (j=0; j<length_to_display; j++) {
			if (useIndexLive)
        cout  << hap_list.at(index_live.at(i)).alleles.at(j) << " ";
      else {
        cout  << hap_list.at(i).alleles.at(j) << " ";
      }
    }
		cout << "\n";
	}
	if(display_indices)
	{
		cout << "\nThe indices of the live lineages are:\n";
		for (i=0; i< no_to_display; i++) {
			if (useIndexLive) {
        cout  << index_live.at(i) << " ";
      }
      else {
        cout  << i << " ";
      }
      cout << "\n";	
    }
  }
  return;
}

//void Ancestry::calc_choice_quotients(double& levent_weight, vector<double>& choice_quotients, int event, const double& wtng_time, const int& rec_site)
void Ancestry::calc_choice_quotients(double& levent_weight, int event,
    const double& wtng_time, const int& rec_site) {
  //double time_density;

  // Note that the coefficient will be cancelled out in the next line.
  // time_density = evnt_rate * exp(-1.0*(evnt_rate * wtng_time));
  switch (event) {
  case 1:
    // The coalescence rate cancels out with itself for a SPECIFIC coalescence event.
    prior_weights.push_back(1.0 / evnt_rate);
    break;
  case 2:
    // The mutation rate cancels out with itself leaving theta_per_site for a SPECIFIC mutation event.
    prior_weights.push_back((theta_per_site / 2.0) / evnt_rate);
    break;
  case 3:
    // The recombination rate cancels out with itself leaving rho for a SPECIFIC rho_per_site event.
    prior_weights.push_back(
      ((rec_map.at(rec_site + 1) - rec_map.at(rec_site)) / 2.0)
      /evnt_rate
    );
    break;
  default:
    cout << "Warning, could not decide on a weight for the current event. \n";
    getchar();
  }
  // importance weight for this (event) = priorProbability(event)/proposalProbability(event)
  // The prior weights are pushed back each event, so the one at the end refers to the current event.
  choice_quotients.push_back(prior_weights.at(prior_weights.size() - 1) /
      exp(levent_weight));
  if(abs(log(choice_quotients.at(choice_quotients.size()-1)))>1e50 && (DEBUG>2 || verbose >=2)) {
    cout << "  Warning, event very unlikely under the prior (and (DEBUG>2 || verbose >=2))\n";
    cout << "  choice_quotients.at(" << choice_quotients.size()-1 << ") = " <<
            choice_quotients.at(choice_quotients.size()-1) << "\n";
    cout << "  levent_weight           = " << levent_weight << "\n";
    cout << "  prior_weights.at(prior_weights.size()-1 = " << prior_weights.at(prior_weights.size() - 1) << "\n";
    cout << "  event    = " <<  event << " (1,2,3 = coal,mut,rec)" << ", evnt_rate = " << evnt_rate << "\n";
    cout << "  recRate  = 1/2(" << rec_map.at(rec_site + 1) << " - " << rec_map.at(rec_site) << ")\n";
    cout << "  rec_site = " << rec_site << "\n";
    if(DEBUG>=1.999)
      getchar();
  }
  //if (choice_quotients.at(choice_quotients.size()-1) <0 || (DEBUG || display_weights))
  if (choice_quotients.at(choice_quotients.size() - 1) < 0 || display_weights) {
    cout << "event_weight = " << exp(levent_weight) << "\n";
    cout << "prior_weights.at(" << prior_weights.size() - 1 << ") = "
        << prior_weights.at(prior_weights.size() - 1) << "\n";
    cout << "Choice_quotients at " << choice_quotients.size() - 1 << " = "
        << choice_quotients.at(choice_quotients.size() - 1) << "\n";
    //getchar();
  }
}

void Ancestry::calc_densities(int event, const double& wtng_time, const int& rec_site)
{
	 switch(event)
     {
     case 1:
         event_densities.push_back(exp(-1.0*(evnt_rate * wtng_time))); // The coalescence rate cancels out with itself for a SPECIFIC coalescence event.
		 break;
     case 2:
		 event_densities.push_back((theta_per_site/2.0) * exp(-1.0*(evnt_rate * wtng_time))); // The mutation rate cancels out with itself leaving theta_per_site for a SPECIFIC mutation event.
		 break;
     case 3:
         event_densities.push_back(rho_per_site/2.0 * exp(-1.0*(evnt_rate * wtng_time))); // The recombination rate cancels out with itself leaving rho for a SPECIFIC rho_per_site event.
		 break;
     default:
             cout << "Warning, could not decide on a weight for the current event. \n";
			 getchar();
	}
}

void Ancestry::Define_Bridge_Rhos()
{
	int i;
	if(bridge_resolution<1) {
	  cout << "  Warning! bridge_resolution must be greater than zero. Problem with program arguments or defaults file?\n";
	  cout << "  Press any key to continue:";
	  getchar();
	  cout << "\n";
	}
	bridge_rhos.resize(bridge_resolution);

	if(bridge_resolution > 4)
	{
		bridge_rhos.at(0) = (double)1e-9;			
		for(i=1; i<bridge_resolution; i++)
		{
			if(i<10)
				bridge_rhos.at(i) = (double)(i)/10.0;
			else
			{
				if(10 <= i && i < 20)
					bridge_rhos.at(i) = 1.0 + (double)(i-10)/5.0;
				else
					bridge_rhos.at(i) = i-17;
			}
		}
		bridge_rhos.at(bridge_rhos.size()-1) = 100;
	}
	else
	{
		bridge_rhos.at(0) = (double)1e-9;					
		for(i=1; i<bridge_resolution; i++)
			bridge_rhos.at(i) = (double)(i+1);
	}
}

void Ancestry::Bridge_Sample(const int& event, const int& rec_site, const double& levent_weight)
{
	 int i;
	 double bridge_rho;
	 double bridge_density = std::numeric_limits<double>::quiet_NaN();
	 // Need to specify a sampler - Otherwise I don't know the probability of suggesting an ARG. Decided (arbitrarily) on rho=1 as best for this.
	 for (i=0; i<bridge_resolution; i++)
	 {
		 bridge_rho = bridge_rhos.at(i);
		 switch(event)
		 {
		 case 1:
			 bridge_density = (1.0/(mut_rate + coal_rate + (rec_rate * (bridge_rho/rho)))); // The coalescence rate cancels out with itself for a SPECIFIC coalescence event.
			 break;
		 case 2:
			 bridge_density = ((theta_per_site/2.0)/(mut_rate + coal_rate + (rec_rate * (bridge_rho/rho)))); // The mutation rate cancels out with itself leaving theta_per_site for a SPECIFIC mutation event.
			 break;
		 case 3:
			 bridge_density = (((rec_map.at(rec_site+1)-rec_map.at(rec_site))*(bridge_rho/rho)/2.0)/(mut_rate + coal_rate + (rec_rate * (bridge_rho/rho)))); // The recombination rate cancels out with itself leaving rho for a SPECIFIC rho_per_site event.
			 break;
		 default:
				 cout << "Warning, could not decide on a weight for the current event. \n";
				 getchar();
		}		
		lbridge_weights.at(rho_step).at(i)	+= log(bridge_density/exp(levent_weight));		
	 }
}

// *Warning* What on earth is this doing? Should it ever get called?
void Ancestry::calc_time_in_tree()
{
    int j=0;
	while (j<rec_sites)
		j++;	
}     

void Ancestry::Coal_Error(int new_lineage, int daughter2)
{	
	cout << "Warning, Both sequences are non zero and yet they differ here. \nCoalescence should only occurr when sequence are the same." << "\n"; 
	cout << "Daughter 1 = " << new_lineage << " and daughter2 = " << daughter2 << "\n";
	cout << "\nIndex live reads:  ";
	for (int i=0; i<actv_lngs; i++)
	{
		cout << index_live.at(i) << " ";
	}
	cout << "\n";
	getchar();
}

void Ancestry::Display_Hap_info(int hap1, int hap2, int hap3)
{
	int i,k;
	
	if((hap1+1) || (hap2+1))
	{		
		for (k=0; k<rec_sites; k++) {cout << hap_list.at(hap1).alleles.at(k) << " ";}
		cout << "\n";
		cout << "Left edge on this lineage, " << hap1 << ", is: " << hap_list.at(hap1).left_rec_edge << " \t and right edge is: " << hap_list.at(hap1).right_rec_edge << "\n";	
		cout << "Rec_Dist on this lineage, " <<  hap1 << ", is: " << hap_list.at(hap1).rec_dist << "\n" << "\n";				
		for (k=0; k<rec_sites; k++) {cout << hap_list.at(hap2).alleles.at(k) << " ";}
		cout << "\n";
		cout << "Left edge on this lineage, " << hap2 << ", is: " << hap_list.at(hap2).left_rec_edge << " \t and right edge is: " << hap_list.at(hap2).right_rec_edge << "\n";	
		cout << "Rec_Dist on this lineage, "  << hap2 << ", is: " << hap_list.at(hap2).rec_dist << "\n" << "\n";	
		if((hap3+1))
		{
			for (k=0; k<rec_sites; k++) {cout << hap_list.at(hap3).alleles.at(k) << " ";}
			cout << "\n";
			cout << "Left edge on this lineage, " << hap3 << ", is: " << hap_list.at(hap3).left_rec_edge << " \t and right edge is: " << hap_list.at(hap3).right_rec_edge << "\n";	
			cout << "Rec_Dist on this lineage, "  << hap3 << ", is: " << hap_list.at(hap3).rec_dist << "\n" << "\n";				
		}
	}
	else
	{	
		cout << "********************************************************************" << "\n";
		cout << "Rec_rate in epoch " << evnt_nmbr << " = " << rec_rate << "\n" << "\n";
		cout << "In epoch: " << evnt_nmbr << " there are " << actv_lngs << " active lineages and the haplotypes read: " << "\n";
			
		for(i=0; i<actv_lngs; i++)
		{
			for (k=0; k<rec_sites; k++) {cout << hap_list.at(index_live.at(i)).alleles.at(k) << " ";}
			cout << "\n";
			//cout << "Left edge on this lineage, " << index_live.at(i) << ", is: " << hap_list.at(index_live.at(i)).left_rec_edge << " \t and right edge is: " << hap_list.at(index_live.at(i)).right_rec_edge << "\n";	
			//cout << "Rec_Dist on this lineage, "  << index_live.at(i) << ", is: " << hap_list.at(index_live.at(i)).rec_dist << "\n" << "\n";				
		}
	}
}

void Ancestry::Display_Tree(bool all_nodes)
{
	int i,j;
	int start_node;

	if(all_nodes)
		start_node = 0;
	else
		start_node = smpl_size;
	
	
	//To draw tree nodewise:
	//while ( Chromosome->n1 != NULL && Chromosome->n2 != NULL)
	//{ draw next line and update Chromosome}


	//cout << "Displaying tree at site 0: " << "\n";
	//cout << "Displaying types history: " << "\n";
	cout << "node \tDscndt1 Dscndt2 Time \tTotal time below \tActive?" << "\n" << "\n";
	//for (i=0; i<2*smpl_size-1; i++)
	j=0;
	while (j<rec_sites)
	{
		cout << "Site: " << j << " at which rho = " << rec_map.at(j) << "\n" << "\n";
		for (i=start_node; i<2*smpl_size-1; i++)
		{
			cout << i+1 << "\t" << "   " ;
			if (marg_gens.at(i).at(j).n1)
				cout << marg_gens.at(i).at(j).n1->index_epoch+1		<< "\t   ";
			else
				cout << "x\t   ";
			if (marg_gens.at(i).at(j).n2)
				cout << marg_gens.at(i).at(j).n2->index_epoch+1		<< "\t";		
			else
				cout << "x\t";

			if ( i >smpl_size-1 ) {cout.precision(3); cout << marg_gens.at(i).at(j).c_time		<< "\t    ";}
			//if ( i >smpl_size-1 ) {cout.width(2); cout << marg_gens.at(i).at(0).c_time		<< "\t";}
			else cout << marg_gens.at(i).at(j).c_time		<< "\t\t    ";		
			if ( i >smpl_size-1 ) cout << marg_gens.at(i).at(j).time_below<< "\t\t   ";
			else cout << marg_gens.at(i).at(j).time_below<< "\t\t";
			cout << marg_gens.at(i).at(j).active	<< "\n";		
		}
		cout << "\n"<< "\n";
		if(rec_sites>3)
			j += (rec_sites/4);
		else
			j++;
	}
	cout << "\n" << "\n";
}

void Ancestry::Update_Event_Vector(const int& event, int& daughter1, int& daughter2,
    int& rec_lineage, int& rec_site, int&    mut_lineage,
    int& mut_site   , int& new_type, double& levent_weight)
{
	CEvent this_event;

	this_event.daughter1     = daughter1;
	this_event.daughter2     = daughter2;
	this_event.event_type    = event;
	this_event.mut_lineage   = mut_lineage;
	this_event.mut_site      = mut_site;
	this_event.new_type      = new_type;
	this_event.rec_lineage   = rec_lineage;
	this_event.rec_site      = rec_site;
	this_event.levent_weight = levent_weight;
	event_vector.push_back(this_event);
}

// This function is used to identify sites which are ancestral. Even those which have reached their MRCA.
inline int Ancestry::Ancestral(int allele)
{
	switch(allele)
	{
	case -2:		//This has reached its MRCA. It is ancestral. (and of type 2, (if(know_derived) this is mutant))
		return 1;
	case -1:		//This has reached its MRCA. It is ancestral. (and of type 1)
		return 1;
	case 0:			//This site is non ancestral.
		return 0;
	case 1:			//This site is ancestral (and of type 1)
		return 1;
	case 2:			//This site is ancestral (and of type 2) (if(know_derived) this is mutant)
		return 1;
	default:
		cout << "Unrecognised type in alleles vector.\n";
		getchar();
		return -1;
	}
}

// This function is used to identify sites which are ancestral and have not reached their MRCA.
inline int Ancestry::Live_Site(int allele)
{
	switch(allele)
	{
	case -2:		//This has reached its MRCA. It is ancestral. (and of type 2, (if(know_derived) this is mutant))
		return 0;
	case -1:		//This has reached its MRCA. It is ancestral. (and of type 1)
		return 0;
	case 0:			//This site is non ancestral.
		return 0;
	case 1:			//This site is ancestral (and of type 1)
		return 1;
	case 2:			//This site is ancestral (and of type 2) (if(know_derived) this is mutant)
		return 1;
	default:
		cout << "Unrecognised type in alleles vector.\n";
		getchar();
		return -1;
	}
}

inline int Ancestry::Find_Left_Edge(const haplotype& daughter_hap, const haplotype& new_hap)
{
	int j;
	for(j=daughter_hap.left_rec_edge; j<rec_sites; j++)
	{
		if(Live_Site(new_hap.alleles.at(j)))
			return j;
	}
	//All ancestral sites on this lineage have now reached their MRCA, this lineage is finished with for simulation purposes
	return no_segsts + 1;
}

inline int Ancestry::Find_Right_Edge(const haplotype& daughter_hap, const haplotype& new_hap)
{
	int j;
	for(j=daughter_hap.right_rec_edge; j>=0; j--)
	{
		if(Live_Site(new_hap.alleles.at(j)))
			return j;
	}
	//All ancestral sites on this lineage have now reached their MRCA, this lineage is finished with for simulation purposes
	return no_segsts + 1;
}

void Ancestry::Output_Sampler()
{
	switch(sampler)
	{
	case 1:
		cout << "Using the Cardin Simple IS scheme\n";
		break;
	case 2:
		cout << "Using Li and Stephens style LDaC Importance Weights\n";
		break;
	case 3:
		cout << "Using Cardin and McVean style LDaC Importance Weights\n";
		break;
	case 4:
		cout << "Using Griffiths and Marjoram Importance Weights\n";
		break;
	default:
		cout << "Error, unable to identify which importance sampler to use, using Li and Stephens style one.\n";
		getchar();
		break;
	}	
}

void Ancestry::Write_Likelihoods(string Likelihoods_Filename, const double& laverage_IS_weight,
    const int& no_runs, const double& lsum_squared_IS_weights,
    const time_t& Beg_time, time_t& Current_time, time_t& Last_time, time_t& End_time)
{
  if(DEBUG>2) {
	  cout << "likelihoods_file = " << Likelihoods_Filename << "\n";
	}
  ofstream likelihoods_file;
	likelihoods_file.open(Likelihoods_Filename.c_str(),ios::app);
	if(likelihoods_file.fail())
	{
		cout << "Warning, opening the likelihoods file failed. \n";
		getchar();
	}
	double l_bridge_variance, l_bridge_var_Coef, bridge_ESS, l_bridge_average_weight;

	if(!(likelihoods_file.good()))
	{
		cout << "Warning, There has been at least one error with the likelihoods file. \n";
		getchar();
	}
	
	double lvariance_IS_weights, coefficient_of_variation=std::numeric_limits<double>::quiet_NaN();
		
	if(rho_step<change_rho || change_rho == 0)
	{
		if(rho_step == 0)
		{
			likelihoods_file	<< "No seg sites:           \t"		<< no_segsts << "\n" 
								<< "No sequences:			\t"		<< smpl_size << "\n"
								<< "Using SMC? (0=Coal):    \t"		<< SMC << "\n" 
								<< "Choosing random hap? (0=no): "	<< choose_hap_first << "\n"
								<< "Using Fast_Coal? (0=no):\t"		<< coal_fast << "\n"
								<< "Fast_Coal THreshold?:   \t"		<< threshold << "\n"
								<< "No Runs:\t\t\t\t"				<< no_runs << "\n"
								<< "Theta (per sequence):\t\t"		<< theta << "\n"
								<< "Data Set            :\t\t"		<< which_set << "\n";							
			if(sampler==1)
				likelihoods_file		<< "Using Simple IS Conditionals \n\n";
			if(sampler==2)
				likelihoods_file		<< "Using Li and Stephens Conditionals \n\n";
			if(sampler==3)
				likelihoods_file		<< "Using Cardin and McVean Conditionals \n\n";
			if(sampler==4)
				likelihoods_file		<< "Using Griffiths and Marjoram Conditionals \n\n";

			likelihoods_file			<< "The site positions were at: \n";
			int data_cols;
			for(data_cols=0; data_cols< no_segsts; data_cols++)
				likelihoods_file << site_posns.at(data_cols) << " ";
			likelihoods_file << "\n\n";
			likelihoods_file			<< "The Data input this time was (Data set" << which_set << "): \n";

			int data_rows;
			for(data_rows = 0; data_rows < smpl_size; data_rows++)
			{
				for(data_cols = 0; data_cols < no_segsts; data_cols++)
					likelihoods_file << data.at(data_rows).at(data_cols);
				likelihoods_file << "\n";
			}

			likelihoods_file << "\nRho\t\tAverageWeight\tESS\t\tAverageRecs\tWeightedRecs\tTimeForRun\n";
			Current_time = Beg_time;
		}

		if((lsum_squared_IS_weights - log((double)no_runs)) - (2.0 * laverage_IS_weight) > 1E-250)
		{
			lvariance_IS_weights	 = Subtract_Log(lsum_squared_IS_weights - log((double)no_runs), 2.0 * laverage_IS_weight);
			coefficient_of_variation = (lvariance_IS_weights/2.0) - laverage_IS_weight;
		}
		likelihoods_file << setprecision(6)<< rho << "   \t" << laverage_IS_weight << "   \t\t"; 			
		
		if(lsum_squared_IS_weights - log((double)no_runs) - 2.0 * laverage_IS_weight > 1E-250)
			likelihoods_file << 1.0/(1.0 + exp(2.0*coefficient_of_variation)) * (double)no_runs << "  \t";
		
		Last_time	 = Current_time;
		Current_time = time(NULL);
		likelihoods_file << total_no_recs/(double)no_runs << "   \t\t";
		likelihoods_file << weighted_no_recs/(double)no_runs << "   \t\t";
		//likelihoods_file << exp(weighted_no_recs-log((double)no_runs)) << "   \t\t";
		likelihoods_file <<  Current_time - Last_time<< "\n";

		if(change_rho == 0)
		{
			End_time = time(NULL); 
			likelihoods_file << "\nTime at end = " << End_time << ", so time taken = " << End_time - Beg_time << "\n";
		}
		likelihoods_file.close();
	}			
	else
	{
		//likelihoods_file.open(Likelihoods_Filename.c_str(),ios::app);	
		likelihoods_file << "\n\nBridge Sampling likelihoods: \n";
		
		int i,j;
		for(j=0; j<change_rho; j++)
		{
			likelihoods_file << "Values sampled using rho = " << rho_vector.at(j) << ":\n";
			likelihoods_file << "Rho:\t";
			for(i=0; i<bridge_resolution; i++)
				likelihoods_file << bridge_rhos.at(i) << "\t\t";
			likelihoods_file << "\n";
			likelihoods_file << "Likelihood:\t";		
			
			for(i=0; i<bridge_resolution; i++)
			{
				l_bridge_average_weight = lsum_bridge_weights.at(j).at(i) - log((double)no_runs);
				likelihoods_file << l_bridge_average_weight << "\t";
			}
			likelihoods_file << "\n";
			likelihoods_file << "ESS:\t\t";
			for(i=0; i<bridge_resolution; i++)
			{
				l_bridge_average_weight = lsum_bridge_weights.at(j).at(i) - log((double)no_runs); 
				//cout << "The sum of the weights is:" << exp(lsum_bridge_weights.at(j).at(i)) << "\n";
				//cout << "The average of the weights is:" << exp(l_bridge_average_weight) << "\n";
				//cout << "The sum_squared of the weights is:" << exp(lsum_bridge_weights_squared.at(j).at(i)) << "\n";
				
				if(fabs(2.0 * l_bridge_average_weight - lsum_bridge_weights_squared.at(j).at(i) - log((double)no_runs)) > 1e-250)
					l_bridge_variance	= Subtract_Log(lsum_bridge_weights_squared.at(j).at(i) - log((double)no_runs), 2.0 * l_bridge_average_weight);
				else 
					l_bridge_variance = -1E308;
			//	cout << "lsum_bridge_weights.at(" << i << ") = " << exp(lsum_bridge_weights.at(i)) << " \n";
			//	cout << "lsum_bridge_weights_squared.at(" << i << ") = " << exp(lsum_bridge_weights_squared.at(i)) << " \n";
				l_bridge_var_Coef	= (l_bridge_variance/2.0) - l_bridge_average_weight; // Can only use one sampler.
				bridge_ESS			= ((double)no_runs)/(1.0 + exp(2.0*l_bridge_var_Coef));			
				if(bridge_ESS < 0.95)
				{
					cout << "ESS should never be less than 1 I think.\n";
					getchar();
				}
				likelihoods_file << bridge_ESS << "\t";
			}
			likelihoods_file << "\n\n";
		}
		
		End_time = time(NULL); 
		likelihoods_file << "\nTime at end = " << End_time << ", so time taken = " << End_time - Beg_time << "\n";
		likelihoods_file.close();
	}
}

void Ancestry::Write_Full_Data(string Full_Data_Filename, const int& run, const int& no_runs, const int& no_recs)
{
	ofstream Full_Data_file; 
	Full_Data_file.open(Full_Data_Filename.c_str(),ios::app);
	if(Full_Data_file.fail())
	{
		cout << "Warning, opening the Full_Data file failed. \n";
		getchar();
	}
	if(!(Full_Data_file.good()))
	{
		cout << "Warning, There has been at least one error with the Full Data file. \n";
		getchar();
	}
	
	
	
	if(run == 0)
	{
		Full_Data_file	<< "No seg sites:           \t"		<< no_segsts << "\n" 
						<< "No sequences:			\t"		<< smpl_size << "\n"
						<< "Using SMC? (0=Coal):    \t"		<< SMC << "\n" 
						<< "Choosing random hap? (0=no): "	<< choose_hap_first << "\n"
						<< "No Runs:\t\t\t\t"				<< no_runs << "\n"
						<< "Theta (per sequence):\t\t"		<< theta << "\n"
						<< "Rho (per sequence)  :\t\t"		<< rho << "\n"
						<< "Data Set            :\t\t"		<< which_set << "\n";							
		if(sampler==1)
			Full_Data_file		<< "Using Simple IS Conditionals \n\n";
		if(sampler==2)
			Full_Data_file		<< "Using Li and Stephens Conditionals \n\n";
		if(sampler==3)
			Full_Data_file		<< "Using Cardin and McVean Conditionals \n\n";
		if(sampler==4)
			Full_Data_file		<< "Using Griffiths and Marjoram Conditionals \n\n";

		Full_Data_file			<< "The site positions were at: \n\n";
		int data_cols;
		for(data_cols=0; data_cols< no_segsts; data_cols++)
			Full_Data_file << site_posns.at(data_cols) << " ";
		Full_Data_file << "\n\n";
		Full_Data_file			<< "The Data input this time was (Data set" << which_set << "): \n\n";

		int data_rows;
		for(data_rows = 0; data_rows < smpl_size; data_rows++)
		{
			for(data_cols = 0; data_cols < no_segsts; data_cols++)
				Full_Data_file << data.at(data_rows).at(data_cols);
			Full_Data_file << "\n";
		}
		Full_Data_file << "\n\n";
		Full_Data_file << "Run\t" << "log(IS Weight)\t" << "log(ARG Density)\t" << "No Recs\t" << "tMRCA at 1st\t" << "tMRCA at last\t" << "Tot time at 1st\t" << "Tot time at last\t" << "\n";
	}
	//marg_gens[node_no][site][node_attribute]			
	Full_Data_file << run << "\t" << limportance_weight << "\t\t" << lprior_seq_events<< "\t\t" << no_recs << "\t\t" << marg_gens.at(node_MRCA-1).at(0).c_time << "\t\t" << marg_gens.at(node_MRCA-1).at(no_segsts-1).c_time << "\t\t" << marg_gens.at(node_MRCA-1).at(0).time_below << "\t\t" << marg_gens.at(node_MRCA-1).at(no_segsts-1).time_below << "\t\n";
	Full_Data_file.close();
}

void Ancestry::Make_Rho_Vector()
{
	bool use_preset_rhos = 1;
	if(change_rho >19)
		use_preset_rhos = 0;
	rho_vector.resize(max(1,change_rho));
	
	int i;
	double change_factor;
	if(use_preset_rhos)
	{
		if(change_rho && change_rho <=4)
		{
			if(0<change_rho) rho_vector.at(0) = 1e-9;
			if(1<change_rho) rho_vector.at(1) = 1;
			if(2<change_rho) rho_vector.at(2) = 5;
			if(3<change_rho) rho_vector.at(3) = 20;
		}

		if(change_rho >4  && change_rho <=5)
		{
			if(0<change_rho) rho_vector.at(0) = 1e-9;
			if(1<change_rho) rho_vector.at(1) = 0.1;
			if(2<change_rho) rho_vector.at(2) = 1;
			if(3<change_rho) rho_vector.at(3) = 3;
			if(4<change_rho) rho_vector.at(4) = 5;
			if(5<change_rho) rho_vector.at(5) = 10;	
		}
		
		if(change_rho >5 && change_rho <=7)
		{
			if(0<change_rho) rho_vector.at(0) = 1;
			if(1<change_rho) rho_vector.at(1) = 3;
			if(2<change_rho) rho_vector.at(2) = 5;
			if(3<change_rho) rho_vector.at(3) = 7;
			if(4<change_rho) rho_vector.at(4) = 9;
			if(5<change_rho) rho_vector.at(5) = 11;
			if(6<change_rho) rho_vector.at(6) = 13;
			if(7<change_rho) rho_vector.at(7) = 15;
		}

		if(change_rho == 8)
		{
			if(0<change_rho) rho_vector.at(0) = 0.1;
			if(1<change_rho) rho_vector.at(1) = 1;
			if(2<change_rho) rho_vector.at(2) = 2;
			if(3<change_rho) rho_vector.at(3) = 5;
			if(4<change_rho) rho_vector.at(4) = 10;
			if(5<change_rho) rho_vector.at(5) = 20;
			if(6<change_rho) rho_vector.at(6) = 50;
			if(7<change_rho) rho_vector.at(7) = 100;		
		}

		if(change_rho == 9)
		{
			if(0<change_rho) rho_vector.at(0) = 1e-9;
			if(1<change_rho) rho_vector.at(1) = 0.1;
			if(2<change_rho) rho_vector.at(2) = 1;
			if(3<change_rho) rho_vector.at(3) = 10;
			if(4<change_rho) rho_vector.at(4) = 50;
			if(5<change_rho) rho_vector.at(5) = 100;
			if(6<change_rho) rho_vector.at(6) = 200;
			if(7<change_rho) rho_vector.at(7) = 500;
			if(8<change_rho) rho_vector.at(8) = 1000;
		}
		
		if(change_rho >9 && change_rho <=12)
		{
			if(0<change_rho) rho_vector.at(0) = 1e-9;
			if(1<change_rho) rho_vector.at(1) = 0.1;
			if(2<change_rho) rho_vector.at(2) = 0.3;
			if(3<change_rho) rho_vector.at(3) = 0.5;
			if(4<change_rho) rho_vector.at(4) = 1;
			if(5<change_rho) rho_vector.at(5) = 3;
			if(6<change_rho) rho_vector.at(6) = 5;
			if(7<change_rho) rho_vector.at(7) = 7;
			if(8<change_rho) rho_vector.at(8) = 10;
			if(9<change_rho) rho_vector.at(9) = 13;
			if(10<change_rho) rho_vector.at(10) = 16;
			if(11<change_rho) rho_vector.at(11) = 20;
		}

		if(change_rho ==13)
		{
			if(0<change_rho) rho_vector.at(0) = 1e-9;
			if(1<change_rho) rho_vector.at(1) = 0.1;
			if(2<change_rho) rho_vector.at(2) = 0.5;
			if(3<change_rho) rho_vector.at(3) = 1;
			if(4<change_rho) rho_vector.at(4) = 2;
			if(5<change_rho) rho_vector.at(5) = 3;
			if(6<change_rho) rho_vector.at(6) = 4;
			if(7<change_rho) rho_vector.at(7) = 5;
			if(8<change_rho) rho_vector.at(8) = 6;
			if(9<change_rho) rho_vector.at(9) = 7;
			if(10<change_rho) rho_vector.at(10) = 8;
			if(11<change_rho) rho_vector.at(11) = 10;
			if(12<change_rho) rho_vector.at(12) = 15;
		}


		if(change_rho >13 && change_rho <=18)
		{
			if(0<change_rho) rho_vector.at(0)   = 1e-9;
			if(1<change_rho) rho_vector.at(1)   = 0.1;
			if(2<change_rho) rho_vector.at(2)   = 0.3;
			if(3<change_rho) rho_vector.at(3)   = 0.5;
			if(4<change_rho) rho_vector.at(4)   = 1;
			if(5<change_rho) rho_vector.at(5)   = 2;
			if(6<change_rho) rho_vector.at(6)   = 3;
			if(7<change_rho) rho_vector.at(7)   = 4;
			if(8<change_rho) rho_vector.at(8)   = 5;
			if(9<change_rho) rho_vector.at(9)   = 6;
			if(10<change_rho) rho_vector.at(10) = 8;
			if(11<change_rho) rho_vector.at(11) = 10;
			if(12<change_rho) rho_vector.at(12) = 12.5;
			if(13<change_rho) rho_vector.at(13) = 15;
			if(14<change_rho) rho_vector.at(14) = 17.5;
			if(15<change_rho) rho_vector.at(15) = 20;
			if(16<change_rho) rho_vector.at(16) = 25;
			if(17<change_rho) rho_vector.at(17) = 30;
			if(18<change_rho) rho_vector.at(18) = 50;
		}
	
		if(change_rho == 19)
		{
			if(0<change_rho) rho_vector.at(0)   = 1e-9;
			if(1<change_rho) rho_vector.at(1)   = 0.1;
			if(2<change_rho) rho_vector.at(2)   = 0.3;
			if(3<change_rho) rho_vector.at(3)   = 0.5;
			if(4<change_rho) rho_vector.at(4)   = 1;
			if(5<change_rho) rho_vector.at(5)   = 2;
			if(6<change_rho) rho_vector.at(6)   = 3;
			if(7<change_rho) rho_vector.at(7)   = 4;
			if(8<change_rho) rho_vector.at(8)   = 5;
			if(9<change_rho) rho_vector.at(9)   = 7;
			if(10<change_rho) rho_vector.at(10) = 10;
			if(11<change_rho) rho_vector.at(11) = 15;
			if(12<change_rho) rho_vector.at(12) = 20;
			if(13<change_rho) rho_vector.at(13) = 50;
			if(14<change_rho) rho_vector.at(14) = 100;
			if(15<change_rho) rho_vector.at(15) = 200;
			if(16<change_rho) rho_vector.at(16) = 500;
			if(17<change_rho) rho_vector.at(17) = 1000;
			if(18<change_rho) rho_vector.at(18) = 5000;
		}
	}	
	else
	{
		rho_vector.at(0) = rho;			
		if(rho !=0)
		{
			for(i=1; i<change_rho; i++) //This loop ensures that we never divide by change_rho when change_rho == 0.
			{
				if(change_rho<10)
					change_factor	 = pow(10.0/rho, 1.0/(double)(change_rho-1));
				else
					change_factor	 = pow(100.0/rho, 1.0/(double)(change_rho-1));
				rho_vector.at(i) = rho * pow(change_factor, (double)i);
			}
		}
		else
		{
			cout << "Error: General changing rho function cannot yet cope with an initial rho = 0\n";
			getchar();
		}
	}
}	

void Ancestry::Change_Rho(double& rho)
{
	rho = rho_vector.at(rho_step);	
}

void Ancestry::Collate_ARG_Info(const int& run)
{
	int event_index; 
	int i;
	vector<double> 	temp_times;

	if((write_full_data || no_runs < too_many_runs) && importance_sample)
	{
		for(event_index = 0; event_index<(int)event_vector.size(); event_index++)
			if(event_vector.at(event_index).event_type == 3)
				rec_positions.at(run).push_back(event_vector.at(event_index).rec_site);
		lIS_Weights.at(run)	   	= limportance_weight;
		  lPr_Probs.at(run)	   	= lprior_seq_events;
		lsum_IS_weights	    		= Add_Log(lsum_IS_weights, limportance_weight, "Collate_ARG_Info1");
		lsum_squared_IS_weights	= Add_Log(lsum_squared_IS_weights, 2.0 * lIS_Weights.at(run), "Collate_ARG_Info2");
	}
	if(collect_times && record_every_time)
		for(i=0; i<rec_sites; i++)
			MRCA_times.at(run).at(i) = marg_gens.at(2*smpl_size-2).at(i).c_time;	
	//On every run the bridge sampler updates the total bridge weight for each value of rho.
	if(importance_sample)
	{
		for(i=0; i<bridge_resolution; i++)
		{
			lsum_bridge_weights.at(rho_step).at(i)			= Add_Log(lsum_bridge_weights.at(rho_step).at(i),
			    lbridge_weights.at(rho_step).at(i), "Collate_ARG_Info3");
			lsum_bridge_weights_squared.at(rho_step).at(i)	= Add_Log(lsum_bridge_weights_squared.at(rho_step).at(i),
			    2.0 * lbridge_weights.at(rho_step).at(i), "Collate_ARG_Info4");
		}
	}
}

void Ancestry::Write_Densities(const int& run, const time_t& Beg_time, time_t& End_time)
{
	if(run == 0)
	{
		if(rho_step == 0)
		{
			densities_file		<< "No seg sites				\t: "	<< no_segsts		<< "\n" 
								<< "No sequences				\t: "	<< smpl_size		<< "\n"
								<< "Using SMC? (0=Coal)			\t: "	<< SMC				<< "\n" 
								<< "Choosing random hap? (0=no)   	\t: "	<< choose_hap_first << "\n"
								<< "No Runs:					\t: "	<< no_runs			<< "\n"
								<< "Theta (per sequence)	      	\t: "	<< theta			<< "\n"
								<< "Rho			   	            \t: "	<< rho				<< "\n"
								<< "Data Set					\t: "	<< which_set		<< "\n";							
			densities_file << "\n\n";			
			densities_file << "Run: \tNo_Recs \t Density\n";
		}
		else
		{
				densities_file << "Rho now = " << rho << "\n";
		}
	}

	densities_file << run << "\t" << rec_records.size() << "\t\t" << lARG_density << "\n";

	if(run == no_runs-1)
	{
		average_recs = total_no_recs/(double)no_runs;
		densities_file << "\n";
		densities_file << "Average no_recs = " << average_recs << "\n";
		densities_file << "\n\n";		
		if((rho_step == change_rho-1) || change_rho == 0)
		{
			End_time = time(NULL); 
			densities_file << "\nTime at end = " << End_time << ", so time taken = " << End_time - Beg_time << "\n";
			densities_file << "\n";
			densities_file << "\n\n";
			densities_file.close();
		}		
	}
}

void Ancestry::Write_Times(string times_Filename, const int& run, const time_t& Beg_time, time_t& End_time)
{
	int i;
	
	ofstream times_file; 
	times_file.open(times_Filename.c_str(),ios::app);
	if(times_file.fail())
	{
		cout << "Warning, opening the time file failed. \n";
		getchar();
	}
	if(!(times_file.good()))
	{
		cout << "Warning, There has been at least one error with the times file. \n";
		getchar();
	}
	
	if(run==0 || record_every_time)
	{
		if(rho_step == 0)
		{
			times_file			<< "No seg sites				\t: "	<< no_segsts		<< "\n" 
								<< "No sequences				\t: "	<< smpl_size		<< "\n"
								<< "Using SMC? (0=Coal)			\t: "	<< SMC				<< "\n" 
								<< "Choosing random hap? (0=no)	\t: "	<< choose_hap_first << "\n"
								<< "No Runs:					\t: "	<< no_runs			<< "\n"
								<< "Theta (per sequence)		\t: "	<< theta			<< "\n"
								<< "Rho							\t: "	<< rho				<< "\n"
								<< "Data Set					\t: "	<< which_set		<< "\n";							
			times_file << "\n\n";			
			times_file << "MRCA times at each site:\n";
		}
	}

	int this_run;
	for(this_run=0; this_run<no_runs; this_run++)
	{
		times_file << "Run " << this_run << " ";
		for(i=0; i<rec_sites; i++)
			times_file << MRCA_times.at(this_run).at(i) << " ";
	}
	times_file << "\n\n";
	End_time = time(NULL); 
	times_file << "\nTime at end = " << End_time << ", so time taken = " << End_time - Beg_time << "\n";
	times_file << "\n";
	times_file.close();
}

void Ancestry::Summarise_Simulations()
{
	int run;
	if(importance_sample)
	{
		if(write_full_data || no_runs < too_many_runs)
		{
			laverage_IS_weight = lsum_IS_weights - log((double)no_runs);		
			cout << "Sum of weights = " << exp(lsum_IS_weights) << "\n";
		}
		if(verbose>=1) {
		  if(no_runs<11) {cout << "The importance weights were: \n"; for(run=0; run<no_runs ; run++){cout << exp(lIS_Weights.at(run)) << ", logged = " << lIS_Weights.at(run) << "\n";} }
		  if(no_runs<11) {cout << "The prior probs        were: \n"; for(run=0; run<no_runs ; run++){cout << exp(lPr_Probs.at(run))   << ", logged = " << lPr_Probs.at(run)   << "\n";} }
		  if(no_runs>1)
			  cout << "The average of the importance weights was: " << exp(laverage_IS_weight) << ", (logged = " << laverage_IS_weight << ")\n"; 
    }
		double lvariance_IS_weights, coefficient_of_variation;
		
		if(no_runs>5 && (write_full_data || no_runs < too_many_runs) && verbose>=1)
		{
			if(((lsum_squared_IS_weights - log((double)no_runs)) - (2.0 * laverage_IS_weight) > 1E-250) && importance_sample)
			{
				lvariance_IS_weights	 = Subtract_Log(lsum_squared_IS_weights - log((double)no_runs), 2.0 * laverage_IS_weight);
				coefficient_of_variation = (lvariance_IS_weights/2.0) - laverage_IS_weight;			
				cout << "The Variation Coefficient of the weights was: " << exp(coefficient_of_variation) << ", (logged = " << coefficient_of_variation << ")\n"; 	 
				cout << "The (estimate of) effective sample size was: " << 1.0/(1.0 + exp(2.0*coefficient_of_variation))  << " * " << no_runs << " = " << 1.0/(1.0 + exp(2.0*coefficient_of_variation)) * (double)no_runs << "\n";
			}
		}
		weighted_no_recs = 0;
		if((write_full_data || no_runs < too_many_runs) && verbose>=1)
		{
			for(run=0; run<no_runs; run++)
				weighted_no_recs += exp(lIS_Weights.at(run)) * (double)(rec_positions.at(run).size());
			weighted_no_recs /=  exp(laverage_IS_weight); // This is to normalise by the average weighting.
			cout << "The weighted average number of recombinations over these runs was: " << weighted_no_recs/(double)no_runs << "\n";
			cout << "The average number of recombinations over these runs was: " << total_no_recs/(double)no_runs << "\n";
		}
	}
	else
		//cout << "The final ARG density was: " << lARG_density;
	if(collect_times)
		Write_Q_Info();	
	if(collect_densities && verbose>=1)
		cout << "The Average Number of Recombination Events was: " << (double)total_no_recs/(double)no_runs << "\n";
}

void Ancestry::Write_Q_Info()
{
	if(rho_step == 0)
	{
		if(Q_file.fail())
		{
			cout << "Warning, opening the time file failed. \n";
			getchar();
		}
		if(!(Q_file.good()))
		{
			cout << "Warning, There has been at least one error with the times file. \n";
			getchar();
		}
		//Q_file << " ";
		Q_file		<< "No seg sites		\t\t: "			<< no_segsts		<< "\n"
							<< "No sequences		\t\t: "			<< smpl_size		<< "\n"
							<< "Using SMC? (0=Coal)			\t: "	<< SMC				<< "\n" 
							<< "Choosing random hap? (0=no)	\t: "	<< choose_hap_first << "\n"
							<< "No Runs:			\t\t: "			<< no_runs			<< "\n"
							<< "Theta (per sequence)		\t: "	<< theta			<< "\n"
							<< "Rho					\t\t: "			<< rho				<< "\n"
							<< "Data Set			\t\t: "			<< which_set		<< "\n";							
		Q_file << "\n\n";			
		Q_file << "No seqs\tRho  \tP(Rho)   \t%NM      \tQ*           \t%Recombined      \tNo Runs  \n";
	}
	double proportion_NM_runs		= (double)non_Markovian_runs/(double)no_runs;
	double proportion_ends_the_same	= (double)ends_the_same/(double)no_runs;
	double proportion_ends_same_rec	= (double)ends_same_rec/(double)no_runs;
	double proportion_had_rec		= (double)had_rec/(double)no_runs;

	if(proportion_ends_the_same)
		Q_file << smpl_size << "     \t" << rho << "   \t" << 100*proportion_ends_the_same << "     \t"<< 100*proportion_NM_runs << "   \t" << (double)non_Markovian_runs/(double)ends_the_same << "   \t"  << 100*(proportion_had_rec) << "   \t"  << no_runs<< endl;
	else
		Q_file << smpl_size << "     \t" << rho << "   \t" << 100*proportion_ends_the_same << "     \t"<< 100*proportion_NM_runs << "   \t" << "N/A" << "   \t"  << 100*(proportion_had_rec)  << "   \t"  << no_runs << endl;


	if(rho_step == change_rho-1)
	{
		Q_file << "\n\n";
		Q_file.close();
	}		
	cout << "The percentage of runs with no recombination:            " << 100*(1-proportion_had_rec) << "%\n";
	cout << "The percentage of runs with recombination yet ends same: " << 100*(proportion_ends_same_rec) << "%\n";
	cout << "The Percentage of Runs with Ends the Same:               " << 100*proportion_ends_the_same << "%\n";
	cout << "The Total number of Gil style Non Markovian Runs was:    " << non_Markovian_runs << "\n"
		 << "As a percentage of the total runs this is:               " << 100*proportion_NM_runs << "%\n";
	cout << "So Q* is:                                                " << 100*((double)non_Markovian_runs/(double)ends_the_same) << "%\n";
}

//string outputDir - not needed as it's part of Ancestry! (which is derived from argumentsOfProgram)
string Ancestry::Create_Likelihoods_Filename()
{
	stringstream			create_Likelihoods_Filename; string Likelihoods_Filename;
	string model;
	if(SMC)
	  model = "SMC";
	else
	  model = "Coal";
	create_Likelihoods_Filename <<  outputDir   << "Likelihoods_" << model
                              << "_numSites=" << no_segsts
                              << "_numInds="  << smpl_size      << "_numRuns="  << no_runs
                              << "_Data_Set=" << which_set      << "_Analysis=" << file_tag<< ".txt";

	Likelihoods_Filename	= create_Likelihoods_Filename.str(); 
	return Likelihoods_Filename;
}

string Ancestry::Create_Times_Filename(const bool& Q_info)
{
	stringstream			create_Times_Filename; string Times_Filename;
	create_Times_Filename <<  outputDir << file_tag;
	if(Q_info)
		create_Times_Filename <<  "Q_Info_";
	else
		create_Times_Filename <<  "Times_";
  string model;
  if(SMC)
    model = "SMC"; else model = "Coal";
  create_Times_Filename << model << "_n=" << smpl_size << "_No_Runs=" << no_runs << "_Data_Set=" << which_set << "_rho=" << rho << ".txt";
	Times_Filename	= create_Times_Filename.str(); 

	return Times_Filename;
}

string Ancestry::Create_Densities_Filename()
{
	stringstream			create_densities_Filename; string densities_Filename;
		create_densities_Filename <<  outputDir <<  "Densities - ";
	  string model;
    if(SMC)
      model = "SMC"; else model = "Coal";
		create_densities_Filename <<  model << "_n=" << smpl_size << "_No_Runs=" << no_runs << "_Data_Set=" << which_set << "_rho=" << rho << ".txt";
	densities_Filename	= create_densities_Filename.str(); 

	return densities_Filename;
}

string Ancestry::Create_Full_Data_Filename()
{
	stringstream		create_Full_Data_Filename; string Full_Data_Filename;
  string model;
  if(SMC)
    model = "SMC"; else model = "Coal";
		create_Full_Data_Filename <<  outputDir <<  "Full_Data_" << model << "_rho=" << rho << "_n=" << smpl_size << "_No_Runs=" << no_runs << "_Data_Set=" << which_set << "_Analysis=" << file_tag<< ".txt";
	Full_Data_Filename	= create_Full_Data_Filename.str(); 	

	return Full_Data_Filename;
}

void Ancestry::Display_Run(const int& run, const int& rho_steps, int approxNumRuns)
{
	if(run == 0)
		cout << "  On Run 0";
	else 
	{
	  if(approxNumRuns==-1)
	    approxNumRuns = ceil(pow(10, round(log10(no_runs))));
    if(((1000 * run)%approxNumRuns) == 0)
      cout << "\r  On Run " << run;
    //cout << ", " << run;
      cout.flush();
	}
	if(run == no_runs-1)
	{
		cout << "\rFinal run of all " << no_runs << " runs!";
		cout << "\n";
	}	
}

int Ancestry::Prior_Choose_Event(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight)
{
	//This is not an importance sampling function, and this function cannot deal with data with mutations.
	// Never choose to mutate!
	if(theta > 0)
	{
		cout << "Error! Mutation rate is not zero and yet no method to deal with mutations has been implemented\n";
		getchar();
	}
	int reject_coal;
	double  choose_evnt;
	int event = std::numeric_limits<int>::quiet_NaN();
	
	reject_coal = 1;
	choose_evnt	= ran->U(); 
	if (choose_evnt< coal_rate/evnt_rate) 
	{
		if (SMC)
		{
			while(reject_coal)
			{
				Prior_Choose_Pair(daughter1, daughter2); // Note that daughter1 is chosen to be < daughter2 (swapped if not)
				if(Check_Coal(daughter1, daughter2))
				{
					reject_coal = 0;
					event = 1; //Coalescence
				}
				else
				{
					Prior_Choose_Pair(daughter1, daughter2); // Note that daughter1 is chosen to be < daughter2 (swapped if not)		
				}
			}
		}
		else
			Prior_Choose_Pair(daughter1, daughter2); // Note that daughter1 is chosen to be < daughter2 (swapped if not)				
		event = 1; //Coalescence
		rec_lineage = -1;
		rec_site	= -1;

	}
	else 
	{
		reject_coal = 0;
		if (choose_evnt< (coal_rate + rec_rate)/evnt_rate) 
		{
			event = 3; //Recombination
			rec_lineage = Prior_Choose_Rec_Lineage();
			Prior_Choose_position(rec_lineage, rec_site);
			daughter1 = -1;
			daughter2 = -1;

		}
		else {
			cout << "Error! Mutation rate is not zero and yet no method to deal with mutations has been implemented\n";
			getchar();
		}
	}

	levent_weight = 0; //Choosing from the prior!
	mut_lineage	  = -1;
	new_type	  = -1;
	mut_site	  = -1;
	return event;
}

// The vector Live_lineages should agree with the nodes that think they are active.
void Ancestry::Prior_Choose_Pair(int& daughter1, int& daughter2)
{
	int i,index;
	int seq1, seq2,seq_temp;

	// Choosing a pair of lineages to coalesce from the live lineages.
	seq1 = ran->discrete(0,actv_lngs-1);
	seq2 = ran->discrete(0,actv_lngs-2);
	if (seq2 >= seq1) seq2 ++;// This line ensures that we don't pick the same lineage twice.
	if(seq2<seq1){seq_temp = seq1; seq1=seq2; seq2 = seq_temp;} // This ensures that seq2>seq1.
	 

	// There may be a more efficient way of doing this loop. Such as updating a list of pointers/references? to the active lineages 
	daughter2 = -1; i=0; index =0;
	while(daughter2 == -1)
	{
		if (index <= seq1)// Then assign daughter1
		{
			if (index == seq1 && Lineage_Alive(i))	daughter1 = i;	// This returns the hap_list index of the daughter haplotype.
			if (Lineage_Alive(i))					index ++;	// index counts up to the number of live lineages									
		}
		else			 // Then assign daughter 2
		{
			if (index == seq2 && Lineage_Alive(i))	daughter2 = i;	// This returns the hap_list index of the daughter haplotype.
			if (Lineage_Alive(i))					index ++;						
		}
		i++;		//i counts through all elements of live_lineages to find the correct index in haplist.
	}
	if(daughter1 >= (int)hap_list.size() || daughter2 >= (int)hap_list.size()) {cout << "Warning! Have attempted to choose a non existant daughter lineage" << endl;getchar();exit(1);}
}

// This function takes in the sequence undergiong recombination and outputs (via a reference) the site at which to recombine
void Ancestry::Prior_Choose_position(int rec_lineage, int& rec_site)
{
	int i;
	double rec_dist_done;
	int left_edge;

	left_edge = hap_list.at(rec_lineage).left_rec_edge;
	rec_dist_done = 0;
	for(i=left_edge ; i<rec_sites-1; i++)	// Note: rec_map[0] = 0 and rec_map[1] = distance between the first 2 sites.
	{
		rec_dist_done += (rec_map.at(i+1) - rec_map.at(i)); //Rec_Map stores the genetic distance from the 0th site of each site. So it's cumulative.
		if (rec_dist_done > rec_pos_dist)
		{
			rec_site = i;					// This means that the recombination event happens between the ith and the (i+1)th sites.		
			break;
		}
	}
	if(rec_site != i){cout << "Warning! The locus for recombination was not chosen properly" << "\n";getchar();exit(1);}
}

// This function takes in nothing and outputs a random sequence to recombine on based on the ancestral material and rec_rate
int Ancestry::Prior_Choose_Rec_Lineage()
{
	int i;
	int stop;
	double rec_event;
	double dist_covered; 
	
	int rec_seq;
	
	rec_event = ran->uniform(0,rec_rate)*2.0; // The rec_rate is Sum(rec_dist/2.0), this is really a position along sum(rec_dist);
	stop = 0; i=0; dist_covered = 0;
	while(!stop)
	{
		if(i >= (int)hap_list.size()){
			cout	<< "Warning! Unable to choose lineage on which to recombine properly. " << "\n"
					<< "Problem occurred while recombining in event: " << evnt_nmbr << endl;
			getchar();
			exit(1);
		}
		if(Lineage_Alive(i))
		{
			if(DEBUG>3)
			{
				if(evnt_nmbr == 100){
				cout << "Event 100: Recombining and about to choose a lineage: " << "\n";
				for (int k=0; k<rec_sites; k++) {cout << hap_list.at(i).alleles.at(k) << " ";}
				cout << "\n";
				cout << "Left edge is: " << hap_list.at(i).left_rec_edge << " \t and right edge is: " << hap_list.at(i).right_rec_edge << "\n";
				cout << "Rec_Dist on this haplotype is: " << hap_list.at(i).rec_dist << "\n" << "\n";}
			}
			
			dist_covered += hap_list.at(i).rec_dist;
			if(rec_event < dist_covered)	
			{
				rec_seq = i; 
				stop = 1;
			}
		}
		i++;		
	}
	rec_pos_dist = rec_event - (dist_covered - hap_list.at(i-1).rec_dist); // This is the distance along the chosen sequence that the recombination occurred at.
	if(rec_pos_dist>(rec_map.at(rec_sites-1)-rec_map.at(0))){cout << "Warning! Returning a genetic distance along a sequence greater than total rho/seq" << "\n";getchar();exit(1);}
	return rec_seq;
}

// This function takes in nothing and outputs the number of segregating sizes in the data
int Ancestry::getDataSize()
{
  return data.size();
}

// This function takes in data and rate for a simple poisson and outputs the probability of the data given the rate
double Ancestry::lNaive_Calc_Poisson(int no_events, double rate)
{
	double l_nbang = 0.0;
	double lpois_prob;
	int i;
	for(i=0; i<no_events; i++)
	{
   	l_nbang += log((double)no_events - i);
	}            
   lpois_prob = (double)no_events * log(rate) - rate - l_nbang;

	//return exp(lpois_prob);
	return lpois_prob;
}  

// This function takes in two logged likelihoods and outputs the log of the sum of the likelihoods.
double Ancestry::Add_Log(double Summand1, double Summand2, string extraInfo)
{
	//if (Summand1 == numeric_limits<double>::infinity() || Summand2 == numeric_limits<double>::infinity() || -Summand1 == numeric_limits<double>::infinity() || -Summand2 == numeric_limits<double>::infinity())
	if (Summand1 == numeric_limits<double>::infinity())
	{
	  cout << "Summand 1 seems to be infinity in Add_Log." << "\n";
	  cout << "Extra Info (probably the calling function: " << extraInfo << ")\n";
	  getchar();
	}
	if (Summand2 == numeric_limits<double>::infinity())
	{
	  cout << "Summand 2 seems to be infinity in Add_Log." << "\n";
    cout << "Extra Info (probably the calling function: " << extraInfo << "\n";
	  getchar();
	}
  // Note that when either (or both!) of these is -Inf (underflow) it's safe to return the other with no loss of accuracy
	if(-Summand1 == numeric_limits<double>::infinity())
		return Summand2;
	if(-Summand2 == numeric_limits<double>::infinity())
		return Summand1;

	if (Summand1 < Summand2)
		return Summand2 + log(1 + exp(Summand1 - Summand2));
	else
		return Summand1 + log(1 + exp(Summand2 - Summand1));
}

// This function takes in two logged likelihoods and outputs the log of the difference of the likelihoods.
double Ancestry::Subtract_Log(double Summand, double subtrahend)
{
	if (Summand < subtrahend)
	{
		cout << "This is impossible, the subtrahend should be the smaller of the two numbers. The program should not work now." << endl;
		getchar();
		return 99; // An impossible value for a logged probability. Hopefully 99 will stick out.
	}
	else
		if (Summand == subtrahend)
		{ 
			cout << "The program cannot deal with such close values once they have been logged. \nOr they are equal." << endl; 
			getchar();
			return -1E308;
		}
	return Summand + log(1 - exp(subtrahend - Summand));
}

