#ifndef ANCESTRY_H
#define ANCESTRY_H

//Memory leak detection:
//#define CRTDBG_MAP_ALLOC
#include <stdlib.h>
//#include <crtdbg.h>

#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <string>
#include <time.h> // This is so I can see how long runs take
#include <limits> // This is so that I can set NaN values as defaults

#include <sstream>
#include <fstream>

// I think the below introduces a new version of 'pow' so sometimes I will need to use std::pow instead of pow
//#include "nr.h" // numerical recipies stuff
#include "random.h"
#include "parameters.h"
#include "haplotype.h"
#include "recomb_event.h"
#include "node.h"
#include "Imp_Sampler.h"
#include "Simple_IS.h"
#include "Li_and_Stephens.h"
#include "LookD_Copy.h"
#include "Cardin_McV.h"
//#include "Niall_Utils.h"
#include "read_data.h" // moved from ancestry.cpp to enable hapmapParams in this file
#include "event.h"
#include "hapmapParams.h"
#include "argumentsOfProgram.h"
#include "control.h"


#ifndef min
  #define min(a,b) ((a) <= (b) ? (a) : (b))
#endif
#ifndef max
  #define max(a,b) ((a) >= (b) ? (a) : (b))
#endif
			
using namespace std;


//class Ancestry
class Ancestry : public argumentsOfProgram
{
public:

  //ChapmapParams  hapmapParams; //now part of AOP class
  Ancestry(Random* Inst_of_random, const argumentsOfProgram &arguments) :
    argumentsOfProgram(arguments), con(arguments)
  {
    //cout << "Ancestry.hapmapParams.haps_file.fName = " << hapmapParams.haps_file.fName << "\n";
    //cout << "Ancestry.bridge_resolution            = " << bridge_resolution            << "\n";
    //cout << "arguments.bridge_resolution           = " << arguments.bridge_resolution  << "\n";
    //cout << "arguments.display_haps                = " << arguments.display_haps       << "\n";
    //cout << "Ancestry.display_haps                 = " << display_haps                 << "\n";

    //static_cast<argumentsOfProgram&>(*this) = par; //line (and associated lines) replaced by initialiser above
    //Input_data(data_Fname, which_set);
    set_ran(Inst_of_random);

    getHapmapDat(
        data_Fname  ,
        hapmapParams,
        hapmapDat   ,
        genmap      ,
        physmap     ,
        n_gen       ,
        n_hap
      );
    setMyDatFromHapmap(
        hapmapDat    ,
        data         ,
        site_posns   ,
        con          ,
        DEBUG
    );
    using_data    = true             ;
    smpl_size     = con.smpl_size    ;
    no_segsts     = con.no_segsts    ;
    all_sites     = con.all_sites    ;
    using_rec_map = con.using_rec_map;


    if(data.size()==0) {
      cout << "  Warning! data is of size 0 (in constructor1). Problem reading in data?\n";
      getchar();
    }
    if(set_seed)
    {
      ran->setseed(-set_seed);
      cout << "Warning, the seed has been SET, multiple executions will give the same results!!\n";
    }
    ran->print_seed();
    if(data.size()==0)
      cout << "  Warning! data is of size 0 (in constructor2). Problem reading in data?\n";
  };

	~Ancestry()
	{
		//cout << "Called Ancestry Destructor" << endl;
	};
		
	void set_ran(Random *ran_in){
	  ran=ran_in;
	};
	Con			   con;

	// Variables that have been introduced to use the read_hapmap code.
	map<double, snp>  hapmapDat;
  vector<double>    genmap;
  vector<double>    physmap;
  int               n_gen;
  int               n_hap;

  bool    display_tree;
  bool    know_derived;
  bool    print_data;   //If this flag is !=0 then the data is output to the console - otherwise it is not
  int     too_many_runs;
  bool    using_data;

  // I have now checked that every required commented variable from this file is implemented.
  // Checked that every commented out variable from this block is implemented elsewhere where it should be
  /*double  DEBUG;                   // 1
    bool    useHapmap;               // 2
    string  dataFormat;              // 3
    bool		display_haps;            // 4
    bool		display_events;          // 5
    bool		display_weights;         // 6
    int			write_trees;             // 7
    bool		write_likelihoods;       // 8
    bool		collect_times;           // 9
    bool		collect_densities;       // 10
    bool		write_full_data;
    bool		choose_hap_first;
    bool		coal_fast;
    double		threshold;
    bool		choose_nonrec;           // 15
    bool		first_nonrec;
    bool		use_logs;
    bool		cheap_coal_upwt;
    bool		sim_data;
    bool		using_rec_map;          // 20
    bool		importance_sample;
    //bool		using_types; //unused variable
    int			set_seed;
    bool		all_nodes;
    bool		SMC;                    // 25
    int			sampler; // 0 = No special scheme (What to do here?), 1 = Simple Events Scheme, 2 = Li and Stephens basic scheme
    int			which_set;
    string  file_tag;
    //char*		tree_file_prefix;
    //char*		SNP_names_file;      //30
    //string		tree_file_prefix;
    //string		snp_names_file;*/  //32

	// 'Input_params'function currently does nothing, it is obsolete:
  // void Input_params(string con_Fname);
	void Input_data(string con_Dname, const int& which_set);
	void getHapmapDat(
	    const string&     con_Dname,
	    const ChapmapParams&  hapmapParams,
	    map<double, snp>& hapmapDat,
	    vector<double>&   genmap,
	    vector<double>&   physmap,
	    int& n_gen,
	    int& n_hap
	  );
	void Initialise(const int& i);
    void initialiseHapList(const int& run);
    void initialiseMargGens();
    void initialiseRhoAndTheta(const int& run);
	void Controller();
	void Simulator();
	void Display_Tree(bool all_nodes);
	void Display_Haplotypes(bool useIndexLive=true);
	void Display_Hap_info(int hap1, int hap2, int hap3);
	int stop_SMC();
	inline bool Stop_Simulation();
	inline bool Lineage_Alive(const int& lineage_index);
	inline bool Junk_Lineage(const int& lineage_index);
	void Coalesce(double e_time, int daughter1, int daughter2);
		int Check_Coal(int daughter1, int daughter2);
		void Calc_index_live();
		void Coal_Error(int new_lineage, int daughter2);
		double Update_Rates(int event, int new_lineage1, int new_lineage2, double old_rec_dist1, double old_rec_dist2, double new_rec_dist);
			int Count_Coal_Pairs();
	void Recombine(double wtng_time, int& rec_lineage, int& rec_site);
		int    Create_new_Lineage(int rec_lineage);
		void   Update_hap_list(int rec_lineage, int new_lineage, int& rec_site);
	void Mutate(const int& mut_lineage, const int& mut_site, const int& new_type);
		double Calc_Mut_Rate();
  int getDataSize();

	
		//Older functions, not used when importance sampling:

		int		Prior_Choose_Event(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight);
		void	Prior_Choose_Pair(int& daughter1, int& daughter2);
		int		Prior_Choose_Rec_Lineage();		
		void	Prior_Choose_position(int rec_lineage, int& rec_site);
		

protected:

	Random *ran;
	
	//CSimple_IS	Simple_IS;	//May need to think about naming of importance sampler. May wish to call all importance samplers something but initialise a different type of one dynamically, somehow.	
	//Imp_Sampler&	This_Sampler;
	//double		theta;
	double		theta_per_site;
	//double		rho;
	double		rho_per_site;
	int				smpl_size;
	int				node_MRCA;
	int				all_sites;
	int				no_segsts;
	int				rec_sites;
	int				actv_lngs;
	int				dead_lngs;
	int				MRCA_lngs;
	double		crnt_height;
	double		rec_rate;
	double		coal_rate;
	double		mut_rate;
	double		evnt_rate;
	int				evnt_nmbr;
	// The below are now in AOP
	/*
    int       no_runs;
    int       change_rho;
    bool      change_theta;
    bool      output_MLE;
    bool      using_wattersons;
    bool      greedy_mle;
    bool      record_every_time;
	*/
	int				non_Markovian_runs;
	int				ends_the_same;
	int				ends_same_rec;
	int				had_rec;
	double		average_recs;
		
	vector<int>				      live_lineages;
	vector<int>				      index_live;
	vector < vector<int> >	data;
	vector<haplotype>		    hap_list;
  vector<double>          site_posns;
  vector<double>          physPosns;
	vector<recomb_event>	  rec_records;
	vector<double>			    mut_times; //This records when each segregating site returned to the ancestral state.
	vector<double>			    rec_map;		//Rec_Map stores the genetic distance from the 0th site of each site. So it's cumulative. rec_map[0] = 0 and rec_map[1] = distance between the first 2 sites.
	vector< vector<node> >	marg_gens;

    double                  calc_likelihood();
    void                    calc_time_in_tree();
    //void                    calc_choice_quotients(double& event_weight, vector<double>& choice_quotients, int event, const double& wtng_time, const int& rec_site);
	void                    calc_choice_quotients(double& event_weight, int event, const double& wtng_time, const int& rec_site);
	void					calc_densities(int event, const double& wtng_time, const int& rec_site);
 	vector<double>			choice_quotients; // This is the IS weight for each separate event
	vector<double>			prior_weights;
	vector<double>			event_densities;
	vector<double>			bridge_rhos;
	vector<vector<double> > lbridge_weights;
	vector<vector<double> > lsum_bridge_weights;
	vector<vector<double> > lsum_bridge_weights_squared;
	//int bridge_resolution;
	
 	double                  lLikelihood_Data;
  vector<double>          time_in_trees;

	double lprior_seq_events;
	double lARG_density;
  double limportance_weight;
  double laverage_likelihood;
  double lsum_IS_weights;
	double laverage_IS_weight;
	double lsum_squared_IS_weights;
	vector<double> lIS_Weights;
	vector<double> lPr_Probs;
	vector<vector<double> > MRCA_times;
	int total_no_recs; 
	double weighted_no_recs;
	
	string con_Fname;
	//string data_Fname;
	ofstream Q_file; 
	ofstream densities_file; 

private:
	double		rec_pos_dist;	// This refers to the distance along the recombining sequence at which it recombines.
	double		rec_mat_lost;	// This checks for trapped non-ancestral material which has been broken up again by recombination, and stores the total lost rec material after a recombination.
	int			no_nodes;
	int			rho_step; 
	vector<int>	which_node;
	vector<int>	count_type_1;
	vector<bool> still_segregating;
	vector<CEvent>	event_vector;
	vector<vector<int> > rec_positions;
	vector<double> rho_vector;

	CEvent temp_event;
	
	double lNaive_Calc_Poisson(int no_events, double rate);
	double Add_Log(double Summand1, double Summand2, string extraInfo="No extra info supplied");
	double Subtract_Log(double Summand, double subtrahend);
	void Output_Sampler();
	void Update_Event_Vector(const int& event, int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight);
	inline int Live_Site(int allele);
	inline int Ancestral(int allele);
	inline int Find_Left_Edge(const haplotype& daughter_hap, const haplotype& new_hap);
	inline int Find_Right_Edge(const haplotype& daughter_hap, const haplotype& new_hap);
	void Collate_ARG_Info(const int& run);	
	void Summarise_Simulations();
	void Bridge_Sample(const int& event, const int& rec_site, const double& levent_weight);
		void Define_Bridge_Rhos();
	void Write_Likelihoods(string Likelihoods_Filename, const double& laverage_IS_weight, const int& no_runs, const double& lsum_squared_IS_weights, const time_t& Beg_time, time_t& Current_time, time_t& Last_time, time_t& End_time);
	void Write_Full_Data(string Full_Data_Filename, const int& run, const int& no_runs, const int& no_recs);
	void Write_Times(string times_Filename, const int& run, const time_t& Beg_time, time_t& End_time);
	void Write_Densities(const int& run, const time_t& Beg_time, time_t& End_time);
		void Update_Gil_Markov();
		void Write_Q_Info();
	void Change_Rho(double& rho);
	void Make_Rho_Vector();	
	string Create_Likelihoods_Filename();
	string Create_Full_Data_Filename();
	string Create_Times_Filename(const bool& Q_info);
	string Create_Densities_Filename();
	void Display_Run(const int& run, const int& rho_steps, int approxNumRuns = -1);

	void Write_Times();
};	


#endif

