//#pragma warning(disable: 4786)
#include "Cardin_McV.h"

using namespace std;


vector<double> CC_and_McV::CandMcV2_Mutation_Cost(int quad_points)
{	
	double k;			// k = # haps can be copied from.
	int time;
	vector<double> prob_mutation;
	prob_mutation.resize(quad_points);
	
	if (time_correction)
	{
		k = (double)actv_lngs;				// This is the Li&Stephens time correction.
	}
	else
	{
		k = ((double)actv_lngs+1)/2;		// This is our new time correction.
	}	
	
	for (time=0; time<quad_points; time++)
	{
		//prob_mutation.at(time) = 0.5 * (1 - exp(-1.0 * (abscissae.at(time)/k) * con.theta)); 
		if(using_ms_theta)
		{
			prob_mutation.at(time) = 0.5 * (1 - exp(-1.0 * (abscissae.at(time)/(double)k) * con.theta)); 
		}
		else
		{
			if(watterson == -1)
			{
				cout << "Warning, using a Watterson estimate of theta is not ready yet\n";
				getchar();
			}
			prob_mutation.at(time) = (2.0*abscissae.at(time)/(double)k) / (watterson + 2.0*abscissae.at(time)/(double)k); 
		}
		
		if (-prob_mutation.at(time) == numeric_limits<double>::infinity() || prob_mutation.at(time) == numeric_limits<double>::infinity())
		{ cout << "prob_mutation.at(" << time << ") seems to be (+/-)infinity here." << endl; getchar();}
	}
                                                                                       	
	if(DEBUG>1 && CandMcV2_DEBUG)
	{
		int display;
		//	cout << "Would you like to see the mutation costs? (0 for no and 1 for yes)" << endl;
		// cin << display
		display = 0;
		if (display)
		{
			for (int i=0; i<quad_points; i++)
			{
				cout << "The mutation cost at the " << i << "th abscissa in C&McV2 is: " << prob_mutation.at(i) << endl;
			}
		}
	}
	return prob_mutation;
}











long double CC_and_McV::R_bar(vector<double> R, int quad_points)
{
	int time;
	long double r_bar = 0;

	for (time=0; time<quad_points; time++)
	{
		r_bar += weights.at(time) * R.at(time);
	}
	return r_bar;
}











vector<double> CC_and_McV::CandMcV2_Recombination_Cost(int current_site, int quad_points, int increment)
{
	int time;
	double k;		// k = # haps can be copied from.
	double distance_times_rho;
	vector<double> prob_recombination;
	prob_recombination.resize(quad_points);
	
	distance_times_rho = fabs(rec_map.at(current_site) - rec_map.at(current_site - increment));
	
	if (time_correction)
	{
		k = (double)actv_lngs;				// This is the Li&Stephens time correction.
	}
	else
	{
		k = ((double)actv_lngs+1)/2;		// This is our new time correction.
	}	
	
	for (time = 0; time < quad_points; time++) 
	{	
		//prob_recombination.at(time) = 1 - (exp((-1.0) * abscissae.at(time)/((double)k) * (distance_times_rho/(double)2)));	// This assumes rho = 4Nr
		prob_recombination.at(time) = 1 - exp(-1.0 * (abscissae.at(time)/k) * distance_times_rho);	// This assumes rho = 4Nr
	}
	
	if(DEBUG>1 && CandMcV2_DEBUG)
	{
		int display;
		//cout << "Would you like to see the recombination costs? (0 for no and 1 for yes)" << endl;
		// cin << display
		display = 1;
		if (display)
		{
			for (int i=0; i<quad_points; i++)
			{
				cout << "In C&McV2 Rec[" << i << "] = " << prob_recombination.at(i) << endl;
			}
		}
	}
	return prob_recombination;
}








void CC_and_McV::Look_Down_and_Copy(const haplotype& alpha, const bool& reverse, vector<double>& lProb_up_to, double& loutput_likelihood)
{
	double log_2 = 0.69314718055994530941;
	int i,j,time, previous_time;
	double no_recomb, recomb;
	int quad_points=4;
	long double r_bar;		// weighted average recombination rate = Sum(w_i*R[i])
	int Beg, End, increment;
	
	vector<double> M;	M.resize(quad_points);		// M = Mutation Cost. 
	vector<double> R;	R.resize(quad_points);		// R = Recombination Cost. 
	
	if(!reverse)
	{
		Beg	  = 0; // This is site that the algorithm (Begins at) first considers
		End	  = no_segsts-1; // This is the index last site that the algorithm looks at.
		increment = 1;
	}
	else
	{
		Beg	  = no_segsts-1;
		End	  = 0;
		increment = -1;
	}
	
	vector < vector < vector<double> > > lPartial_from;
	
	vector < vector<double> > lsummed_over_seqs;
	lsummed_over_seqs.resize(no_segsts);
	for (i=0; i<no_segsts; i++)
	{
		lsummed_over_seqs.at(i).resize(quad_points, -1E308);
	}
		
	vector < vector < vector<double> > > lPartial_Likelihood;
	lPartial_Likelihood.resize(actv_lngs);			// Partial LogLikelihood at each site | who copied from at last site & time at this site
	for (i=0; i<actv_lngs; i++)
	{
		lPartial_Likelihood.at(i).resize(no_segsts);
	}
	for (i=0; i<actv_lngs; i++)
	{
		for (j=0; j<no_segsts; j++)
		{
			lPartial_Likelihood.at(i).at(j).resize(quad_points,-1E308);
		}
	}
	vector < vector<double> > ltime_summed_prtl;		// Partial LogLikelihood at each site | who copied from at last site
	ltime_summed_prtl.resize(actv_lngs);
	for (i=0; i<actv_lngs; i++)
	{
		ltime_summed_prtl.at(i).resize(no_segsts, -1E308);
	}

	//vector<double> lProb_up_to;						// Parital LogLIkelihood at each site.
	lProb_up_to.clear();
	lProb_up_to.resize(no_segsts,-1E308);
	
	if (CandMcV2_DEBUG) 	{cout << endl << "________________________________________________________________________________"<< endl;}

	Quadrature(quad_points);
	M	= CandMcV2_Mutation_Cost(quad_points);	

	for (i=0; i<actv_lngs; i++)
	{
		if (alpha.alleles.at(Beg) > 0 && hap_list.at(index_live.at(i)).alleles.at(Beg) > 0)
		{
			for (time=0; time<quad_points; time++)
			{
				if (alpha.alleles.at(Beg) == hap_list.at(index_live.at(i)).alleles.at(Beg))
				{
					lPartial_Likelihood.at(i).at(Beg).at(time) =  log(weights.at(time));
					lPartial_Likelihood.at(i).at(Beg).at(time) += log(1.0-(M.at(time)));				
					lPartial_Likelihood.at(i).at(Beg).at(time) += log(1.0/(double)actv_lngs);
				}
				else
				{
					lPartial_Likelihood.at(i).at(Beg).at(time) =  log(weights.at(time));
					lPartial_Likelihood.at(i).at(Beg).at(time) += log(M.at(time));				
					lPartial_Likelihood.at(i).at(Beg).at(time) += log(1.0/(double)actv_lngs);
				}
				ltime_summed_prtl.at(i).at(Beg)	 = Add_Log(ltime_summed_prtl.at(i).at(Beg)   , lPartial_Likelihood.at(i).at(Beg).at(time));
				lsummed_over_seqs.at(Beg).at(time) = Add_Log(lsummed_over_seqs.at(Beg).at(time), lPartial_Likelihood.at(i).at(Beg).at(time));
			}
			lProb_up_to.at(Beg) = Add_Log(lProb_up_to.at(Beg), ltime_summed_prtl.at(i).at(Beg));		
		}
		else
		{
			if (alpha.alleles.at(Beg) < 0 || hap_list.at(index_live.at(i)).alleles.at(Beg) < 0)
			{
				cout << "Warning, should not have negative allele types as far as 'Cardin and McVean' is concerned.\nBeing treated as non-ancestral.";
				getchar();
				Display_Haplotypes();
			}
			if(alpha.alleles.at(Beg) == 0)
			{
				for (time=0; time<quad_points; time++)
				{
					lPartial_Likelihood.at(i).at(Beg).at(time) =  log(weights.at(time));
					//***There is no information here so the site is copied with probability 1.***
					lPartial_Likelihood.at(i).at(Beg).at(time) += log(1.0);				//(this being the sum of M and 1-M)
					lPartial_Likelihood.at(i).at(Beg).at(time) -= log((double)actv_lngs);
					
					ltime_summed_prtl.at(i).at(Beg)	 = Add_Log(ltime_summed_prtl.at(i).at(Beg)   , lPartial_Likelihood.at(i).at(Beg).at(time));
					lsummed_over_seqs.at(Beg).at(time) = Add_Log(lsummed_over_seqs.at(Beg).at(time), lPartial_Likelihood.at(i).at(Beg).at(time));
				}
				lProb_up_to.at(Beg) = Add_Log(lProb_up_to.at(Beg), ltime_summed_prtl.at(i).at(Beg));
			}
			else //so it shold be the thing being copied from that's 0 here.
			{
				for (time=0; time<quad_points; time++)
				{
					lPartial_Likelihood.at(i).at(Beg).at(time) =  log(weights.at(time));
					lPartial_Likelihood.at(i).at(Beg).at(time) -= log_2;				//This needs to be updated to frequency or something.
					lPartial_Likelihood.at(i).at(Beg).at(time) -= log((double)actv_lngs);
				
					ltime_summed_prtl.at(i).at(Beg)	 = Add_Log(ltime_summed_prtl.at(i).at(Beg)   , lPartial_Likelihood.at(i).at(Beg).at(time));
					lsummed_over_seqs.at(Beg).at(time) = Add_Log(lsummed_over_seqs.at(Beg).at(time), lPartial_Likelihood.at(i).at(Beg).at(time));
				}
				lProb_up_to.at(Beg) = Add_Log(lProb_up_to.at(Beg), ltime_summed_prtl.at(i).at(Beg));
			}			
			if (exp(lProb_up_to.at(Beg)) > 1.00001)
			{
				cout << "The probabilities have summed to: " << exp(lProb_up_to.at(Beg))<< ", more than 1!!." << endl;
				CandMcV2_DEBUG = 1;
				getchar();
			}			
		}
	}
	if (CandMcV2_DEBUG)
			cout << "Prob_up_to[" << Beg << "] = " << exp(lProb_up_to.at(Beg)) << endl;
	
	for (j=Beg + increment; (!reverse * (j<no_segsts)) + (reverse * (j>=0)); j = j + increment) //This is meant to implement the 'condition' function.
	{	
		R = CandMcV2_Recombination_Cost(j, quad_points, increment);	
		r_bar	= R_bar(R, quad_points);					
		for (i=0; i<actv_lngs; i++)					// This loop steps through  which sequence is being copied from
		{
			for (time=0; time<quad_points; time++)			// This loop adds up the likelihoods for each quadrature point.
			{
				if (CandMcV2_DEBUG)
				{
					cout << "weight = " << weights.at(time) << "\t Cost of zero mutations = " << 1-M.at(time) << endl;
					cout << "C&McV2 Likelihood[" << i << "][" << j-increment << "][" << time << "]: " << exp(lPartial_Likelihood.at(i).at(j-increment).at(time)) << endl;
				}				
				if (rho !=0)
				{
					no_recomb	= lPartial_Likelihood.at(i).at(j-increment).at(time)	+ log(1.0-R.at(time));  // Can get R.(time)==1 (so get infinity here)
					
					if (-no_recomb == numeric_limits<double>::infinity()) // A bit of a hack?
						no_recomb = -1E308;
				
					recomb = -1E308;
					for (previous_time=0; previous_time < quad_points; previous_time ++)
						recomb	= Add_Log(recomb, lsummed_over_seqs.at(j-increment).at(previous_time) + log(R.at(previous_time)) + log(weights.at(time)*R.at(time)/r_bar));
					recomb = recomb - log((double)actv_lngs);
				}
				else
				{
					no_recomb = lPartial_Likelihood.at(i).at(j-increment).at(time);
					recomb = -1E308;
				}

				if (alpha.alleles.at(j) == hap_list.at(index_live.at(i)).alleles.at(j))		// Has there been a mutation?
				{	
					if (rho != 0)
					{
						if (alpha.alleles.at(Beg) > 0 && hap_list.at(index_live.at(i)).alleles.at(Beg) > 0)
							lPartial_Likelihood.at(i).at(j).at(time) = Add_Log(recomb,no_recomb) + log(1.0-M.at(time));
						else
						{
							if (alpha.alleles.at(Beg) < 0 || hap_list.at(index_live.at(i)).alleles.at(Beg) < 0)
							{
								cout << "Warning, should not have negative allele types as far as 'Cardin and McVean' is concerned.\nBeing treated as non-ancestral.";
								getchar();
							}
							if(alpha.alleles.at(Beg) == 0)
								lPartial_Likelihood.at(i).at(j).at(time) = Add_Log(recomb,no_recomb);
							else //So we're copying from something that = 0. Need to use FREQUENCY info or similar for this later
								lPartial_Likelihood.at(i).at(j).at(time) = Add_Log(recomb,no_recomb) - log_2;
						}
					}
					else
					{
						lPartial_Likelihood.at(i).at(j).at(time) = no_recomb + log(1.0-M.at(time));
					}
					if (CandMcV2_DEBUG)
					{
						cout << "weight = " << weights.at(time) << "\t Cost of zero mutations = " << 1-M.at(time) << endl;
						cout << "C&McV2 Likelihood[" << i << "][" << j << "][" << time << "]: " << exp(lPartial_Likelihood.at(i).at(j).at(time)) << endl;
					}
				}
				else
				{	
					lPartial_Likelihood.at(i).at(j).at(time) = Add_Log(recomb,no_recomb) + log(M.at(time));
					if (CandMcV2_DEBUG && 0)
					{
						cout << "weight = " << weights.at(time) << "\t Mut cost = " << M.at(time) << endl;
						cout << "C&McV2 Likelihood[" << i << "][" << j << "][" << time << "]: " << exp(lPartial_Likelihood.at(i).at(j).at(time)) << endl;
						cout << "Prob_up_to[" << j << "] = " << exp(lProb_up_to.at(j)) << endl;
					}
				}

				lsummed_over_seqs.at(j).at(time) = Add_Log(lsummed_over_seqs.at(j).at(time), lPartial_Likelihood.at(i).at(j).at(time));
				ltime_summed_prtl.at(i).at(j) = Add_Log(ltime_summed_prtl.at(i).at(j), lPartial_Likelihood.at(i).at(j).at(time));
				if (CandMcV2_DEBUG)
				{
					cout << endl;
					cout << "Summing over time, copying from: " << i << " at time: " << time << " gives: " << exp(ltime_summed_prtl.at(i).at(j)) << endl;
					cout << endl;
				}
			}
			if (CandMcV2_DEBUG)
				cout << "time_summed_prtl[" << i << "][" << j << "] = " << exp(ltime_summed_prtl.at(i).at(j)) << endl;				
			lProb_up_to.at(j) = Add_Log(lProb_up_to.at(j), ltime_summed_prtl.at(i).at(j));
			if (exp(lProb_up_to.at(j)) > 1.00001)
			{
				cout << "The probabilities have summed to: " << exp(lProb_up_to.at(j))<< ", more than 1!!." << endl;
				CandMcV2_DEBUG = 1;
				getchar();
			}
			
			if (CandMcV2_DEBUG)
			{cout << "Partial up to " << j << " is: " << exp(lProb_up_to.at(j)) << endl;
			}
		}
		if (CandMcV2_DEBUG)
			cout << "Prob_up_to[" << j << "] = " << exp(lProb_up_to.at(j)) << endl;
	}
	loutput_likelihood =-1E308;
	for (i=0; i<actv_lngs; i++)
	{
		loutput_likelihood = Add_Log(loutput_likelihood,ltime_summed_prtl.at(i).at(End));
	}
}




















// This functino should perform Gaussian quadrature for a limmited selection of quadrature points, for integrations on the REAL LINE with exponential weighting functions. 
// It uses the roots (and the corresponding weightings) of the Laguerre polynomials.
void CC_and_McV::Quadrature(int quad_points)
{
	switch (quad_points)
	{
	case 1:
		{
			weights.resize(quad_points);		// where quad_points is 4 in this case
			abscissae.resize(quad_points);		// where quad_points is 4 in this case

			weights.at(0)   = 1;
			abscissae.at(0) = 1;			
			break;

		}
	case 2:
		{cout << "This sounds like too few." << endl; getchar();}
	case 4:
		{
			weights.resize(quad_points);		// where quad_points is 4 in this case
			abscissae.resize(quad_points);		// where quad_points is 4 in this case

			weights.at(0) = 0.603154104341633602;
			weights.at(1) = 0.357418692437799687;
			weights.at(2) = 0.388879085150053843E-1; // Note there are zeroes after the decimal points here!
			weights.at(3) = 0.539294705561327450E-3; // Note there are zeroes after the decimal points here!

			abscissae.at(0) = 0.322547689619392312;
			abscissae.at(1) = 1.74576110115834658;
			abscissae.at(2) = 4.53662029692112798;
			abscissae.at(3) = 9.39507091230113313;

			break;

		}
	case 8:
		{cout << "This sounds like quite a few, I haven't put in the quadrature points for this yet" << endl; getchar();break;}
	case 16:
		{cout << "This sounds like quite a few, I haven't put in the quadrature points for this yet" << endl; getchar();break;}
	default:
		{
			cout << "The have not used a standard number of quadrature points, you must input the abscissae and weights yourself" << endl;
			getchar();
			break;
		}

	}		
	/*integral = 0;
	for (i=0; i<quad_points; i++)
	{
		integral += weights.at(i) * FandD_function(abscissae.at(i));
	}*/
}
