#include "Imp_Sampler.h"
#include "Ancestry.h"


void Imp_Sampler::Update_Imp_Sampler(int input_actv_lngs, int input_dead_lngs, vector<int> input_live_lineages, vector<int> input_index_live, double input_rec_rate, 
										double input_coal_rate, double input_mut_rate, double input_evnt_rate, int input_evnt_nmbr, vector<haplotype>& input_hap_list)
{	
    int i;
	actv_lngs		= input_actv_lngs;
	dead_lngs		= input_dead_lngs;
	live_lineages	= input_live_lineages;
	index_live		= input_index_live;
	evnt_nmbr		= input_evnt_nmbr;
	
	pr_rec_rate		= input_rec_rate;
	pr_coal_rate	= input_coal_rate;
	pr_mut_rate		= input_mut_rate;
	pr_evnt_rate	= input_evnt_rate;
	
	//hap_list		= input_hap_list;

	// I wonder if there is a more efficient way of doing these:
	// can_coals[smaller_index][larger_index]
	can_coals.clear(); // This is necessary, otherwise the resize function doesn't do anything (so vector not repopulated with 1's)
	can_coals.resize(actv_lngs);
	for(i=0; i<actv_lngs; i++)
	{
		can_coals.at(i).resize(actv_lngs,1);
	}
	// can_muts[site][lineage]
	can_muts.clear();
	can_muts.resize(no_segsts+1); 
	for(i=0; i<no_segsts+1; i++)
		can_muts.at(i).resize(actv_lngs+1,0);
}




int Imp_Sampler::Choose_from_all_events(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& event_weight)
{
	cout << "You have called the base class version of this function. Only the derived class virtual overloaded functions should be called.\n";
	getchar();
	return 0;
}


//int CSimple_IS::count_coal_pairs(const int& actv_lngs, const int& SMC, const int& no_segsts, const vector<haplotype>& hap_list, const int& rec_sites)
int Imp_Sampler::count_coal_pairs(CEvent& cheap_event)
{
	int i, j, k;
	int no_can_coal;

	//if(DEBUG>4) Display_Haplotypes();
	if (0)
	{
		cout << "The matrix of coalesceable pairs, BEFORE recounting, is:";
		for (int i=0; i< actv_lngs; i++)
		{
			for(int j=0; j<i; j++)
			{
				cout << can_coals.at(j).at(i) << " ";
			}
			cout << "\n";
		}
	}
	
	// The algorithm works by starting with the maximum possible number of coalescences and then reducing the number for every combination that 'doesn't work'
	no_can_coal = (actv_lngs)*(actv_lngs-1)/2;
	// This is potentially where the SMC makes a lot of gains. It stop wasting time coalescing pairs that can only coalesce because they don't overlap?
	if (SMC)
	{
		for (i=1; i<actv_lngs; i++)
		{
			for(j=0; j<i; j++)
			{
				//if(Check_Coal(i,j, hap_list)) // This lines checks that the two lineages share ancestral material. Doing it this way SAVES ENCAPSULATION.
				if(Check_Coal(index_live.at(i),index_live.at(j))) // This lines checks that the two lineages share ancestral material.
				{
					for(k=0; k<no_segsts; k++)
					{
						// This loops checks whether the haplotypes are the SAME wherever they share ancestral material.
						if(Ancestral(hap_list.at(index_live.at(i)).alleles.at(k)) && Ancestral(hap_list.at(index_live.at(j)).alleles.at(k))) // This line checks that both sequences have ancestral material at this site.
						{
							if(hap_list.at(index_live.at(i)).alleles.at(k) != hap_list.at(index_live.at(j)).alleles.at(k))
							{
								//can_coals[smaller_index][larger_index]
								can_coals.at(j).at(i) = 0;
								no_can_coal --;
								break;
							}
						}
					}
				}
				else
				{
					can_coals.at(j).at(i) = 0;	
					no_can_coal --;								
				}
				if(can_coals.at(j).at(i))
				{
					cheap_event.evnt_seq = i;
					if(con.first_nonrec)
					{
						cheap_event.event_type = 1;
						cheap_event.daughter1	= i;
						cheap_event.daughter2	= j;
						cheap_event.levent_weight= 1;
						return(1);
					}
				}
			}
		}
	} 
	else
	{
		for (i=1; i<actv_lngs; i++)
		{
			for(j=0; j<i; j++)
			{
				for(k=0; k<no_segsts; k++)
				{
					// This loops checks whether the haplotypes are the SAME wherever they share ancestral material.
					if(Ancestral(hap_list.at(index_live.at(i)).alleles.at(k)) && Ancestral(hap_list.at(index_live.at(j)).alleles.at(k)))
					{
						if(hap_list.at(index_live.at(i)).alleles.at(k) != hap_list.at(index_live.at(j)).alleles.at(k))
						{
							can_coals.at(j).at(i) = 0;
							no_can_coal --;
							break;
						}

					}
				}
				if(can_coals.at(j).at(i))
				{
					cheap_event.evnt_seq = i;
					if(con.first_nonrec)
					{
						cheap_event.event_type = 1;
						cheap_event.daughter1	= i;
						cheap_event.daughter2	= j;
						cheap_event.levent_weight= 1;
						return(1);
					}
				}
			}
		}
	}
	if (0)
	{
		cout << "The matrix of coalesceable pairs is: \n";
		for (int i=0; i< actv_lngs; i++)
		{
			for(int j=0; j<i; j++)
			{
				cout << can_coals.at(j).at(i) << " ";
			}
			cout << "\n";
		}
	}
	return no_can_coal;
}



// This is OVERRIDEN in Li & Stephens.
// Should also have a null event where nothing changes (this is for when SMC and rho is low and want to leave coal rate unchanged, should increase speed)
// This chooses an event based on the prior rates and the IS upweightings.
int Imp_Sampler::Choose_Event(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight, const double& rec_rate, const double& coal_rate, const double& mut_rate, const double& evnt_rate, const double& threshold)
{
	int reject_coal;
	double  choose_evnt;
	int event = std::numeric_limits<double>::quiet_NaN();

	
	if(importance_sample)
	{
		//Here is where the real importance sampling work is done:
		//********************************************************************
		calc_upweightings(daughter1, daughter2, rec_lineage, rec_site, mut_lineage, mut_site, new_type);
		//********************************************************************
		IS_coal_rate = 	pr_coal_rate * coal_upweight;
		IS_mut_rate  = 	pr_mut_rate  * mut_upweight;
		IS_rec_rate  = 	pr_rec_rate  * rec_upweight;
	}
	else
	{
		IS_coal_rate = pr_coal_rate;
		IS_mut_rate  = 	pr_mut_rate;
		IS_rec_rate  = 	pr_rec_rate;
	}
	IS_evnt_rate = IS_coal_rate + IS_mut_rate + IS_rec_rate;

	
	reject_coal = 1; int crash = 0;
	while(reject_coal && crash < 10000)
	{	
		crash ++;
		choose_evnt	= ran->U(); 
		if (choose_evnt < IS_coal_rate/IS_evnt_rate) 
		{
			// This needs to take TYPES into account:
			Choose_Pair(daughter1, daughter2); // Note that daughter1 is chosen to be < daughter2 (swapped if not)
			if (SMC)
			{
				if(Check_Coal(daughter1, daughter2))
				{
					reject_coal = 0;
					event = 1; //Coalescence
				}					
			}
			else 
			{
				reject_coal = 0;
				event = 1; //Coalescence
			}
		}
		else 
		{
			reject_coal = 0;
			if (choose_evnt< (IS_coal_rate + IS_rec_rate)/IS_evnt_rate) event = 2; //Recombination
			else 
			{
				event = 3;
				// The mutation details are now chosen within 'calc_upweightings'
				// Choose_Mutation(mut_lineage, mut_site, new_type);
			}
		}
	}
	if (crash > 9999)
	{
		cout << "Was unable to find a pair of sequences to coalesce.\nThis indicates a disparity between can_coals and the checking code here.\n";
		getchar();
	}
	return event;
}



//Pick two sequences of the same type to coalesce.	
void Imp_Sampler::Choose_Pair(int& daughter1, int& daughter2)
{
	int seq1, seq2,seq_temp;

	if (0)
	{
		cout << "The matrix of coalesceable pairs is: \n";
		for (int i=0; i< actv_lngs; i++)
		{
			for(int j=0; j<i; j++)
			{
				cout << can_coals.at(j).at(i) << " ";
			}
			cout << "\n";
		}
	}
	
	// This method is inefficient when there are many sequences, most pairs of which cannot coalesce.
	// Better to choose amongst those that can! Yes, needs to be UPDATED.
	int stop = 0; int crash = 0;	
	while(!stop && crash<10000)
	{
		// Choosing a pair of lineages to coalesce from the live lineages.
		seq1 = ran->discrete(0,actv_lngs-1);
		seq2 = ran->discrete(0,actv_lngs-2);
		if (seq2 >= seq1) seq2 ++;// This line ensures that we don't pick the same lineage twice.
		if(seq2<seq1){seq_temp = seq1; seq1=seq2; seq2 = seq_temp;} // This ensures that seq2>seq1.
		
		// can_coals[smaller_index][larger_index]   These smaller, often resized (slow?) vectors have no gaps in them.
		if(can_coals.at(seq1).at(seq2))
		{
			daughter1 = index_live.at(seq1);
			daughter2 = index_live.at(seq2);	
			stop = 1;	 
		}
		crash ++;
	}
	if(!stop)
	{
		cout << "Error: Unable to choose pair to coalesce\n";
		getchar();
	}
	
	if(daughter1 >= (int)hap_list.size() || daughter2 >= (int)hap_list.size()) {
		cout << "Warning! Have attempted to choose a non existant daughter lineage" << endl;
		getchar();
		exit(1);
	}
}






void Imp_Sampler::Calc_Prior_Rates(const vector<double>& lineage_rates, vector<int>& pr_coal_rates, vector<int>& pr_mut_rates, vector<int>& pr_rec_rates)
{
	//use index_live
	
	// This looks slow! May have to think about ways to update these.
	pr_mut_rates.resize(actv_lngs,0);
	pr_rec_rates.resize(actv_lngs,0);
	pr_coal_rates.resize(actv_lngs,0);

	//Calculate raw rates of events per chromosome.
	for(int i=0; i<actv_lngs; i++)
	{
		//Coal
		if(SMC)
		{
			// This needs to be done faster, there are other things that ask this same question, should not do it repeatedly. 
			for(int j=0; j<actv_lngs; j++)
			{
				if(j != i)
					pr_coal_rates.at(i) += Check_Coal(index_live.at(i), index_live.at(j));
			}
		}

		else
			pr_coal_rates.at(i) = actv_lngs - 1;

		//Mut
		for(int j=0; j<no_segsts; j++)
		{
			//if(Live_SiteI(hap_list.at(index_live.at(i)).alleles.at(j)))
			if(hap_list.at(index_live.at(i)).alleles.at(j)>0)
				pr_mut_rates.at(i) ++;
		}
		//Rec [Note, rec_dist already has the recombination map in it!!]
		//pr_rec_rates.at(i) = (int)(hap_list.at(index_live.at(i)).rec_dist/rho_per_site)+0.1);
		//cout << "Right edge: " << hap_list.at(index_live.at(i)).right_rec_edge << ", and left edge: " << hap_list.at(index_live.at(i)).left_rec_edge << "\n";
		pr_rec_rates.at(i) = hap_list.at(index_live.at(i)).right_rec_edge - hap_list.at(index_live.at(i)).left_rec_edge;
	}
}








int Imp_Sampler::Choose_haplotype(vector<double>& lineage_rates, vector<int>& pr_coal_rates, vector<int>& pr_mut_rates, vector<int>& pr_rec_rates, double& levent_weight, const vector<bool>& can_nonrec)
{
	int i;
    double random_number;
	random_number = ran->U();
	double sum  = 0  ;
	int  alpha  = std::numeric_limits<double>::quiet_NaN();
	bool chosen = 0  ;

	double theta_per_site = con.theta/(double)no_segsts; //This is already 'known' in ancestry, but couldn't be bothered to pass it around.
	for(i=0; i<actv_lngs; i++)
	{
		lineage_rates.at(i) = pr_coal_rates.at(i) + (hap_list.at(index_live.at(i)).rec_dist) + (theta_per_site * pr_mut_rates.at(i));
		if(con.choose_nonrec)
		{
			if(con.greedy_mle)
			{
				if(!can_nonrec.at(i))
					lineage_rates.at(i) = lineage_rates.at(i) / 1E10;
			}
			else
			{
				if(!can_nonrec.at(i))
					lineage_rates.at(i) = lineage_rates.at(i) * con.rho / con.no_segsts;
			}
		}
	}
	//NOTE: ALL of the above rates have been MULTIPLIED BY TWO. The total rate is HALF of the above.
	
	//Normalise Rates (and, now, find out the most likely one)
    int		best_hap		= 0;
	double	best_hap_val	= -1E308;

	sum_of_rates = 0;
	for(i=0; i<actv_lngs; i++)
	{
		sum_of_rates += lineage_rates.at(i);
		if(lineage_rates.at(i)>best_hap_val)
		{
			best_hap     = i;
			best_hap_val = lineage_rates.at(i);
		}
	}
	for(i=0; i<actv_lngs; i++)
		lineage_rates.at(i) = lineage_rates.at(i)/sum_of_rates;
	
	for(i=0; i<actv_lngs; i++)
	{
		sum += lineage_rates.at(i);
		//if((con.greedy_mle && i==best_hap)||(!con.greedy_mle && sum>random_number)) //This approach leads to repeatedly applying events to the same lineage until it has to recombine.
		if(sum>random_number)
		{
			alpha = i; //Note this is NOT the index that will be used in 'hap_list'
			levent_weight = log(lineage_rates.at(i)); // This is the probability of choosing this lineage.
			chosen =1;
			break;
		}
	}
	if(chosen != 1)
	{
		cout << "Warning, have failed to choose a haplotype in the LDaC scheme where the haplotype is chosen according to the prior\n";
		getchar();
	}
	if(DEBUG>=2) cout << "Choosing (unindexed) haplotype: " << alpha << "";
	return alpha;
 }


void Imp_Sampler::Enum_Events(vector<int>& pos_coals, vector<int>& pos_muts, const int& unindexed_alpha, int& no_pos_coals, int& no_pos_muts)
{
		// For each value of i need to find a lineage which alpha can naively coalesce with
		// Then determine whether it really can
				// If it can need to store this info so as to run the L_and_S cond function on it
				// Otherwise need to assign it a weight of zero and not waste time with L_and_S
	Hap_can_coal(pos_coals, unindexed_alpha, no_pos_coals);
	if(DEBUG>3) for(int i=0; i<actv_lngs; i++)
		cout << "pos_coals.at(" << i << ") = " << pos_coals.at(i) << "\n";
		

		// For each value of i need to find a site at which alpha is ancestral
		// Then need to determine whether alphas type at this site is singleton
				// If it is need to store this info so as to run the L_and_S cond function on removing this singleton
				// Otherwise need to assign a mutation event here a weight of zero and not waste L_and_S time on it.
	int check;
	CEvent non_event;
	count_singletons(non_event);
	
	// The following isn't really needed, can just use can_muts matrix all the way through!
	if(can_muts.at(no_segsts).at(unindexed_alpha) == 9)
	{
		check = 0;
		for(int k=0; k<no_segsts; k++)
		{
			if(can_muts.at(k).at(unindexed_alpha)) // I think this should be the type that we mutate to
			{
				pos_muts.at(k) = can_muts.at(k).at(unindexed_alpha);
				no_pos_muts ++;
				check = 1;
			}
		}
		if(!check)
			cout << "Warning, failed to find a site at which to mutate\n";
	}
}




void Imp_Sampler::Hap_can_coal(vector<int>& alpha_coals, const int& unindexed_alpha, int& no_pos_coals)
{
	for(int j=0; j<actv_lngs; j++)
	{
		if(unindexed_alpha != j)
		{
			// Will C++ still waste time no the second condition if the first has been satisfied?
			if(!SMC || Check_Coal(index_live.at(unindexed_alpha),index_live.at(j))) // This lines checks that the two lineages share ancestral material.
			{
				for(int k=0; k<no_segsts; k++)
				{
					// This loops checks whether the haplotypes are the SAME wherever they share ancestral material.
					if(Ancestral(hap_list.at(index_live.at(unindexed_alpha)).alleles.at(k)) && Ancestral(hap_list.at(index_live.at(j)).alleles.at(k))) // This line checks that both sequences have ancestral material at this site.
					{
						if(hap_list.at(index_live.at(unindexed_alpha)).alleles.at(k) != hap_list.at(index_live.at(j)).alleles.at(k))
						{
							alpha_coals.at(j) = 0;
							break;
						}
					}
					if(k == no_segsts-1)
					{
						alpha_coals.at(j) = 1;
						no_pos_coals ++;
					}
				}
			}
			else
			{
				alpha_coals.at(j) = 0;
			}
		}
		else 
			alpha_coals.at(j) = -2;
	}
	if(DEBUG>3)
	{
		cout << "The lineages that alpha can coalesce with are: \n";
		for(int j=0; j<actv_lngs; j++)
		{
			if(unindexed_alpha != j)
			{
				cout << alpha_coals.at(j) << " ";
			}
			else cout << " This is alpha ";
		}
		cout << "\n";
	}
}





void Imp_Sampler::Mut_create_alpha(haplotype& alpha, haplotype& beta, const int& index_alpha, int i)
{
	//Is it worth adding another check here to see that this is a singleton of the right form?
	//check
	if(!Junk_Lineage(index_alpha)) {cout << "Warning, this lineage should be dead\n"; getchar();}
	
	/*alpha.alleles.resize(no_segsts);
	alpha.left_rec_edge  = hap_list.at(index_alpha).left_rec_edge;
	alpha.right_rec_edge = hap_list.at(index_alpha).right_rec_edge;
	alpha.nodes			 = hap_list.at(index_alpha).nodes;			//probably not required.
	alpha.rec_dist		 = hap_list.at(index_alpha).rec_dist;*/
	//alpha = hap_list.at(index_alpha);
	beta  = hap_list.at(index_alpha);
	
	// This only happens at site *i*, the one passed to the function for mutation.
	if(hap_list.at(index_alpha).alleles.at(i) == 1) // Note that this is referring to a DEAD lineage, but one that has not been touched.
		beta.alleles.at(i) = 2;
	else
		beta.alleles.at(i) = 1;
}






// This arranges that all of the weightings on ONE haplotype sum to 1. This is what the event_weighting code in Pick_Event assumes.
void Imp_Sampler::Normalise_Weightings(vector<double>& lcoal_weights, vector<double>& lmut_weights, vector<double>& lrec_weights)
{
	double lsum_weightings	= -1E308;

	int event_counter		= 0;
	best_event_value        = -1E308;
	best_event_index		= 0;

	int i;
    for(i=0; i<(int)lcoal_weights.size(); i++)
	{
		// Do not need to do anything to use prior rate of coalescence
		lsum_weightings			 = Add_Log(lsum_weightings, lcoal_weights.at(i));
		
		if(lcoal_weights.at(i) > best_event_value)
		{
			best_event_index = event_counter;
			best_event_value = lcoal_weights.at(i);
		}
		event_counter++;
	}
	for(i=0; i<(int)lmut_weights.size(); i++)
	{
		lsum_weightings			 = Add_Log(lsum_weightings,  lmut_weights.at(i));

		if(lmut_weights.at(i) > best_event_value)
		{
			best_event_index = event_counter;
			best_event_value = lmut_weights.at(i);
		}
		event_counter++;
	}
	for(i=0; i<(int)lrec_weights.size(); i++)
	{
		lsum_weightings			 = Add_Log(lsum_weightings,  lrec_weights.at(i));
		if(lrec_weights.at(i) > best_event_value)
		{
			best_event_index = event_counter;
			best_event_value = lrec_weights.at(i);
		}
		event_counter++;
	}

	if(lsum_weightings<-1E305)
	{
		cout << "Warning! Probable loss of Precision in IS Weightings.\nFailure to Choose event properly suspected\n";
		getchar();
	}
	//Display_Haplotypes();	

	for(i=0; i<(int)lcoal_weights.size(); i++)
		lcoal_weights.at(i) = lcoal_weights.at(i) - lsum_weightings;
	for(i=0; i<(int)lmut_weights.size(); i++)
		lmut_weights.at(i)  = lmut_weights.at(i)  - lsum_weightings;
	for(i=0; i<(int)lrec_weights.size(); i++)
		lrec_weights.at(i)  = lrec_weights.at(i)  - lsum_weightings;
}



void Imp_Sampler::Normalise_All_Weightings(vector<vector<double> > & matrix_lcoal_weights, vector<vector<double> > & matrix_lmut_weights, vector<vector<double> > & matrix_lrec_weights, vector<double>& lhaplo_weights)
{
	double lsum_weightings = -1E308;
	int i, alpha;
	int passed_alpha;	
	double theta_per_site	= con.theta/(double)no_segsts;
    
	for(alpha=0; alpha<actv_lngs; alpha++)
	{
		passed_alpha = 0;
		for(i=0; i<actv_lngs; i++)
		{
			if(i!=alpha)
			{
				// Do not need to do anything to use prior rate of coalescence
				lsum_weightings						= Add_Log(lsum_weightings, matrix_lcoal_weights.at(alpha).at(i-passed_alpha));
			}
			else
				passed_alpha = 1;
		}
		for(i=0; i<no_segsts; i++)
		{
			//cout << "matrix_lmut_weights.at("<< alpha << ").at(" << i<< ") = " << matrix_lmut_weights.at(alpha).at(i) << "\n";			
			if(fabs(con.theta)> 1e-20)
				matrix_lmut_weights.at(alpha).at(i)		+= log(theta_per_site/2.0);
			else
				matrix_lmut_weights.at(alpha).at(i)		= -1E308;
			//cout << "matrix_lmut_weights.at("<< alpha << ").at(" << i<< ") = " << matrix_lmut_weights.at(alpha).at(i) << "\n";
			lsum_weightings							= Add_Log(lsum_weightings,  matrix_lmut_weights.at(alpha).at(i));
		}
		for(i=0; i<no_segsts-1; i++)
		{
			if(site_posns.at(i+1) - site_posns.at(i) > 0)
			{
				if(fabs(rho)> 1e-20 && (rec_map.at(i+1) - rec_map.at(i)>1E-10))
					matrix_lrec_weights.at(alpha).at(i)		+= log((rec_map.at(i+1) - rec_map.at(i))/2.0);			//matrix_lrec_weights.at(alpha).at(i)		+= log(rho_per_site);
				else
					matrix_lrec_weights.at(alpha).at(i)		 = -1E308;
			}
			else
			{
				if(fabs(rho)> 1e-20)
					matrix_lrec_weights.at(alpha).at(i)		+= log((rho * 0.00005)/2.0);
				else
					matrix_lrec_weights.at(alpha).at(i)		= -1E308;
			}
			//cout << matrix_lrec_weights.at(alpha).at(i) << " " << rec_map.at(i+1) << " " << rec_map.at(i) << " " << rec_map.at(i+1) - rec_map.at(i) << "\n";
			lsum_weightings							= Add_Log(lsum_weightings,  matrix_lrec_weights.at(alpha).at(i));
			//if(DEBUG>4)
			//	cout <<"recmap:" << rec_map.at(0) << " " << rec_map.at(1) << " " << rec_map.at(2) << " " << rec_map.at(3) << "\n";
		}
	}
	
	for(alpha=0; alpha<actv_lngs; alpha++)
	{
		passed_alpha = 0;
		lhaplo_weights.at(alpha) = -1E308;		
		for(i=0; i<actv_lngs; i++)
		{
				if(i!=alpha)
				{
					matrix_lcoal_weights.at(alpha).at(i-passed_alpha)	-= lsum_weightings;
					lhaplo_weights.at(alpha) 			= Add_Log(lhaplo_weights.at(alpha), matrix_lcoal_weights.at(alpha).at(i-passed_alpha));
				}
				else
					passed_alpha = 1;
		}
		for(i=0; i<no_segsts; i++)
		{
			matrix_lmut_weights.at(alpha).at(i)							-= lsum_weightings;
			lhaplo_weights.at(alpha) 			= Add_Log(lhaplo_weights.at(alpha), matrix_lmut_weights.at(alpha).at(i));
		}
		for(i=0; i<no_segsts-1; i++)
		{
			matrix_lrec_weights.at(alpha).at(i)							-= lsum_weightings;
			lhaplo_weights.at(alpha) 			= Add_Log(lhaplo_weights.at(alpha), matrix_lrec_weights.at(alpha).at(i));
		}
	}
	//Normalising haploweights - to make importance weight calculation easier (they now correspond to the probability of picking this haplotype)
	double lsum_haplo_weights = -1E308;
	for(alpha=0; alpha<actv_lngs; alpha++)
		lsum_haplo_weights = Add_Log(lsum_haplo_weights, lhaplo_weights.at(alpha));
	for(alpha=0; alpha<actv_lngs; alpha++)
		lhaplo_weights.at(alpha) -= lsum_haplo_weights;
}






// The 'unindexed_alpha' used here is the UNINDEXED_alpha used in Li&Stephens
void Imp_Sampler::Pick_Event(int& event, const int& unindexed_alpha, int& unin_daught2, int& daughter2, int& rec_site, int& mut_site, int& new_type, vector<double>& lcoal_upweightings, vector<double>& lmut_upweightings, vector<double>& lrec_upweightings, const vector<int>& pos_muts, double& levent_weight, const double& lhaplo_weighting)
{
	double lsum_weightings	= -1E308;
	bool chosen				= 0;
	int event_counter		= 0;
	
	double random_number;
	double lrandom_number;
	random_number = ran->U();
	lrandom_number = log(random_number);
	//lrandom_number += lhaplo_weighting; //This calibrates the number for the total weighting of this haplotype.

	//Coalescences
	int passed_alpha = 0;
	int i;
    for(i=0; i<actv_lngs && !chosen; i++)
	{
		if(i != unindexed_alpha)
		{
			lsum_weightings = Add_Log(lsum_weightings, lcoal_upweightings.at(i-passed_alpha));
			if((con.greedy_mle && event_counter==best_event_index) || 
			  (!con.greedy_mle && lrandom_number < lsum_weightings))
			{
				event = 1;
				unin_daught2 = i;
				daughter2 = index_live.at(i); 
				rec_site = -1;
				mut_site = -1;
				new_type = -1;
				// There is a change here as either of these two sequences could have been chosen for this one event.
				//levent_weight += log_2 + lcoal_upweightings.at(i-passed_alpha); // The probability of having chosen this haplotype comes from the code to choose the haplotype
				levent_weight += lcoal_upweightings.at(i-passed_alpha); // The probability of having chosen this haplotype comes from the code to choose the haplotype
				//******// I BACKTRACK here when 'choosing hap first' and I double this weighting in the LDaC function when doing it properly.
				chosen	 = 1;
				break;
			}
			event_counter++;
		}
		else
		{
			passed_alpha = 1;
			//When coal fast is used (and rec==0) the indexing isn't altered, when dynamic programming happens it is
			//The problem being that Normalise weightings works the same in both cases and so calculates weights for more events when coal_fast is used
			//This check is to make sure pick event skips events ONLY WHEN Normalise events had a reduced event number.
			//if(!use_dynamic_programming)
			//	event_counter++;
		}
	}

	//Mutations
	for(i=0; i<no_segsts && !chosen; i++)
	{
		lsum_weightings = Add_Log(lsum_weightings, lmut_upweightings.at(i));
		if((con.greedy_mle && event_counter==best_event_index) || 
		  (!con.greedy_mle && lrandom_number < lsum_weightings))
		{	
			event = 2;
			daughter2 = -1;
			rec_site = -1;
			mut_site =	i;
			new_type = pos_muts.at(i);			
			int index_alpha = index_live.at(unindexed_alpha);				
			if(new_type == hap_list.at(index_alpha).alleles.at(i))
			{
				cout << "Warning, have chosen an invalid mutation in the importance sampler - mutating `to' current type..\n"; 
				Sequence_Mutabilities();			
				getchar();			
			}
			if(new_type <1){
				cout << "Warning, have chosen an invalid mutation in the importance sampler - mutating at non-ancestral material.\n"; 
				cout << "The haplotype reads: "; Display_a_hap(hap_list.at(index_alpha)); cout << "\n";
				Sequence_Mutabilities();
				cout << "lmut_upweightings.at(" << i<< ") = " << lmut_upweightings.at(i) << "\n";
				getchar();
			}
			levent_weight += lmut_upweightings.at(i); // The probability of having chosen this haplotype comes from the code to choose the haplotype
			chosen	 = 1;
      		break;
		}
		event_counter++;
	}

	//Recombinations
	for(i=0; i<no_segsts-1 && !chosen; i++)
	{
		lsum_weightings = Add_Log(lsum_weightings, lrec_upweightings.at(i));
		if((con.greedy_mle && event_counter==best_event_index) || 
		  (!con.greedy_mle && lrandom_number < lsum_weightings))
		{	
			event = 3;
			daughter2 = -1;
			rec_site =  i;
			mut_site = -1;
			new_type = -1;
			levent_weight += lrec_upweightings.at(i); // The probability of having chosen this haplotype comes from the code to choose the haplotype
			chosen	 = 1;
			break;
		}
		event_counter++;
	}
	if(!chosen)
	{
		cout << "Program was unable to choose an event to happen in epoch: " << evnt_nmbr << "\n";
		getchar();
	}	
}


void Imp_Sampler::Choose_Mutation(int& mut_lineage, int& mut_site, int& new_type, const int& no_singletons)
{
	double sum_muts_passed = 0;
	bool chosen = 0;
	
	double random_number;
	random_number = ran->U();
	
	int site; int lineage;
	for(site=0; site<no_segsts && !chosen; site++)
	{
		if(can_muts.at(site).at(actv_lngs) == 9)
		{
			for (lineage=0; lineage<actv_lngs; lineage++)
			{
				if(can_muts.at(site).at(lineage) != 0)
					sum_muts_passed += 1.0/(double)no_singletons;
				if(random_number < sum_muts_passed)
				{
					mut_lineage = lineage;
					mut_site	= site;
					new_type	= can_muts.at(site).at(lineage);
					chosen = 1;
					break;
				}	
			}
		}
		else
		{
			if(can_muts.at(site).at(actv_lngs) != 0)
			{
				cout << "Warning, This value should only be 0 or 9, unable to determine whether there is a singleton at site: " << site << "\n";
				getchar();
			}
		}		
	}
}





// Currently this function counts singletons, however we may only wish to know if there are ANY in the full Sampler.
int Imp_Sampler::count_singletons(CEvent& cheap_event)
{	
	int i,k;
	int no_can_mut = no_segsts; //This is the number of SITES that can mutate. It is decreased by 1 every time a site with no (derived) singletons is encountered.
	
	if(know_derived)
	{
		int no_derived;
		int derived_type_on = -999999999; //std::numeric_limits<double>::quiet_NaN();
		for(k=0; k<no_segsts; k++)
		{
			no_derived  = 0;
			for (i=0; i<actv_lngs; i++)
			{	
				if(hap_list.at(index_live.at(i)).alleles.at(k) == 2)
				{
					no_derived++;
					if(no_derived == 1)
						derived_type_on = i;
					if(no_derived>1)
					{
						// can_muts[site][lineage]
						//can_muts.at(i).at(k) = 0; //Should be redundant as initialised to zero.
						no_can_mut --;
						i = actv_lngs;
					}
				}				
			}
			if(no_derived == 1)
			{
				can_muts.at(k).at(derived_type_on) = 1;
				can_muts.at(no_segsts).at(derived_type_on) = 9; // This is to indicate which lineages have a singleton on them.
				can_muts.at(k).at(actv_lngs) = 9; // This is to indicate which sites have a singleton on some lineage.
				cheap_event.evnt_seq = derived_type_on;
				if(con.first_nonrec)
				{
					cheap_event.event_type = 2;
					cheap_event.mut_site = k;
					cheap_event.new_type = 1;	// This is the type mutated TO when a mutation occurs.	
					cheap_event.levent_weight = 1;
					cheap_event.mut_lineage = derived_type_on; // This is the chosen sequence that recombines.
					return(1);
				}
			}
			if(no_derived == 0)
			{
				//can_muts.at(i).at(k) = 0;
				no_can_mut --;
			}
		}
	}
	else //Use first sequence as a reference type. // Have both DIFFERENT and SAME (looks redundant) in case of just two lineages.
	{
		int same_as_first=0;
		int different_from_first=0;
		int first_type;
		int first_index;
		int second_index;
		int second_type;
		
		for(k=0; k<no_segsts; k++)
		{
			same_as_first = 0; different_from_first = 0; first_type = 0; second_type =0; first_index = -1; second_index = -1;
			for (i=0; i<actv_lngs; i++)
			{	
				//if(!Live_SiteI(hap_list.at(index_live.at(i)).alleles.at(k)))// Non Live - so either non ancestral or reached MRCA (only ancestral lineage at this site)
				if(!(hap_list.at(index_live.at(i)).alleles.at(k)>0))// Non Live - so either non ancestral or reached MRCA (only ancestral lineage at this site)
				{	}
				else
				{
					if(first_type == 0) 
					{
						first_type = hap_list.at(index_live.at(i)).alleles.at(k);
						first_index = i;
					}

					if(hap_list.at(index_live.at(i)).alleles.at(k) == first_type)
						same_as_first ++; // Note that this will always be >= 1 (first ancestral sequence is always the same as itself)
					else
					{
						different_from_first ++;
						second_index = i; // Note this may be be the SINGLETON. That is what really decides which is mutant.
						second_type = hap_list.at(index_live.at(i)).alleles.at(k);
					}
					
					if (same_as_first>1 && different_from_first>1)
					{
						//can_muts.at(i).at(k) = 0;
						//i = actv_lngs; // this just jumps out of the loop.
						no_can_mut --;
						break; // This attempts to end only the most inner loop, it seems better practice than the above.
					}
				}
			}
			if(different_from_first == 0)
			{
				no_can_mut --;
			}
			if (same_as_first == 1 && different_from_first > 0) // So this type is singleton
			{
				// can_muts[site][lineage]
				can_muts.at(k).at(first_index) = second_type;
				can_muts.at(no_segsts).at(first_index) = 9; // This is to indicate which lineages have a singleton on them.
				can_muts.at(k).at(actv_lngs) = 9; // This is to indicate which sites have a singleton on some lineage.
				cheap_event.evnt_seq = k;
				if(con.first_nonrec)
				{
					cheap_event.event_type = 2;
					cheap_event.mut_site = k;
					cheap_event.new_type = second_type;	// This is the type mutated TO when a mutation occurs.	
					cheap_event.levent_weight = 1;
					cheap_event.mut_lineage = first_index; // This is the chosen sequence that recombines.
					return(1);
					}
			}
			if(different_from_first == 1) // So only one lineage was different to the first type, ie. that type is singleton.
			{
				// setting 'can_muts' to the type needed to mutate back to, here type on lineage 1 is at least as common.
				can_muts.at(k).at(second_index) = first_type;
				can_muts.at(no_segsts).at(second_index) = 9; // This is to indicate which lineages have a singleton on them.
				can_muts.at(k).at(actv_lngs) = 9; // This is to indicate which sites have a singleton on some lineage.
				cheap_event.evnt_seq = k;
				if(con.first_nonrec)
				{
					cheap_event.event_type = 2;
					cheap_event.mut_site = k;
					cheap_event.new_type = first_type;	// This is the type mutated TO when a mutation occurs.	
					cheap_event.levent_weight = 1;
					cheap_event.mut_lineage = second_index; // This is the chosen sequence that recombines.
					return(1);
				}
			}		
		}
	}
	if(DEBUG>3)
		Sequence_Mutabilities();
	return no_can_mut;
}

void Imp_Sampler::Sequence_Mutabilities()
{
	int i,j;
	cout << "The sequence mutabilities are:\n";
	for(j=0; j<actv_lngs+1; j++)			
	{
		for (i=0; i<no_segsts+1; i++)
		{
			// can_muts[site][lineage]	
			cout << can_muts.at(i).at(j);
			if (i== no_segsts-1) cout << " ";
		}
		cout <<"\n";
		if (j== actv_lngs-1) cout << "\n";
	}
}

void Imp_Sampler::calc_nonrec_sequences(vector<bool>& can_nonrec)
{
	int i,j;
	
  if(DEBUG>=2)
    cout << "sites at which mutation can occur: \n";
	for(i=0; i<actv_lngs; i++)
	{
	    if(DEBUG>3)
	    	cout << "Sequence (followed by sites that can mutate): i = " << i << ", j = ";
		for(j=0; j<no_segsts; j++)
		{
			if(can_muts.at(j).at(i))
			{
			  if(DEBUG>=2)
          cout << j << " ";
				can_nonrec.at(i) = true;
				break;
			}
		}
		if(DEBUG>3) cout << "\n";
	}
	for(i=0; i<actv_lngs; i++)
	{
		for(j=0; j<actv_lngs; j++)
		{
			if(can_coals.at(i).at(j) | can_nonrec.at(i))
			{
				can_nonrec.at(i) = true;
				break;
			}
		}
	}
}



void Imp_Sampler::Calc_index_live()
{
	int i;
	index_live.resize(0);
	//Compile a list of indexes of active lineages.
	for(i=0; i<(int)hap_list.size(); i++)
	{
		if(Lineage_Alive(i))
			index_live.push_back(i);
	}
}





	








void Imp_Sampler::Choose_alpha_properly(int& unindexed_alpha, const vector<vector<double> >& matrix_lcoal_weights, const vector<vector<double> >& matrix_lmut_weights, const vector<vector<double> >& matrix_lrec_weights, const vector<double>& lhaplo_weights, double& levent_weight)
{
    double random_number;
    double lrandom_number;
    
    random_number  = ran->U();
    lrandom_number = log(random_number);
    
    double lsum_weights;
    
    int i; int j; int passed_i;
    lsum_weights = -1E308;
    unindexed_alpha = -1;
	for(i=0; i<actv_lngs; i++)
    {
		passed_i = 0;    
		for(j=0; j<actv_lngs; j++)
		{
			if(i!=j)
			{
				lsum_weights = Add_Log(lsum_weights, matrix_lcoal_weights.at(i).at(j - passed_i));
				if(lsum_weights > lrandom_number)
				{
					unindexed_alpha = i;
					levent_weight	= lhaplo_weights.at(i);
					break;
				}
			}
			else
				passed_i = 1;
		}
		if(unindexed_alpha != -1)
			break;
		for(j=0; j<no_segsts; j++)
		{
			lsum_weights = Add_Log(lsum_weights, matrix_lmut_weights.at(i).at(j));
			if(lsum_weights > lrandom_number)
			{
				unindexed_alpha = i;
				levent_weight	= lhaplo_weights.at(i);
				break;
			}                
		}
		if(unindexed_alpha != -1)
			break;
		for(j=0; j<no_segsts-1; j++)
		{
            lsum_weights = Add_Log(lsum_weights, matrix_lrec_weights.at(i).at(j));
			if(lsum_weights > lrandom_number)
			{
				unindexed_alpha = i;
				levent_weight	= lhaplo_weights.at(i);
				break;
			}                
       }
       if(unindexed_alpha != -1)
          break;        
    }  
	if(unindexed_alpha == -1)
	{
		cout << "Warning, was unable to choose a haplotype to act on in Full alpha_choosing code" << endl;
		getchar();
	}
}






// This function checks for shared ancestral sites on ALREADY (Live) INDEXED lineages.
int Imp_Sampler::Check_Coal(int daughter1, int daughter2)
{
	if(!(hap_list.at(daughter1).right_rec_edge < hap_list.at(daughter2).left_rec_edge || hap_list.at(daughter2).right_rec_edge < hap_list.at(daughter1).left_rec_edge))
		return 1; //There is overlapping material, so they can coalesce
	else
		return 0; //There is no overlapping material
}









//Copied from Ancestry to check things in the Importance Sampler.
void Imp_Sampler::Display_Haplotypes()
{
	int i,j;
	int no_to_display;
	int max_lngs = 15;
	int length_to_display;
	int max_length = 15;
	int display_indices = 0;
	
	if(actv_lngs>max_lngs)
	{
		cout << "Warning, there are a lot of lineages to display, displaying first " << max_lngs << endl;
		no_to_display = max_lngs;
	}
	else no_to_display = actv_lngs;

	if(no_segsts>max_length)
	{
		cout << "Warning, there are a lot of lineages to display, displaying first " << max_lngs << endl;
		length_to_display = max_length;
	}
	else length_to_display = no_segsts;

	cout << "The haplotypes in the importance sampler are: \n";
	// Isn't there a problem here of displaying INACTIVE lineages?
	for (i=0; i< no_to_display; i++)
	{
		cout << "hap[" << index_live.at(i) << "(" << i << ")]  \t";
		for (j=0; j<length_to_display; j++)
		{
			cout  << hap_list.at(index_live.at(i)).alleles.at(j) << " ";
		}
		cout << "\n";
	}
	if(display_indices)
	{
		cout << "\nThe indices of the live lineages are:\n";
		for (i=0; i< no_to_display; i++)
			cout  << index_live.at(i) << " ";
		cout << "\n in this epoch are.\n";	
	}

}





void Imp_Sampler::Display_a_hap(const haplotype& this_hap)
{
	int i;
	//cout << "This hap = ";
	for(i=0; i<(int)this_hap.alleles.size(); i++)
	{
		cout << this_hap.alleles.at(i) << " ";
	}
	cout << "\n";

}





/*For possible problems with this method see end*/
void Imp_Sampler::calc_upweightings(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type)
{
	cout << "You have called the Imp_Sampler class version of this function. Only the Simple_IS class version should be called.\n";
	getchar();
}









// This function is used to identify sites which are ancestral and have not reached their MRCA.
/*bool Imp_Sampler::Live_SiteI(const int& allele)
{
	switch(allele)
	{
	case -2:		//This has reached its MRCA. It is ancestral. (and of type 2, (if(know_derived) this is mutant))
		return false;
	case -1:		//This has reached its MRCA. It is ancestral. (and of type 1)
		return false;
	case 0:			//This site is non ancestral.
		return false;
	case 1:			//This site is ancestral (and of type 1)
		return true;
	case 2:			//This site is ancestral (and of type 2) (if(know_derived) this is mutant)
		return true;
	default:
		cout << "Unrecognised type in alleles vector.\n";
		getchar();
		return false;
	}
}*/





// This function is used to identify sites which are ancestral. Even those which have reached their MRCA.
bool Imp_Sampler::Ancestral(const int& allele)
{
	switch(allele)
	{
	case -2:		//This has reached its MRCA. It is ancestral. (and of type 2, (if(know_derived) this is mutant))
		return true;
	case -1:		//This has reached its MRCA. It is ancestral. (and of type 1)
		return true;
	case 0:			//This site is non ancestral.
		return false;
	case 1:			//This site is ancestral (and of type 1)
		return true;
	case 2:			//This site is ancestral (and of type 2) (if(know_derived) this is mutant)
		return true;
	default:
		cout << "Unrecognised type in alleles vector.\n";
		getchar();
		return false;
	}
}



bool Imp_Sampler::Lineage_Alive(const int& lineage_index)
{
	if(live_lineages.at(lineage_index) == 1)
		return 1;
	else 
		return 0;
}


bool Imp_Sampler::Junk_Lineage(const int& lineage_index)
{
	if(live_lineages.at(lineage_index) == 0)
		return 1;
	else 
		return 0;
}
	







// This currently removes a mutation DETERMINISTICALLY and it should do it randomly (or according to some scheme, unlikely).
void Imp_Sampler::Remove_Mutation(int& mut_lineage, int& mut_site, int& new_type)
{
	int i,j;  //j is lineages, i is sites.
	int success = 0;
	if(no_segsts<actv_lngs)
	{
		for(i=0; i<no_segsts; i++)
		{		
			if(success)
				cout << "Warning: have already had a successful mutation and seem to be about to mutate again.\n";	//can_muts[site][lineage]			
			if(can_muts.at(i).at(actv_lngs) == 9)
			{
				for(j=0; j<actv_lngs; j++)
				{	//(haplist is a list of haplotypes so it's hap_list[lineage].alleles[site])
					if(can_muts.at(i).at(j) != 0)
					{   //check	:
						if(hap_list.at(index_live.at(j)).alleles.at(i) != can_muts.at(i).at(j))
						{
							//hap_list.at(index_live.at(j)).alleles.at(i) = can_muts.at(i).at(j);
							cout << "The wrong function for the job is trying to mutate, this should be done in ancestry\n";
							getchar();
							success = 1;
							break;
						}
						else
						{
							cout << "The haplotype should be different from can_muts at this site\n as it has been chosen to mutate to the can_muts value.\n";
							cout << "Hap: " << hap_list.at(index_live.at(j)).alleles.at(i) << " , Mut_table: " << can_muts.at(i).at(j) << "\n";
							getchar();
						}					
					}					
				}
				if(success == 0)
				{
					cout << "Warning, should have found a lineage on which to mutate at this site.\n";
					getchar();
				}
				break;
			}
			else
			{   //can_muts[site][lineage]			
				if(can_muts.at(i).at(actv_lngs) != 0)
				{
					cout << "can_muts should only take values of 0 or 9 at the high extremes\n";
					getchar();
				}
			}
		}
	}
	else
	{
		//Note the different ordering of these loops!
		for(j=0; j<actv_lngs; j++)
		{
			if(success)
				cout << "Warning: have already had a successful mutation and seem to be about to mutate again.\n";	//can_muts[site][lineage]			
			if(can_muts.at(no_segsts).at(j) == 9)
			{
				for(i=0; i<no_segsts; i++)
				{
					if(can_muts.at(i).at(j) != 0)
					{	//check
						if(can_muts.at(i).at(j) != 0 && hap_list.at(index_live.at(j)).alleles.at(i) != can_muts.at(i).at(j))
						{
							//hap_list.at(index_live.at(j)).alleles.at(i) = can_muts.at(i).at(j);
							cout << "The wrong function for the job is trying to mutate, this should be done in ancestry\n";
							getchar();							
							success = 1;
							break;
						}
						else
						{
							cout << "The haplotype should be different from can_muts at this site\n as it has been chosen to mutate to the can_muts value.\n";
							getchar(); DEBUG = 2;
						}					
					}
				}
				if(success == 0)
				{
					cout << "Warning, should have found a site to mutate at on this lineage.\n";
					getchar(); DEBUG = 2;
				}
				break;
			}
			else
			{
				if(can_muts.at(no_segsts).at(j) != 0)
				{
					cout << "can_muts should only take values of 0 or 9 at the high extremes\n";
					getchar(); DEBUG = 2;
				}
			}
		}
	}
	if(DEBUG>3)
	{
		cout << "The sequence mutabilities were:\n";
		for(j=0; j<actv_lngs+1; j++)			
		{
			for (i=0; i<no_segsts+1; i++)
			{
				cout << can_muts.at(i).at(j);
				if (i== no_segsts-1) cout << " ";
			}
			cout <<"\n";
			if (j== actv_lngs-1) cout << "\n";
		}
	}
	if(success == 0)
	{
		cout << "Error: Was unable to find a mutation to remove!\n";	
		getchar(); DEBUG = 2;
	}
}













// This function retains the method to pick amongst live lineages without using 'index_live'. It does NOT take TYPES into account.
void Imp_Sampler::old_Choose_Pair(int& daughter1, int& daughter2)
{
	int i,index;
	int seq1, seq2,seq_temp;

	// Choosing a pair of lineages to coalesce from the live lineages.
	seq1 = ran->discrete(0,actv_lngs-1);
	seq2 = ran->discrete(0,actv_lngs-2);
	if (seq2 >= seq1) seq2 ++;// This line ensures that we don't pick the same lineage twice.
	if(seq2<seq1){seq_temp = seq1; seq1=seq2; seq2 = seq_temp;} // This ensures that seq2>seq1.
	 

	// There may be a more efficient way of doing this loop. Such as updating a list of pointers to the active lineages 
	daughter2 = -1; i=0; index =0;
	while(daughter2 == -1)
	{
		if (index <= seq1)// Then assign daughter1
		{
			if (index == seq1 && Lineage_Alive(i))	daughter1 = i;	// This returns the hap_list index of the daughter haplotype.
			if (Lineage_Alive(i))					index ++;	// index counts up to the number of live lineages									
		}
		else			 // Then assign daughter 2
		{
			if (index == seq2 && Lineage_Alive(i))	daughter2 = i;	// This returns the hap_list index of the daughter haplotype.
			if (Lineage_Alive(i))					index ++;						
		}
		i++;		//i counts through all elements of live_lineages to find the correct index in haplist.
	}
	//if(daughter1 >= hap_list.size() || daughter2 >= hap_list.size()) {cout << "Warning! Have attempted to choose a non existant daughter lineage" << endl;getchar();exit(1);}
}


double Imp_Sampler::Add_Log(double Summand1, double Summand2)
{
	if (Summand1 == numeric_limits<double>::infinity() || Summand2 == numeric_limits<double>::infinity() || -Summand1 == numeric_limits<double>::infinity() || -Summand2 == numeric_limits<double>::infinity())
	{	cout << "A Summand seems to be (+/-) infinity here." << endl; getchar();}
	if (Summand1 < Summand2)
		return Summand2 + log(1 + exp(Summand1 - Summand2));
	else
		return Summand1 + log(1 + exp(Summand2 - Summand1));
}
