#ifndef IMPORTANCE_H
#define IMPORTANCE_H
//Memory leak detection:
//#define CRTDBG_MAP_ALLOC
#include <stdlib.h>
//#include <crtdbg.h>
#include <stdio.h>
#include <iostream>
#include <vector>
#include <limits> // This is so that I can set NaN values as defaults

//#include "nr.h" // numerical recipies stuff
#include "random.h"

#include "haplotype.h"
#include "recomb_event.h"
#include "node.h"
#include "control.h"
#include "event.h"


#define m_min(a,b) ((a) <= (b) ? (a) : (b))
#define m_max(a,b) ((a) >= (b) ? (a) : (b)) 

using namespace std;


// This class is not properly ENCAPSULATED. To save passing hap_list around I pass a reference to it once, and this is changed in Ancestry.
// This means that ANCESTRY in effect changes a member variable of IMP_SAMPLER.
class Imp_Sampler
{
public:
	//Variables that need to be changed on every event
	int evnt_nmbr;
	int actv_lngs;
	int dead_lngs;
	vector<int> live_lineages;
	vector<int> index_live;
	const vector<double>& site_posns;
	const vector<bool>& still_segregating;
	vector<haplotype>& hap_list;
	

	void Update_Imp_Sampler(int input_actv_lngs, int input_dead_lngs, vector<int> input_live_lineages, vector<int> index_live, double input_rec_rate, double input_coal_rate, double input_mut_rate, double input_evnt_rate, int input_evnt_nmbr, vector<haplotype>& input_hap_list);
	virtual void calc_upweightings(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type);
	int count_coal_pairs(CEvent& cheap_event);
	int count_singletons(CEvent& cheap_event);
	void calc_nonrec_sequences(vector<bool>& can_nonrec);
		void Sequence_Mutabilities();
	virtual int	Choose_Event(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight, const double& rec_rate, const double& coal_rate, const double& mut_rate, const double& evnt_rate, const double& threshold);
	        virtual int Choose_from_all_events(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& event_weight);
		
	int	 Check_Coal(int daughter1, int daughter2);
	void old_Choose_Pair(int& daughter1, int& daughter2);

	void Calc_index_live();
	
	void Calc_Prior_Rates(const vector<double>& lineage_rates, vector<int>& pr_coal_rates, vector<int>& pr_mut_rates, vector<int>& pr_rec_rates);
	int	 Choose_haplotype(vector<double>& lineage_rates, vector<int>& pr_coal_rates, vector<int>& pr_mut_rates, vector<int>& pr_rec_rates, double& levent_weight, const vector<bool>& can_nonrec);
	void Choose_alpha_properly(int& unindexed_alpha, const vector<vector<double> >& matrix_lcoal_weights, const vector<vector<double> >& matrix_lmut_weights, const vector<vector<double> >& matrix_lrec_weights, const vector<double>& lhaplo_weights, double& levent_weight);
	void Enum_Events(vector<int>& pos_coals, vector<int>& pos_muts, const int& unindexed_alpha, int& no_pos_coals, int& no_pos_muts);
		void Hap_can_coal(vector<int>& alpha_coals, const int& alpha, int& no_pos_coals);
		void Hap_can_mut(vector<int>& alpha_muts, const int& alpha);
		void Mut_create_alpha(haplotype& alpha, haplotype& beta, const int& index_alpha, int i);
		void Normalise_Weightings(vector<double>& lcoal_upweightings, vector<double>& lmut_upweightings, vector<double>& lrec_upweightings);
		void Normalise_All_Weightings(vector<vector<double> > & matrix_lcoal_weights, vector<vector<double> > & matrix_lmut_weights, vector<vector<double> > & matrix_lrec_weights, vector<double>& lhaplo_weights);
		void Pick_Event(int& event, const int& unindexed_alpha, int& unin_daught2, int& daughter2, int& rec_site, int& mut_site, int& new_type, vector<double>& lcoal_upweightings, vector<double>& lmut_upweightings, vector<double>& lrec_upweightings, const vector<int>& pos_muts, double& levent_weight, const double& lhaplo_weighting);
	// I expect these 3 functions to be overriden, certainly in the 1-D conditionals scheme
	void Choose_Pair(int& daughter1, int& daughter2);
	void Choose_rec_details(int& lineage, int& rec_from, int& rec_to);
	void Choose_Mutation(int& mut_lineage, int& mut_site, int& new_type, const int& no_singletons);
	
	Imp_Sampler(const Con& input_con, const double& input_rho, const double& input_rho_per_site,
			const int&  input_smpl_size,
			const int&  input_all_sites,
			const int&  input_no_segsts,
			const int&  input_rec_sites,
			double&     input_DEBUG,
			const bool& input_SMC,           const bool& input_cheap_coal_upwt,
			const bool& input_using_rec_map, const bool& input_importance_sample,
			const bool& input_coal_fast,
			const bool& input_greedy_mle,    const bool& input_know_derived,
			const vector<double>& input_site_posns,
			const vector<bool>&   input_still_segregating,
			vector<haplotype>&    input_hap_list,
			const vector<double>& input_rec_map,
			Random* Inst_of_random):
	site_posns(input_site_posns)              ,
	still_segregating(input_still_segregating),
	hap_list(input_hap_list)                  ,
	con(          input_con)                  , rho(            input_rho)            ,
	rho_per_site( input_rho_per_site)         , smpl_size(      input_smpl_size)      ,
	all_sites(    input_all_sites)            , no_segsts(      input_no_segsts)      ,
	rec_sites(    input_rec_sites)            ,
	DEBUG(        input_DEBUG)                ,
	SMC(          input_SMC)                  ,
	cheap_coal_upwt(input_cheap_coal_upwt)    ,
	using_rec_map(input_using_rec_map)        ,
	importance_sample(input_importance_sample),
	coal_fast(input_coal_fast)                ,
	rec_map(input_rec_map), greedy_mle(input_greedy_mle),
	know_derived(input_know_derived)
	{
		set_ran(Inst_of_random); // Need to look at how this line works.
	};
	virtual ~Imp_Sampler(){
	//	cout << "Inside the Imp_Sampler destructor\n";
	}; //destructor.

	 void set_ran(Random *ran_in){ran=ran_in;};

	 
	// can_muts[site][lineage]
	vector< vector<int> > can_muts;  //This is an indicator matrix that tell the user which sites can mutate and which is the singleton sequence.

	//Copied from Ancestry to check things in the Importance Sampler.
	void Display_Haplotypes();
	void Display_a_hap(const haplotype& this_hap);
	
	//Obsolete Function:
	void Remove_Mutation(int& mut_lineage, int& mut_site, int& new_type);
	
	 
protected:
	Random* ran; //I ought to read up on exactly how all of this Random class stuff is happening.	 
	
	//Variables that do not change throughout one ancestry
	const Con&  con;
	const double& rho;
	const double& rho_per_site;
	const int&  smpl_size;
	const int&  all_sites;
	const int&  no_segsts;
	const int&  rec_sites;
	double&	    DEBUG;
	const bool&	SMC;
	const bool& cheap_coal_upwt;
	const bool&	using_rec_map;
	const bool& importance_sample;
	const bool& coal_fast;
	const vector<double>& rec_map;
	const bool& greedy_mle;
	const bool& know_derived;


	double	pr_rec_rate;
	double	pr_coal_rate;
	double	pr_mut_rate;
	double	pr_evnt_rate;
	double  sum_of_rates;

	double	IS_rec_rate;
	double	IS_coal_rate;
	double	IS_mut_rate;
	double	IS_evnt_rate;

	int no_singletons;
	int no_identical;
	int greatest_identity;

	double  mut_upweight;
	double coal_upweight;
	double  rec_upweight;

	static vector<bool> can_nonrec;
	vector< vector<bool> > can_coals; //This is an upper triangular indicator matrix that tell the user when pairs of sequences can coalesce.
	double Add_Log(double Summand1, double Summand2);

	//bool Live_SiteI(const int& allele);		
	bool Ancestral(const int& allele);
	bool Lineage_Alive(const int& lineage_index);
	bool Junk_Lineage(const int& lineage_index);	

	double	best_event_value;
	int		best_event_index;
	bool    use_dynamic_programming; //This only makes a difference in the HMM based approaches when using coal_fast and hap 1st.

};

#endif


