#include "Li_and_Stephens.h"

using namespace std;


// Should return the mutation cost. Need to rethink this in  the conext of an importance sampler?
double CL_and_S::LandS_Prob_Mutation()
{
	double k;
	double prob_mutation;
	double theta_per_site = con.theta/(double)no_segsts;
	
	if (time_correction)
		k = (double)actv_lngs;				// This is the Li&Stephens time correction.
	else
		k = ((double)actv_lngs+1)/2;		// This is our new time correction.
	
	prob_mutation = (1.0/2.0) * theta_per_site/(k + theta_per_site); // I think this is the Li & Stephens mutation cost.
	
	return prob_mutation;
}

// Currently we assume constant recombination rate and the dynamic programming algorithm does not sum over different possible recombination rates at different sites.
// This function uses a function which Li&Stephens claim is designed to MIMIC THE EFFECTS of recombination, it has a slightly strange form as far as I can tell.
// Do I need to RECALIBRATE RHO for this function? Similarly for theta...
vector<double> CL_and_S::LandS_Recombination_Cost(int current_site, int increment)
{
	vector<double> prob_recombination;
	prob_recombination.resize(2);

	double k;
	double exp_part;
	double distance_times_rho;
	if(site_posns.at(current_site) - site_posns.at(current_site - increment) > 0)
		distance_times_rho = fabs(rec_map.at(current_site) - rec_map.at(current_site - increment));
	else
		distance_times_rho = rho * 0.00005; // This is a *hack* to deal with sites reported to be at the same location

	
	if (time_correction)
		k = (double)actv_lngs;				// This is the Li&Stephens time correction.
	else
		k = ((double)actv_lngs+1)/2;		// This is our new time correction.
	
	exp_part = exp(-distance_times_rho / k);	

	//prob_recombination.at(1) =				((double)1 - exp_part)/(double)actv_lngs;							// This is the probability of recombining to a different lineage.
	//prob_recombination.at(0) = exp_part +	((double)1 - exp_part)/(double)actv_lngs;				// This is the probabiltity no_recomb + prob of recomb back onto the same sequence.

	prob_recombination.at(1) = ((double)1 - exp_part)/(double)actv_lngs;							// This is the probability of recombining to a different lineage.
	prob_recombination.at(0) = exp_part;
	if (-log(prob_recombination.at(1)) == numeric_limits<double>::infinity()) // A bit of a hack?
	{
		if(DEBUG>1)
		{
			cout << "Warning: the probability of recombining in Li*Stephens here is 0!!" << endl;
			cout << "The positions here are: pos(" << current_site << ") = << site_posns.at(current_site) << and pos(" << current_site-1 << ") = " << site_posns.at(current_site-1) << endl;
		}
	}
	
	
	return prob_recombination;
}


// This scheme should, given a set of sequences, S, and another seperate sequence, X,  calculate the L&S P(X | S).
// The Li&Stephens scheme is effectively like the 'weather' example Gil described. Site by site.
void CL_and_S::Look_Down_and_Copy(const haplotype& alpha, const bool& reverse, vector<double>& lProb_up_to, double& loutput_likelihood)
{
	double log_2 = 0.69314718055994530941;
	double accuracy_threshold = 8; //16 Is probably lossless - 8 loses all past about 3 s.f. Likelihoods will be smaller
	int i,j; 
	double recomb, no_recomb;
	int Beg, End, increment;

	double M;	// M = Mutation Cost. M needs to be calculated on each iteration of the program.
	double log_R1;
	double log_R0;
	double log_m;
	double log_1minus_m;
	double log_actv_lngs;

	vector<double> R; R.resize(2);							// R = Recombination Cost. R (potentially) changes at every site. For constant rho should calculate it once in the function for speed.
	
	// NEEDS reconsidering.
	//double loutput_likelihood = 0;

	vector < vector <double> >	lPartial_Likelihood;
	lPartial_Likelihood.resize(actv_lngs);			// The partial (log) likelihood at each site needs to be stored assuming you last copied from a particular sequence.
	for (i=0; i<actv_lngs; i++)
	{
		lPartial_Likelihood.at(i).resize(no_segsts,-1E308);
	}
	
	//vector<double> lProb_up_to;
	lProb_up_to.clear();
	lProb_up_to.resize(no_segsts, -1E308);
		

	log_actv_lngs = log((double)actv_lngs);
	M = LandS_Prob_Mutation();
	log_m = log(M);
	log_1minus_m = log(1-M);
	// initialisation step of dynamic programming algorithm:

	if(!reverse)
	{
		if(quick_Li_and_Stephens)
		{
			Beg	  = alpha.left_rec_edge;  // This is site that the algorithm (Begins at) first considers
			End	  = alpha.right_rec_edge; // This is the index last site that the algorithm looks at.
		}
		else
		{
			Beg	  = 0; // This is site that the algorithm (Begins at) first considers
			End	  = no_segsts-1; // This is the index last site that the algorithm looks at.
		}
		increment = 1;
	}
	else
	{
		if(quick_Li_and_Stephens)
		{
			Beg	  = alpha.right_rec_edge;  // This is site that the algorithm (Begins at) first considers
			End	  = alpha.left_rec_edge;  // This is the index last site that the algorithm looks at.
		}
		else
		{
			Beg	  = no_segsts-1; // This is site that the algorithm (Begins at) first considers
			End	  = 0; // This is the index last site that the algorithm looks at.
		}
		increment = -1;
	}

	for (i=0; i<actv_lngs; i++)
	{
		if (alpha.alleles.at(Beg) > 0 && hap_list.at(index_live.at(i)).alleles.at(Beg)>0)
		{			
			if (alpha.alleles.at(Beg) == hap_list.at(index_live.at(i)).alleles.at(Beg))
			{
				lPartial_Likelihood.at(i).at(Beg) = log_1minus_m - log_actv_lngs;
				if (LandS_DEBUG)
					cout << "lPartial_Likelihood.at(" << i << ").at(" << Beg << "): " << exp(lPartial_Likelihood.at(i).at(Beg)) << endl;
			}
			else
			{
				lPartial_Likelihood.at(i).at(Beg) =  log_m - log_actv_lngs;
				if (LandS_DEBUG)
					cout << "lPartial_Likelihood.at(" << i << ").at(" << Beg << "): " << exp(lPartial_Likelihood.at(i).at(Beg)) << endl;
			}
		}
		else
		{
			 if (alpha.alleles.at(Beg) < 0 || hap_list.at(index_live.at(i)).alleles.at(Beg) < 0) //ie, if these sites have reached their MRCA
			 	lPartial_Likelihood.at(i).at(Beg) =  -log_actv_lngs;
			 if(alpha.alleles.at(Beg) == 0 || !(still_segregating.at(Beg))) // Only penalise for copying from non-anc material is site might be different.
				lPartial_Likelihood.at(i).at(Beg) =  -log_actv_lngs;
			 else
			 {
				 if(hap_list.at(index_live.at(i)).alleles.at(Beg) == 0)
					 lPartial_Likelihood.at(i).at(Beg) =  -log_actv_lngs - log_2; // This is an approximation will include frequency info eventually.
			 }

		}
		if(fabs(lProb_up_to.at(Beg) - lPartial_Likelihood.at(i).at(Beg))>accuracy_threshold)
			lProb_up_to.at(Beg) = m_max(lProb_up_to.at(Beg), lPartial_Likelihood.at(i).at(Beg));
		else
			lProb_up_to.at(Beg) = Add_Log(lProb_up_to.at(Beg), lPartial_Likelihood.at(i).at(Beg));
	}
	
	// Recursion step of dynamic programming algorithm:
	//for (j=Beg + increment; Condition(j, reverse); j = j + increment) //Adam says that 'condition' should eventually be 'inline'
	for (j=Beg + increment; (!reverse * (j<no_segsts)) + (reverse * (j>=0)); j = j + increment) //This is meant to implement the 'condition' function.
	{
		if(fabs(rho)> 1e-20)
			R = LandS_Recombination_Cost(j, increment); // Calculate R here until we introduce a real variable recombination rate. (See Recomb_Cost function). Replace '0' by 'j' if variable rec rate used.
			log_R1 = log(R.at(1));
			log_R0 = log(R.at(0));
		
		for (i=0; i<actv_lngs; i++)
		{
			if(fabs(rho)> 1e-20)
			{
				if (-log_R1 == numeric_limits<double>::infinity()) // A bit of a hack?
				{
					if(DEBUG>1)
						cout << "Warning: the probability of recombining at site: " << j << " in LandS when copying the: " << actv_lngs << "th sample was 0!!" << endl;
					recomb = -1E308;
				}			
				else
					recomb		= lProb_up_to.at(j-increment)				+ log_R1;
				
				no_recomb	= lPartial_Likelihood.at(i).at(j-increment) + log_R0;
				if (-no_recomb == numeric_limits<double>::infinity()) // A bit of a hack?
				{
					no_recomb = -1E308;
				}			
					
				if (LandS_DEBUG)
					cout << "Rho at this site = " << rec_map.at(Beg + increment + 1) - rec_map.at(Beg + increment) << " and i = " << i << " and recomb = " << recomb << " and no_recomb = " << no_recomb << endl;
								
				//if (Live_SiteI(alpha.alleles.at(j)) && Live_SiteI(hap_list.at(index_live.at(i)).alleles.at(j)))
				if (alpha.alleles.at(j) > 0 && hap_list.at(index_live.at(i)).alleles.at(j)>0)
				{
					if (alpha.alleles.at(j) == hap_list.at(index_live.at(i)).alleles.at(j))		// Has there been a mutation?
					{
						if(fabs(recomb - no_recomb)>accuracy_threshold)
							lPartial_Likelihood.at(i).at(j) = m_max(recomb, no_recomb) + log_1minus_m;				
						else
							lPartial_Likelihood.at(i).at(j) = Add_Log(recomb, no_recomb) + log_1minus_m;				

						if (LandS_DEBUG)
							cout << "exp(lPartial_Likelihood.at(" << i << ").at(" << j << "))): " << exp(lPartial_Likelihood.at(i).at(j)) << endl;
					}
					else
					{
						if(fabs(recomb - no_recomb)>accuracy_threshold)
							lPartial_Likelihood.at(i).at(j) = m_max(recomb, no_recomb) + log_m; 
						else
							lPartial_Likelihood.at(i).at(j) = Add_Log(recomb, no_recomb) + log_m; 
						if (LandS_DEBUG)
							cout << "exp(lPartial_Likelihood.at(" << i << ").at(" << j << "))): " << exp(lPartial_Likelihood.at(i).at(j)) << endl;
					}
				}
				else
				{
					if (alpha.alleles.at(j) < 0 || hap_list.at(index_live.at(i)).alleles.at(j) < 0)
					{
						if(fabs(recomb - no_recomb)>accuracy_threshold)
							lPartial_Likelihood.at(i).at(j) = m_max(recomb, no_recomb); 
						else
							lPartial_Likelihood.at(i).at(j) = Add_Log(recomb, no_recomb);
					}
					if(alpha.alleles.at(j) == 0 || !(still_segregating.at(j))) // Only penalise for copying from non-anc material is site might be different.
					{
						if(fabs(recomb - no_recomb)>accuracy_threshold)
							lPartial_Likelihood.at(i).at(j) = m_max(recomb, no_recomb); 
						else
							lPartial_Likelihood.at(i).at(j) = Add_Log(recomb, no_recomb);
					}
					else
					{
						if(hap_list.at(index_live.at(i)).alleles.at(j) == 0)
						{
							// This -log2 is my simpler approxiomation to Fearnheads Imputation by Frequency Method
							if(fabs(recomb - no_recomb)>accuracy_threshold)
								lPartial_Likelihood.at(i).at(j) = m_max(recomb, no_recomb)-log_2; 
							else
								lPartial_Likelihood.at(i).at(j) = Add_Log(recomb, no_recomb)-log_2;
						}
					}
				}
			}
			else
			{
				no_recomb	= lPartial_Likelihood.at(i).at(j-increment);
				if (alpha.alleles.at(j) > 0 && hap_list.at(index_live.at(i)).alleles.at(j) > 0)
				{
					if (alpha.alleles.at(j) == hap_list.at(index_live.at(i)).alleles.at(j))		// Has there been a mutation?
					{
						lPartial_Likelihood.at(i).at(j) = no_recomb + log_1minus_m;	
						if (LandS_DEBUG)
							cout << "exp(lPartial_Likelihood.at(" << i << ").at(" << j << "))): " << exp(lPartial_Likelihood.at(i).at(j)) << endl;
					}
					else
					{
						lPartial_Likelihood.at(i).at(j) = no_recomb + log_m; 
						if (LandS_DEBUG)
							cout << "exp(lPartial_Likelihood.at(" << i << ").at(" << j << "))): " << exp(lPartial_Likelihood.at(i).at(j)) << endl;

					}
				}
				else
				{
					if (alpha.alleles.at(0) < 0 || hap_list.at(index_live.at(i)).alleles.at(0) < 0)
						lPartial_Likelihood.at(i).at(j) =  no_recomb;
					if(alpha.alleles.at(j) <= 0 || !(still_segregating.at(j))) // Only penalise for copying from non-anc material is site might be different.
						lPartial_Likelihood.at(i).at(j) =  no_recomb;
					else
					{
						if(hap_list.at(index_live.at(i)).alleles.at(j) == 0)
							lPartial_Likelihood.at(i).at(j) =  no_recomb - log_2; // This is an approximation will include frequency info eventually.
					}
				}
			}
			if(fabs(lProb_up_to.at(j) - lPartial_Likelihood.at(i).at(j))>accuracy_threshold)
				lProb_up_to.at(j) = m_max(lProb_up_to.at(j), lPartial_Likelihood.at(i).at(j));
			else
				lProb_up_to.at(j) = Add_Log(lProb_up_to.at(j), lPartial_Likelihood.at(i).at(j));

			if (LandS_DEBUG)
				cout << "exp(lProb_up_to(" << j << ")) = " << exp(lProb_up_to.at(j)) << endl;
		}
	}

	loutput_likelihood =-1E308;
	for (i=0; i<actv_lngs; i++)
	{
		if(fabs(loutput_likelihood - lPartial_Likelihood.at(i).at(End))>accuracy_threshold)
			loutput_likelihood = m_max(loutput_likelihood,lPartial_Likelihood.at(i).at(End));
		else
			loutput_likelihood = Add_Log(loutput_likelihood,lPartial_Likelihood.at(i).at(End));
	}
	
	//cout << "Li and Stephens output for adding on the " << actv_lngs << "th haplotype is: " << exp(loutput_likelihood) << endl;
	//return lPartial_Likelihood;
}
















// This scheme should, given a set of sequences, S, and another seperate sequence, X,  calculate the L&S P(X | S).
// The Li&Stephens scheme is effectively like the 'weather' example Gil described. Site by site.
void CL_and_S::Look_Down_and_Copy_nolog(const haplotype& alpha, const bool& reverse, vector<double>& lProb_up_to, double& loutput_likelihood)
{
	int i,j; 
	double recomb, no_recomb;
	int Beg, End, increment;


	double M;	// M = Mutation Cost. M needs to be calculated on each iteration of the program.

	vector<double> R; R.resize(2);							// R = Recombination Cost. R (potentially) changes at every site. For constant rho should calculate it once in the function for speed.
	vector < vector <double> >	Partial_Likelihood;
	Partial_Likelihood.resize(actv_lngs);			// The partial (log) likelihood at each site needs to be stored assuming you last copied from a particular sequence.
	for (i=0; i<actv_lngs; i++)
	{
		Partial_Likelihood.at(i).resize(no_segsts,0);
	}
	
	//vector<double> lProb_up_to;
	lProb_up_to.clear();
	lProb_up_to.resize(no_segsts, -1E308);
	vector<double> Prob_up_to;
	Prob_up_to.resize(lProb_up_to.size(),0.0);
	double output_likelihood;
		
	M = LandS_Prob_Mutation();

	if(!reverse)
	{
		Beg	  = alpha.left_rec_edge;  // This is site that the algorithm (Begins at) first considers
		End	  = alpha.right_rec_edge; // This is the index last site that the algorithm looks at.
		increment = 1;
	}
	else
	{
		Beg	  = alpha.right_rec_edge;  // This is site that the algorithm (Begins at) first considers
		End	  = alpha.left_rec_edge;  // This is the index last site that the algorithm looks at.
		increment = -1;
	}

	for (i=0; i<actv_lngs; i++)
	{
		if (alpha.alleles.at(Beg) > 0 && hap_list.at(index_live.at(i)).alleles.at(Beg)>0)
		{			
			if (alpha.alleles.at(Beg) == hap_list.at(index_live.at(i)).alleles.at(Beg))
			{
				Partial_Likelihood.at(i).at(Beg) = (1-M) / ((double)actv_lngs);
				if (LandS_DEBUG)
					cout << "Partial_Likelihood.at(" << i << ").at(" << Beg << "): " << exp(Partial_Likelihood.at(i).at(Beg)) << endl;
			}
			else
			{
				Partial_Likelihood.at(i).at(Beg) =  (M) / ((double)actv_lngs);
				if (LandS_DEBUG)
					cout << "Partial_Likelihood.at(" << i << ").at(" << Beg << "): " << exp(Partial_Likelihood.at(i).at(Beg)) << endl;
			}
		}
		else
		{
			 if (alpha.alleles.at(Beg) < 0 || hap_list.at(index_live.at(i)).alleles.at(Beg) < 0) //ie, if these sites have reached their MRCA
			 	Partial_Likelihood.at(i).at(Beg) =  1.0/((double)actv_lngs);
			 if(alpha.alleles.at(Beg) == 0 || !(still_segregating.at(Beg))) // Only penalise for copying from non-anc material is site might be different.
				Partial_Likelihood.at(i).at(Beg) =  1.0/((double)actv_lngs);
			 else
			 {
				 if(hap_list.at(index_live.at(i)).alleles.at(Beg) == 0)
					 Partial_Likelihood.at(i).at(Beg) =  1.0/((double)actv_lngs*2.0); // This is an approximation will include frequency info eventually.
			 }

		}
		Prob_up_to.at(Beg) = Prob_up_to.at(Beg) + Partial_Likelihood.at(i).at(Beg);
	}
	
	// Recursion step of dynamic programming algorithm:
	//for (j=Beg + increment; Condition(j, reverse); j = j + increment) //Adam says that 'condition' should eventually be 'inline'
	for (j=Beg + increment; (!reverse * (j<no_segsts)) + (reverse * (j>=0)); j = j + increment) //This is meant to implement the 'condition' function.
	{
		if(fabs(rho)> 1e-20)
			R = LandS_Recombination_Cost(j, increment); // Calculate R here until we introduce a real variable recombination rate. (See Recomb_Cost function). Replace '0' by 'j' if variable rec rate used.
		else {
			R.at(0) = 1;
			R.at(1) = 0;
		}
		for (i=0; i<actv_lngs; i++)
		{
			if(fabs(rho)> 1e-20)
			{
				recomb		= Prob_up_to.at(j-increment) * R.at(1);				
				no_recomb	= Partial_Likelihood.at(i).at(j-increment) * R.at(0);
				if (LandS_DEBUG)
					cout << "Rho at this site = " << rec_map.at(Beg + increment + (reverse*1)) - rec_map.at(Beg + increment - 1 + (reverse*1)) << " and i = " << i << " and recomb = " << recomb << " and no_recomb = " << no_recomb << endl;
								
				if (alpha.alleles.at(j) > 0 && hap_list.at(index_live.at(i)).alleles.at(j)>0)
				{
					if (alpha.alleles.at(j) == hap_list.at(index_live.at(i)).alleles.at(j))		// Has there been a mutation?
					{
						Partial_Likelihood.at(i).at(j) = (recomb + no_recomb)*(1-M);				
						if (LandS_DEBUG)
							cout << "Partial_Likelihood.at(" << i << ").at(" << j << "))): " << (Partial_Likelihood.at(i).at(j)) << endl;
					}
					else
					{
						Partial_Likelihood.at(i).at(j) = (recomb + no_recomb) * M; 
						if (LandS_DEBUG)
							cout << "Partial_Likelihood.at(" << i << ").at(" << j << ")): " << (Partial_Likelihood.at(i).at(j)) << endl;
					}
				}
				else
				{
					if (alpha.alleles.at(j) < 0 || hap_list.at(index_live.at(i)).alleles.at(j) < 0)
					{
						Partial_Likelihood.at(i).at(j) = (recomb + no_recomb);
					}
					if(alpha.alleles.at(j) == 0 || !(still_segregating.at(j))) // Only penalise for copying from non-anc material is site might be different.
					{
						Partial_Likelihood.at(i).at(j) = (recomb + no_recomb);
					}
					else
					{
						if(hap_list.at(index_live.at(i)).alleles.at(j) == 0)
						{
							// This -log2 is my simpler approxiomation to Fearnheads Imputation by Frequency Method
							Partial_Likelihood.at(i).at(j) = (recomb + no_recomb)/2.0;
						}
					}
				}
			}
			else
			{
				no_recomb	= Partial_Likelihood.at(i).at(j-increment);
				if (alpha.alleles.at(j) > 0 && hap_list.at(index_live.at(i)).alleles.at(j) > 0)
				{
					if (alpha.alleles.at(j) == hap_list.at(index_live.at(i)).alleles.at(j))		// Has there been a mutation?
					{
						Partial_Likelihood.at(i).at(j) = no_recomb * (1-M);	
						if (LandS_DEBUG)
							cout << "Partial_Likelihood.at(" << i << ").at(" << j << ")): " << (Partial_Likelihood.at(i).at(j)) << endl;
					}
					else
					{
						Partial_Likelihood.at(i).at(j) = no_recomb * M; 
						if (LandS_DEBUG)
							cout << "exp(Partial_Likelihood.at(" << i << ").at(" << j << "))): " << (Partial_Likelihood.at(i).at(j)) << endl;

					}
				}
				else
				{
					if (alpha.alleles.at(0) < 0 || hap_list.at(index_live.at(i)).alleles.at(0) < 0)
						Partial_Likelihood.at(i).at(j) =  no_recomb;
					if(alpha.alleles.at(j) <= 0 || !(still_segregating.at(j))) // Only penalise for copying from non-anc material is site might be different.
						Partial_Likelihood.at(i).at(j) =  no_recomb;
					else
					{
						if(hap_list.at(index_live.at(i)).alleles.at(j) == 0)
							Partial_Likelihood.at(i).at(j) =  no_recomb/2.0; // This is an approximation will include frequency info eventually.
					}
				}
			}
			Prob_up_to.at(j) = (Prob_up_to.at(j) + Partial_Likelihood.at(i).at(j));

			if (LandS_DEBUG)
				cout << "(Prob_up_to(" << j << ")) = " << (Prob_up_to.at(j)) << endl;
		}
	}

	output_likelihood =0;
	for (i=0; i<actv_lngs; i++)
	{
		output_likelihood = (output_likelihood + Partial_Likelihood.at(i).at(End));
	}
	//Post processing before wrapper functions learn to cope with non-logged answers.
	loutput_likelihood = log(output_likelihood);
	for(i=0; i<(int)Prob_up_to.size(); i++)
		lProb_up_to.at(i) = log(Prob_up_to.at(i));
}




//For possible problems with this method see end
/*void CL_and_S::calc_upweightings(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type)
{
	cout << "You have called the Li and Stephens class version of this function. Only the Simple_IS class version should be called.\n";
	getchar();
}*/




