#include "LookD_Copy.h"

using namespace std;


int CLookD_Copy::Choose_Event(
    int& daughter1         , int& daughter2         , int& rec_lineage      , int& rec_site          ,
    int& mut_lineage       , int& mut_site          , int& new_type         , double& levent_weight  ,
    const double& rec_rate , const double& coal_rate, const double& mut_rate, const double& evnt_rate,
    const double& threshold
  )
{
	double log_2 = 0.69314718055994530941;
	double theta_per_site	= con.theta/(double)no_segsts;		
	double fast_adjustment;
	int event = -1; //1 = coal, 2 = mut, 3 = rec
	int i; 
	int index_alpha;
	int unindexed_alpha;
	haplotype alpha;

	vector<double>	lineage_rates;
	// These record the prior rate of each type of event on each lineage. (eg. 4 if 4 sites are ancestral, not 4*theta)
	vector<int> pr_coal_rates;
	vector<int> pr_mut_rates;
	vector<int> pr_rec_rates;
	
	lineage_rates.resize(actv_lngs);
	// Stochastically choose a haplotype, alpha, to act on according to the rate at which events happen to it.
	// Under MLE algorithm it is still essentially stochastic but seed is set to 1 (if not specified by user)
	// A BETTER MLE type approach would be to check for possible mutations/coalescences, but it's dangerous within 
	// an importance sampling program to do things which make the IS weights so complex so not implemented until I have much more time.
    Calc_Prior_Rates(lineage_rates, pr_coal_rates, pr_mut_rates, pr_rec_rates);
    // Hap_can_coal(pos_coals, unindexed_alpha, no_pos_coals); //Checks if the haplotype numbered unindexed_alpha can coal 

    // Hopefully just run count_coal_pairs() and count_singletons and these will update the relevant matrices.
    // can_muts and can_coals are member classes of 'Imp_Sampler'.
    CEvent cheap_event;
    cheap_event.evnt_seq = -1; cheap_event.daughter1 = -1; cheap_event.daughter2 = -1;
    count_singletons(cheap_event); //Updates the matrix can_muts
    if(!con.first_nonrec || cheap_event.evnt_seq == -1)
    	count_coal_pairs(cheap_event); //Updates the matrix can_coals
    //count_singletons(cheap_event); //Updates the matrix can_muts
    can_nonrec.resize(actv_lngs);
    for(i=0; i<(int)can_nonrec.size(); i++)
    	can_nonrec.at(i) = false;
    calc_nonrec_sequences(can_nonrec); //Returns which sequences can undergo events other than recombinations, these should be picked more often.
    if(DEBUG>3) {
    	cout << "Sequences that can do things other than recombine:\n";
    	for(int tmp=0; tmp<(int)can_nonrec.size(); tmp++)
    		cout << can_nonrec.at(tmp) << " ";
        cout << "\n";
    }
	if(con.first_nonrec && cheap_event.evnt_seq != -1)
	{
        if(cheap_event.event_type==1) daughter1   = index_live.at(cheap_event.daughter1);
        else daughter1 = -1;
        if(cheap_event.event_type==1) daughter2   = index_live.at(cheap_event.daughter2);  
        else daughter2 = -1;
        if(cheap_event.event_type==2) mut_lineage = index_live.at(cheap_event.mut_lineage) ; 
        else mut_lineage = -1;
        if(cheap_event.event_type==2) mut_site    = cheap_event.mut_site ;
        else mut_site = -1;
        if(cheap_event.event_type==2) new_type      = cheap_event.new_type;
        else new_type = -1;
        rec_lineage   = -1; rec_site    = -1 ;
        levent_weight = 0                    ; 
		return(cheap_event.event_type);
	}
	else
    {
		unindexed_alpha = Choose_haplotype(lineage_rates, pr_coal_rates, pr_mut_rates, pr_rec_rates, levent_weight, can_nonrec);
    }
 
    /*cout << "Lineage Rates:\n";
	for(int tmp=0; tmp<lineage_rates.size(); tmp++)
	{
		for(int tmp2=0; tmp2<4; tmp2++)
		{
			cout << lineage_rates.at(tmp) << " ";
		}
		cout << "\n";
	}
	cout << "\n\n";*/

	index_alpha = index_live.at(unindexed_alpha);
	if(DEBUG>=2)
	  cout << ", indexed = " << index_alpha << "\n";

	// Determine all of the events that can happen to this haplotype.
	// These record which exact events can actually can happen (eg which sites can mutate on alpha)
	int				no_pos_coals = 0; 
	int				no_pos_muts  = 0;
	vector<int>		pos_coals;	
	vector<int>		pos_muts;
	//vector<bool>		pos_recs;   //Why would I need this? All recombinations can happen within rec_left_edge and right_rec_edge
	pos_coals.resize(actv_lngs,-1); //0 says that alpha cannot coalesce with this lineage. A positive integer indicates with whom coalescence is possible.
	pos_muts.resize(no_segsts,-1);  //0 says that this site cannot mutate. a positive integer indicates with whom coalescence is possible.
	//pos_recs.resize(no_segsts-1);
	
	Enum_Events(pos_coals, pos_muts, unindexed_alpha, no_pos_coals, no_pos_muts);
	
	// These are NOT just importance weights, they are proportional the the chance of being chosen
	// Note that the size of lcoal_weights is smaller than the total number of lineages.
	// This is because a specific lineage gets chosen (and it cannot coalesce with itself)
	// However, this leads to careful bookkeeping so that, say we've chosen the 20th then the 20th coalescence event means 20 coalesces with the _21st_ haplotype
	vector<double> 	lcoal_weights;	lcoal_weights.resize(actv_lngs-1, 4E10); 
	vector<double> 	lmut_weights;  lmut_weights.resize(no_segsts, 4E10);
	vector<double> 	lrec_weights;	lrec_weights.resize(no_segsts-1, 4E10);

	//double threshold = 0.002;		
	bool rec=true; //std::numeric_limits<bool>::quiet_NaN();
	double coal_and_mut_rate; double random_number;		double prior_restricted_event_rate;
	
	//Old Version - produces too much recombination
	//coal_and_mut_rate = no_pos_coals/2.0 + (no_pos_muts * theta_per_site/2.0);
	
	if(no_pos_coals + no_pos_muts > 0)
		coal_and_mut_rate = coal_rate/2.0 + (mut_rate/2.0);
	else
		coal_and_mut_rate = no_pos_coals/2.0 + (no_pos_muts * theta_per_site/2.0);
	if(coal_and_mut_rate < 0 || coal_and_mut_rate > 1E10)
		getchar();

	if(con.greedy_mle)
		random_number = 0.5;
	else
		random_number = ran->U();
	Calc_index_live();					
	haplotype& fast_alpha = hap_list.at(index_alpha);
	//prior_restricted_event_rate = no_pos_coals + (no_pos_muts * theta_per_site) + (((fast_alpha.right_rec_edge-1) - fast_alpha.left_rec_edge) * rho_per_site);
	prior_restricted_event_rate =  (2.0*coal_and_mut_rate)+ (((fast_alpha.right_rec_edge) - fast_alpha.left_rec_edge) * rho_per_site);		
	fast_adjustment = coal_and_mut_rate/(prior_restricted_event_rate/2.0);
	// The below CHOICE must be recorded in the EVENT WEIGHT. Not sure how best to achieve this 
	// with the current frame work. Currently 'Pick Event' does all of this job. Introduce a flag? A bit nasty...
	//Note the !'s here! This check is to see if we can do a quick and dirty move here.
	use_dynamic_programming = !(coal_fast && !(con.rec_rate/con.evnt_rate > con.threshold));
	if(!use_dynamic_programming)
	{
		if(random_number < fast_adjustment)
			rec=false;
		else
		{
			rec=true;
			use_dynamic_programming=1;
		}

		if(rec==false)
		{
			for (i=0; i<no_segsts; i++)
			{
				if(pos_muts.at(i) > 0)
					lmut_weights.at(i) = log(con.theta/(no_segsts*2.0));
				else
					lmut_weights.at(i) = -1E308;
			}	
			int past_alpha = 0;
			for (i=0; i<actv_lngs; i++) // This is because actv_lngs has DECREASED since possible_coals was created and alpha needs to be skipped.
   			{
				if(i != unindexed_alpha)
   				{
   					if(pos_coals.at(i) > 0)
						lcoal_weights.at(i-past_alpha) = -log_2; //lcoal_weightings has no 'space for alpha' so we skip past that. Also, each coal event is counted twice.
  					else
   						lcoal_weights.at(i-past_alpha) = -1E308;
   				}
   				else
   				{
   					if(unindexed_alpha != actv_lngs-1)
						lcoal_weights.at(i) = -1E308; //This is about to get overwritten above because of 'past_alpha' !!
   					past_alpha = 1;
   				}
   			}	
			for (i=0; i<no_segsts-1; i++)
				lrec_weights.at(i) = -1E308;
			if(DEBUG>=2)
			{
				cout << "\nThe weightings (on " << unindexed_alpha << ") before normalisation are: \n";
				Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
			}
		}
		else
		{
			//else: rec==true Which means that a recombination event WILL be the next event (mut/coal set to zero)
			bool reverse = 0;
			
			 Remove_alpha_index(index_alpha);			
			/*live_lineages.at(index_alpha) = 0;		// Must fix all of this once the event is chosen.
			actv_lngs --; dead_lngs ++;	
			Calc_index_live();		// Careful of indices past this point!*/
		   	alpha = hap_list.at(index_alpha);	
			
			if(con.use_logs)
				Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
			else
				Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	
			vector<double>	lforwards_Prob_up_to		 = lProb_up_to; // This stores the L_and_S results    from the 'forward pass'
			//double			lforwards_output_likelihood	 = loutput_likelihood;  // This stores the L_and_S likelihood from the 'forward pass'

			reverse = 1;
			if(con.use_logs)
				Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
			else
				Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	
						const vector<double>& lbackwards_Prob_up_to = lProb_up_to; // This stores the L_and_S results	  from the 'backward pass'
			//double	lbackwards_output_likelihood = loutput_likelihood;							// This stores the L_and_S likelihood from the 'backward pass'
				
			for (i=0; i<no_segsts-1; i++)
			{
				//if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge)
				if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge-1) //Have got right_edge-1 as it seems consistent at the time, should check!
				{
					// Probably need to debug this with a super small data set to check that it's what we want.
					lrec_weights.at(i) = (lforwards_Prob_up_to.at(i) + lbackwards_Prob_up_to.at(i+1)) - lforwards_Prob_up_to.at(no_segsts-1);
				}
				else
					lrec_weights.at(i) = -1E308;
			}

			for(i=0; i<(int)lrec_weights.size(); i++)
			{
				if(site_posns.at(i+1) - site_posns.at(i) > 0)
				{
					if(fabs(rho)> 1e-20)
						lrec_weights.at(i)		+= log((rec_map.at(i+1) - rec_map.at(i))/2.0);
					else
						lrec_weights.at(i) = -1E308;
				}
				else
				{
					if(fabs(rho)> 1e-20)
						lrec_weights.at(i)		+= log((rho * 0.00005)/2.0);
					else
						lrec_weights.at(i) = -1E308;
				}
			}
			/*for (i=0; i<no_segsts; i++)
			{
				if(pos_muts.at(i) > 0)
					lmut_weights.at(i) = log(con.theta/(no_segsts*2.0));
				else
					lmut_weights.at(i) = -1E308;
			}*/	
			//When have chosen to recombine - then set probability of coalescing or mutation to log(0).
			for (i=0; i<(int)lcoal_weights.size(); i++)
   				lcoal_weights.at(i) = -1E308;
   			for (i=0; i<no_segsts; i++)
				lmut_weights.at(i) = -1E308;

			if(DEBUG>=2)
			{
				cout << "\nThe weightings before normalisation but including priors are: \n";
				Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
			}
		}
	}
	else
	{
		//haplotype alpha;
		bool reverse = 0;
		
        Remove_alpha_index(index_alpha);
		/*live_lineages.at(index_alpha) = 0;		// Must fix all of this once the event is chosen.
		actv_lngs --; dead_lngs ++;	
		Calc_index_live();							// Careful of indices past this point!*/

		//temp checking line.
				if(0) 
				{
					if(unindexed_alpha < (int)index_live.size())
						cout << "The index of alpha is:" << index_alpha << " but index_live.at(unindexed_alpha) is now: " << index_live.at(unindexed_alpha) << "\n";
					else
						cout << "Chose the last lineage so now index_live.at(unindexed_alpha) does not exist\n";
					cout << "\n";
				}
				
		haplotype beta; double llikelihood_alpha, llikelihood_beta;
   		alpha = hap_list.at(index_alpha);
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, llikelihood_alpha);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, llikelihood_alpha);	
		for (i=0; i<no_segsts; i++)
		{
			if(pos_muts.at(i) > 0)
			{
				Mut_create_alpha(alpha, beta, index_alpha, i); // Need to cheat here and use the lineage that just got killed
				if(con.use_logs)
					Look_Down_and_Copy(beta, reverse, lProb_up_to, llikelihood_beta);	
				else
					Look_Down_and_Copy_nolog(beta, reverse, lProb_up_to, llikelihood_beta);	
				lmut_weights.at(i) = llikelihood_beta - llikelihood_alpha;
			}
			else
				lmut_weights.at(i) = -1E308;
		}	

		haplotype aTTb, store_alpha;
		store_alpha = hap_list.at(index_alpha);			
		int past_alpha = 0;
   		for (i=0; i<actv_lngs + 1; i++) // This is because actv_lngs has DECREASED since possible_coals was created and alpha needs to be skipped.
   		{
			if(i != unindexed_alpha)
   			{
   				if(pos_coals.at(i) > 0)
				{
					if(cheap_coal_upwt)
						lcoal_weights.at(i-past_alpha) = -log_2;
					else
					{
   						//aUb.alleles.clear(); 
						aTTb.alleles.clear();
						//Coal_Create_aUb(hap_list.at(index_alpha), hap_list.at(index_live.at(i-past_alpha)), aUb); 
						Coal_Create_aTTb(hap_list.at(index_alpha), hap_list.at(index_live.at(i-past_alpha)), aTTb);		
						hap_list.at(index_alpha) = aTTb;
									
						//if(aTTb.left_rec_edge < aTTb.right_rec_edge+1)
						if(aTTb.left_rec_edge < aTTb.right_rec_edge)
						{
							if(con.use_logs)
								Look_Down_and_Copy(aTTb, reverse, lProb_up_to, loutput_likelihood);	
							else
								Look_Down_and_Copy_nolog(aTTb, reverse, lProb_up_to, loutput_likelihood);
						}
						else
							loutput_likelihood = 0;
						hap_list.at(index_alpha) = store_alpha;
						lcoal_weights.at(i-past_alpha) = -loutput_likelihood - log_2; //lcoal_weightings has no 'space for alpha' so we skip past that. Also, each coal event is counted twice.
					}
				}
   				else
   					lcoal_weights.at(i-past_alpha) = -1E308;
   			}
   			else
   			{
   				if(unindexed_alpha != actv_lngs)
					lcoal_weights.at(i) = -1E308;
   				past_alpha = 1;
   			}
   		}	

		alpha = hap_list.at(index_alpha);					
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	
		vector<double>	lforwards_Prob_up_to		 = lProb_up_to; // This stores the L_and_S results    from the 'forward pass'
		//double			lforwards_output_likelihood	 = loutput_likelihood;  // This stores the L_and_S likelihood from the 'forward pass'

		reverse = 1;
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	
		const vector<double>& lbackwards_Prob_up_to = lProb_up_to; // This stores the L_and_S results	  from the 'backward pass'
		//double	lbackwards_output_likelihood = loutput_likelihood;							// This stores the L_and_S likelihood from the 'backward pass'
			
		for (i=0; i<no_segsts-1; i++)
		{
			//if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge)
			if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge-1) //Have got right_edge-1 as it seems consistent at the time, shoudl check!
			{
				// Probably need to debug this with a super small data set to check that it's what we want.
				lrec_weights.at(i) = (lforwards_Prob_up_to.at(i) + lbackwards_Prob_up_to.at(i+1)) - lforwards_Prob_up_to.at(no_segsts-1);
			}
			else
				lrec_weights.at(i) = -1E308;
		}

		//double theta_per_site	= con.theta/(double)no_segsts;
		for(i=0; i<(int)lmut_weights.size(); i++)
		{
			if(con.theta !=0)
				lmut_weights.at(i)		+= log(theta_per_site/2.0);
			else
				lmut_weights.at(i)		= -1E308;
		}
		for(i=0; i<(int)lrec_weights.size(); i++)
		{
			if(site_posns.at(i+1) - site_posns.at(i) > 0)
			{
				if(fabs(rho)> 1e-20)
					lrec_weights.at(i)		+= log((rec_map.at(i+1) - rec_map.at(i))/2.0);
				else
					lrec_weights.at(i) = -1E308;
			}
			else
			{
				if(fabs(rho)> 1e-20)
					lrec_weights.at(i)		+= log((rho * 0.00005)/2.0);
				else
					lrec_weights.at(i) = -1E308;
			}
		}			
		if(DEBUG>=2)
		{
			cout << "\nThe weightings before normalisation but including priors are: \n";
			Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
		}
	}
	// This function also factors in the prior rate of each event
	Normalise_Weightings(lcoal_weights, lmut_weights, lrec_weights);
	if(DEBUG>=2)
	{	
		cout << "The weightings on " << unindexed_alpha << "(" << index_alpha << ") are: \n";
		Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
	}

	//use_dynamic_programming = (calc_event_weights_properly || rec==true);
	//if((!con.coal_fast || (random_number > fast_adjustment) || (con.rec_rate/con.evnt_rate > con.threshold) || (no_pos_muts+no_pos_coals==0))!=use_dynamic_programming)
	//{
	//	cout << "Warning! bookkeeping of whether or not have used DP has gone wrong in LDaC function with coal_fast and hap_1st.\n";
	//	getchar();
	//}
	if(use_dynamic_programming)
	{
		Restore_alpha_index(index_alpha);
		/*live_lineages.at(index_alpha) = 1;			// Live lineages has the same size and indexing as hap_list.
		actv_lngs ++; dead_lngs --;	
		Calc_index_live();	*/
	}
	
	int unin_daught2;
	Pick_Event(event, unindexed_alpha, unin_daught2, daughter2, rec_site, mut_site, new_type, lcoal_weights, lmut_weights, lrec_weights, pos_muts, levent_weight, 0.0);// stochastically choose an event from all of those exact events
	
	if(event == 1)
	{
		double daughter2_rate = lineage_rates.at(unin_daught2);
		double lsymmetric_likelihood;
		int passed_daught2 = 0; 
		if(unin_daught2 < unindexed_alpha) 
			passed_daught2 = 1;
		if(coal_fast && (rec_rate/evnt_rate < threshold))
		{
			levent_weight += log_2; //This is log(2.0) but faster
		}
		else
		{
			Back_track(unindexed_alpha, unin_daught2, daughter2_rate, lsymmetric_likelihood, passed_daught2);
			levent_weight = Add_Log((lsymmetric_likelihood + log(daughter2_rate)), levent_weight);
		}
	}
	
	if(event == 1)
	{
		daughter1	= index_alpha;				// So this is for hap_list, cannot be used is smaller vectors.
		rec_lineage = -1;
		mut_lineage = -1;
		
		//The following two lines should be double checked, the reason is that the indexing is subtle.
		//if(unind_daughter2<unindexed_alpha) levent_weight += lcoal_weights.at(unind_daughter2); //Note: the probability of choosing THIS ALPHA has already been factored in!!
		//else levent_weight								+= lcoal_weights.at(unind_daughter2 - 1);
	}
	if(event == 2)
	{
		mut_lineage = index_alpha;
		daughter1   = -1;
		rec_lineage = -1;
	}
	if(event == 3)
	{
		rec_lineage = index_alpha;
		daughter1   = -1;
		mut_lineage = -1;
	}

	if(coal_fast  && (rec_rate/evnt_rate < threshold))
	{
		if(rec)
			levent_weight += log(1-fast_adjustment);
		else
			levent_weight += log(fast_adjustment);
	}
	if(levent_weight > (1.0E-9))
	{
		cout << "Warning - suspiciously high event weight: " << exp(levent_weight) << ".\n"; 
		getchar();
	}

	return event;
}



















int CLookD_Copy::Choose_from_all_events(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type, double& levent_weight)
{
	double log_2 = 0.69314718055994530941;
	
	int event = -1; //1 = coal, 2 = mut, 3 = rec
	quick_Li_and_Stephens = 1;
	int index_alpha;
	int unindexed_alpha;
	vector<double>	lineage_rates;
	// These record the prior rate of each type of event on each lineage. (eg. 4 if 4 sites are ancestral, not 4*con.theta)
	vector<int> pr_coal_rates;
	vector<int> pr_mut_rates;
	vector<int> pr_rec_rates;
	
	lineage_rates.resize(actv_lngs);
	
	// Stochastically choose a haplotype, alpha, to act on according to the rate at which events happen to it.
    Calc_Prior_Rates(lineage_rates, pr_coal_rates, pr_mut_rates, pr_rec_rates);
    //These are NOT just importance weights, they are proportional the the chance of bein chosen
   	vector<double> 	lcoal_weights;	lcoal_weights.resize(actv_lngs-1,4E10); 
   	vector<double> 	lmut_weights;  lmut_weights.resize(no_segsts, 4E10);
   	vector<double> 	lrec_weights;	lrec_weights.resize(no_segsts-1, 4E10);
   	vector<int>		pos_coals;	
   	vector<int>		pos_muts;
   	
   
	int i;    
	vector<vector<int> > matrix_lmut_types;    matrix_lmut_types.resize(actv_lngs);
	vector<vector<double> > matrix_lcoal_weights; matrix_lcoal_weights.resize(actv_lngs);
	//for(i=0; i<actv_lngs; i++) matrix_lcoal_weights.at(i).resize(actv_lngs);
	vector<vector<double> > matrix_lmut_weights;  matrix_lmut_weights.resize(actv_lngs);
	//for(i=0; i<actv_lngs; i++) matrix_lmut_weights.at(i).resize(no_segsts);
	vector<vector<double> > matrix_lrec_weights;  matrix_lrec_weights.resize(actv_lngs);
	//for(i=0; i<actv_lngs; i++) matrix_lrec_weights.at(i).resize(no_segsts-1);	
	
	for(unindexed_alpha =0;unindexed_alpha<actv_lngs; unindexed_alpha ++)
	{
   		//These will be overwritten naturally in the program. Not sure if they should be set to -9 or something here though.
		//lcoal_weights.clear(); lmut_weights.clear(); lrec_weights.clear();
   		pos_coals.clear(); pos_muts.clear();
		  index_alpha = index_live.at(unindexed_alpha);
   
   		// Determine all of the events that can happen to this haplotype.
   		// These record which exact events can actually can happen (eg which sites can mutate on alpha)
   		int				no_pos_coals = 0; 
   		int				no_pos_muts  = 0;
   		//vector<bool>		pos_recs;   //Why would I need this? All recombinations can happen within rec_left_edge and right_rec_edge
   		pos_coals.resize(actv_lngs,-1); //0 says that alpha cannot coalesce with this lineage. A positive integer indicates with whom coalescence is possible.
   		pos_muts.resize(no_segsts,-1);  //0 says that this site cannot mutate. a positive integer indicates with whom coalescence is possible.
   		//pos_recs.resize(no_segsts-1);
   		
   		Enum_Events(pos_coals, pos_muts, unindexed_alpha, no_pos_coals, no_pos_muts);
   		
   		haplotype alpha;
   		bool reverse = 0;
   		
   		Remove_alpha_index(index_alpha);			
		/*live_lineages.at(index_alpha) = 0;		// Must fix all of this once the event is chosen.
   		actv_lngs --; dead_lngs ++;	
   		Calc_index_live();							// Careful of indices past this point!*/
   				//temp checking line.
   				if(0) 
   				{
   					if(unindexed_alpha < (int)index_live.size())
   						cout << "The index of alpha is:" << index_alpha << " but index_live.at(unindexed_alpha) is now: " << index_live.at(unindexed_alpha) << "\n";
   					else
   						cout << "Chose the last lineage so now index_live.at(unindexed_alpha) does not exist\n";					
   					cout << "\n";
   				}
		haplotype beta; double llikelihood_alpha, llikelihood_beta;
		alpha = hap_list.at(index_alpha);
		//cout << "Alpha = "; Display_a_hap(alpha); cout << "\n";			
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, llikelihood_alpha);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, llikelihood_alpha);	
		for (i=0; i<no_segsts; i++)
		{
		if(pos_muts.at(i) > 0)
		{
			Mut_create_alpha(alpha, beta, index_alpha, i); // Need to cheat here and use the lineage that just got killed
			//cout << "Beta  = "; Display_a_hap(beta); cout << "\n";
			if(con.use_logs)
				Look_Down_and_Copy(beta, reverse, lProb_up_to, llikelihood_beta);	
			else
				Look_Down_and_Copy_nolog(beta, reverse, lProb_up_to, llikelihood_beta);	
			lmut_weights.at(i) = llikelihood_beta - llikelihood_alpha;
		}
		else
   			lmut_weights.at(i) = -1E308;
		}	

		//haplotype aUb;
		haplotype aTTb, store_alpha;		
	//	cout << "Alpha = "; Display_a_hap(hap_list.at(index_alpha)); cout << "\n";
		store_alpha = hap_list.at(index_alpha);			
		int past_alpha = 0;
   		for (i=0; i<actv_lngs + 1; i++) // This is because actv_lngs has DECREASED since possible_coals was created and alpha needs to be skipped.
   		{
			if(i != unindexed_alpha)
   			{
				if(pos_coals.at(i) > 0)   													
				{
   					if(cheap_coal_upwt)
						lcoal_weights.at(i-past_alpha) = -log_2;
					else
					{
						//aUb.alleles.clear(); 
						aTTb.alleles.clear();
						//Coal_Create_aUb(hap_list.at(index_alpha), hap_list.at(index_live.at(i-past_alpha)), aUb); 
						Coal_Create_aTTb(hap_list.at(index_alpha), hap_list.at(index_live.at(i-past_alpha)), aTTb);		
						hap_list.at(index_alpha) = aTTb;
									
						//if(aTTb.left_rec_edge < aTTb.right_rec_edge+1)
						if(aTTb.left_rec_edge < aTTb.right_rec_edge)
						{
							if(con.use_logs)
								Look_Down_and_Copy(aTTb, reverse, lProb_up_to, loutput_likelihood);	
							else
								Look_Down_and_Copy_nolog(aTTb, reverse, lProb_up_to, loutput_likelihood);	
						}
						else 
							loutput_likelihood = 0;
						hap_list.at(index_alpha) = store_alpha;
   						lcoal_weights.at(i-past_alpha) = -loutput_likelihood -log_2; //lcoal_weightings has no 'space for alpha' so we skip past that. Also, each coal event is counted twice.
					}
				}
   				else
   					lcoal_weights.at(i-past_alpha) = -1E308;
   			}
			else
   			{
   				if(unindexed_alpha != actv_lngs)
					lcoal_weights.at(i) = -1E308;
   				past_alpha = 1;
   			}
   		}	
		//hap_list.at(index_alpha) = store_alpha;
		//Display_a_hap(hap_list.at(index_alpha));
   
   		alpha = hap_list.at(index_alpha);			
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	

		vector<double>	lforwards_Prob_up_to		 = lProb_up_to; // This stores the L_and_S results    from the 'forward pass'
   		//double			lforwards_output_likelihood	 = loutput_likelihood;  // This stores the L_and_S likelihood from the 'forward pass'
   		
   		reverse = 1;
   	
		if(con.use_logs)
			Look_Down_and_Copy(alpha, reverse, lProb_up_to, loutput_likelihood);	
		else
			Look_Down_and_Copy_nolog(alpha, reverse, lProb_up_to, loutput_likelihood);	

   		const vector<double>& lbackwards_Prob_up_to = lProb_up_to; // This stores the L_and_S results	  from the 'backward pass'
		//double	lbackwards_output_likelihood = loutput_likelihood;							// This stores the L_and_S likelihood from the 'backward pass'
   			
   		for (i=0; i<no_segsts-1; i++)
   		{
   			//if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge)
   			if(i>=alpha.left_rec_edge && i<=alpha.right_rec_edge-1) //Have got right_edge-1 as it seems consistent at the time, shoudl check!
   			{
   				// Probably need to debug this with a super small data set to check that it's what we want.
   				lrec_weights.at(i) = (lforwards_Prob_up_to.at(i) + lbackwards_Prob_up_to.at(i+1)) - lforwards_Prob_up_to.at(no_segsts-1);
   			}
   			else
   				lrec_weights.at(i) = -1E308;
   		}
   			
   		if(DEBUG>=2)
   		{
   			cout << "The coalesecence weightings on " << unindexed_alpha << " before normalisation or PRIORS are: \n";
   			for(i=0; i<actv_lngs; i++)
   			{
   				cout << exp(lcoal_weights.at(i)) << " ";
   			}
   			cout << "\nThe Mutation weightings on " << unindexed_alpha << " before normalisation or inclusion of PRIORS are: \n";
   			for(i=0; i<no_segsts; i++)
   			{
   				cout << exp(lmut_weights.at(i)) << " ";
   			}
   			cout << "\nThe Recombination weightings on " << unindexed_alpha << " before normalisation or PRIORS are: \n";
   			for(i=0; i<no_segsts-1; i++)
   			{
   				cout << exp(lrec_weights.at(i)) << " ";
   			}
   			cout << "\n";
   		}
   		matrix_lmut_types.at(unindexed_alpha)    = pos_muts;
   		matrix_lcoal_weights.at(unindexed_alpha) = lcoal_weights;
   		matrix_lmut_weights.at(unindexed_alpha)  = lmut_weights;
   		matrix_lrec_weights.at(unindexed_alpha)  = lrec_weights;
		Restore_alpha_index(index_alpha);
		/*live_lineages.at(index_alpha) = 1;			// Live lineages has the same size and INDEXING as hap_list.
		actv_lngs ++; dead_lngs --;	
		Calc_index_live();							// Careful of indices past this point!		*/
	}

   	// This function also factors in the prior rate of each event
   	vector<double> lhaplo_weights; lhaplo_weights.resize(actv_lngs,-1);
	Normalise_All_Weightings(matrix_lcoal_weights, matrix_lmut_weights, matrix_lrec_weights, lhaplo_weights);

	if(DEBUG>=2)
	{
		for(unindexed_alpha = 0; unindexed_alpha < actv_lngs; unindexed_alpha++)
		{
			cout << "haplo_weights.at(" << unindexed_alpha << ") = " << exp(lhaplo_weights.at(unindexed_alpha)) << "\n";
		}
	}
  if (DEBUG>3)
	{
    int passed_unindexed_alpha;
		for(unindexed_alpha = 0; unindexed_alpha < actv_lngs; unindexed_alpha++)
		{
			cout << "\nThe full coalesecence weightings on " << unindexed_alpha << " are: \n";
   			
			passed_unindexed_alpha = 0;
			for(i=0; i<actv_lngs; i++)
   			{
   				if (i!= unindexed_alpha)
					cout << exp(matrix_lcoal_weights.at(unindexed_alpha).at(i-passed_unindexed_alpha)) << " ";
				else
					passed_unindexed_alpha = 1;
   			}
   			cout << "\nThe full Mutation weightings on " << unindexed_alpha << " are: \n";
   			for(i=0; i<no_segsts; i++)
   			{
   				cout << exp(matrix_lmut_weights.at(unindexed_alpha).at(i)) << " ";
   			}
   			cout << "\nThe full Recombination weightings on " << unindexed_alpha << " are: \n";
   			for(i=0; i<no_segsts-1; i++)
   			{
   				cout << exp(matrix_lrec_weights.at(unindexed_alpha).at(i))  << " ";
   			}
		}
		cout << "\n\n";	
   	}	
	
	unindexed_alpha = -1;
	// This function chooses using all of the information, not just the prior rates.
	Choose_alpha_properly(unindexed_alpha, matrix_lcoal_weights, matrix_lmut_weights, matrix_lrec_weights, lhaplo_weights, levent_weight);
	index_alpha = index_live.at(unindexed_alpha);
	// The following line destroys the relative weighting of the haplotypes (no longer needed) so that Pick_Event is dealing with a haplotype of total weight 1.
	// The weight of this haplotype is added to 'levent_weight' in the choose_alpha properly code.
	Normalise_Weightings(matrix_lcoal_weights.at(unindexed_alpha), matrix_lmut_weights.at(unindexed_alpha), matrix_lrec_weights.at(unindexed_alpha));
	if (DEBUG>=2)
	{	
		cout << "\nThe finally normalised weightings on " << unindexed_alpha << " are: \n";
		Display_Complex_Weights(unindexed_alpha, matrix_lcoal_weights, matrix_lmut_weights, matrix_lrec_weights);
	}

	//Pick_Event(event, unindexed_alpha, daughter2, rec_site, mut_site, new_type, matrix_lcoal_weights.at(unindexed_alpha), matrix_lmut_weights.at(unindexed_alpha), matrix_lrec_weights.at(unindexed_alpha), matrix_lmut_types.at(unindexed_alpha), levent_weight, lhaplo_weights.at(unindexed_alpha));// stochastically choose an event from all of those exact events
	int unin_daught2;

	Pick_Event(event, unindexed_alpha, unin_daught2, daughter2, rec_site, mut_site, new_type, matrix_lcoal_weights.at(unindexed_alpha), matrix_lmut_weights.at(unindexed_alpha), matrix_lrec_weights.at(unindexed_alpha), matrix_lmut_types.at(unindexed_alpha), levent_weight, 0.0);// stochastically choose an event from all of those exact events	
	if(event == 1)
	{
		daughter1		=  index_alpha;				// So this is for hap_list, cannot be used is smaller vectors.
		rec_lineage		=  -1;
		mut_lineage		=  -1;
		levent_weight	+= log_2;
	}
	if(event == 2)
	{
		mut_lineage = index_alpha;
		daughter1   = -1;
		rec_lineage = -1;
	}
	if(event == 3)
	{
		rec_lineage = index_alpha;
		daughter1   = -1;
		mut_lineage = -1;
	}
	//lProb_up_to.~vector();
	return event;
 }
 




















void CLookD_Copy::Back_track(const int& unindexed_alpha, const int& unindexed_daught2, const double& lineage_rate, double& lsymmetric_likelihood, const int& passed_daught2)
{
	double log_2 = 0.69314718055994530941;
	
	int index_daughter2;
	index_daughter2	= index_live.at(unindexed_daught2);

	int				no_pos_coals = 0; 
	int				no_pos_muts  = 0;
	vector<int>		pos_coals;	
	vector<int>		pos_muts;
	pos_coals.resize(actv_lngs,-1); //0 says that alpha cannot coalesce with this lineage. A positive integer indicates with whom coalescence is possible.
	pos_muts.resize(no_segsts,-1);  //0 says that this site cannot mutate. a positive integer indicates with whom coalescence is possible.
	
	Enum_Events(pos_coals, pos_muts, unindexed_daught2, no_pos_coals, no_pos_muts);
	
	vector<double> 	lcoal_weights;	lcoal_weights.resize(actv_lngs-1, 4E10); 
	vector<double> 	lmut_weights;  lmut_weights.resize(no_segsts, 4E10);
	vector<double> 	lrec_weights;	lrec_weights.resize(no_segsts-1, 4E10);
	haplotype daughter2;
	bool reverse = 0;
	
    Remove_alpha_index(index_daughter2);		
	/*live_lineages.at(index_daughter2) = 0;		// Must fix all of this once the event is chosen.
	actv_lngs --; dead_lngs ++;	
	Calc_index_live();							// Careful of indices past this point!*/
			
	int i; haplotype beta; double llikelihood_daught2, llikelihood_beta;
	daughter2 = hap_list.at(index_daughter2);	
	if(con.use_logs)
		Look_Down_and_Copy(daughter2, reverse, lProb_up_to, llikelihood_daught2);	
	else
		Look_Down_and_Copy_nolog(daughter2, reverse, lProb_up_to, llikelihood_daught2);	

	for (i=0; i<no_segsts; i++)
	{
		if(pos_muts.at(i) > 0)
		{
			Mut_create_alpha(daughter2, beta, index_daughter2, i); // Need to cheat here and use the lineage that just got killed
			if(con.use_logs)
				Look_Down_and_Copy(beta, reverse, lProb_up_to, llikelihood_beta);	
			else
				Look_Down_and_Copy_nolog(beta, reverse, lProb_up_to, llikelihood_beta);	
			lmut_weights.at(i) = llikelihood_beta - llikelihood_daught2;
		}
		else
			lmut_weights.at(i) = -1E308;
	}	

	haplotype aTTb, store_daught2;
	store_daught2 = hap_list.at(index_daughter2);			
	int past_daught2 = 0;
   	for (i=0; i<actv_lngs + 1; i++) // This is because actv_lngs has DECREASED since possible_coals was created AND alpha needs to be skipped.
   	{
		if(i != unindexed_daught2)
   		{
   			if(pos_coals.at(i) > 0)
			{
   				if(cheap_coal_upwt)
					lcoal_weights.at(i-past_daught2) = -log_2;
				else
				{
					//aUb.alleles.clear(); 
					aTTb.alleles.clear();
					//Coal_Create_aUb(hap_list.at(index_daughter2), hap_list.at(index_live.at(i-past_daught2)), aUb); 
					Coal_Create_aTTb(hap_list.at(index_daughter2), hap_list.at(index_live.at(i-past_daught2)), aTTb);		
					hap_list.at(index_daughter2) = aTTb;
					
					//It looks like the commented version is correct but I haven't had time to debug it
					//if(aTTb.left_rec_edge < aTTb.right_rec_edge+1)
					if(aTTb.left_rec_edge < aTTb.right_rec_edge)
					{
						if(con.use_logs)
							Look_Down_and_Copy(aTTb, reverse, lProb_up_to, loutput_likelihood);	
						else
							Look_Down_and_Copy_nolog(aTTb, reverse, lProb_up_to, loutput_likelihood);	
					}
					else
						loutput_likelihood = 0;
					hap_list.at(index_daughter2) = store_daught2;   			   				
					lcoal_weights.at(i-past_daught2) = -loutput_likelihood - log_2; //lcoal_weightings has no 'space for alpha' so we skip past that. Also, each coal event is counted twice.
				}
			}
   			else
   				lcoal_weights.at(i-past_daught2) = -1E308;
   		}
   		else
   		{
   			if(unindexed_daught2 != actv_lngs)
				lcoal_weights.at(i) = -1E308;
   			past_daught2 = 1;
   		}
   	}	

	daughter2 = hap_list.at(index_daughter2);						
	if(con.use_logs)
		Look_Down_and_Copy(daughter2, reverse, lProb_up_to, loutput_likelihood);	
	else
		Look_Down_and_Copy_nolog(daughter2, reverse, lProb_up_to, loutput_likelihood);	
	vector<double>	lforwards_Prob_up_to		 = lProb_up_to; // This stores the L_and_S results    from the 'forward pass'
	//double			lforwards_output_likelihood	 = loutput_likelihood;  // This stores the L_and_S likelihood from the 'forward pass'

	reverse = 1;
	if(con.use_logs)
		Look_Down_and_Copy(daughter2, reverse, lProb_up_to, loutput_likelihood);	
	else
		Look_Down_and_Copy_nolog(daughter2, reverse, lProb_up_to, loutput_likelihood);	
	const vector<double>& lbackwards_Prob_up_to = lProb_up_to; // This stores the L_and_S results	  from the 'backward pass'
	//double	lbackwards_output_likelihood = loutput_likelihood;							// This stores the L_and_S likelihood from the 'backward pass'
		
	for (i=0; i<no_segsts-1; i++)
	{
		if(i>=daughter2.left_rec_edge && i<=daughter2.right_rec_edge-1)
			lrec_weights.at(i) = (lforwards_Prob_up_to.at(i) + lbackwards_Prob_up_to.at(i+1)) - lforwards_Prob_up_to.at(no_segsts-1);
		else
			lrec_weights.at(i) = -1E308;
	}

	double theta_per_site	= con.theta/(double)no_segsts;
	for(i=0; i<(int)lmut_weights.size(); i++)
	{
		if(con.theta !=0)
			lmut_weights.at(i)		+= log(theta_per_site/2.0);
		else
			lmut_weights.at(i)		= -1E308;
	}
	for(i=0; i<(int)lrec_weights.size(); i++)
	{
		if(site_posns.at(i+1) - site_posns.at(i) > 0)
		{
			if(fabs(rho)> 1e-20)
				lrec_weights.at(i)		+= log((rec_map.at(i+1) - rec_map.at(i))/2.0);
			else
				lrec_weights.at(i) = -1E308;
		}
		else
		{
			if(fabs(rho)> 1e-20)
				lrec_weights.at(i)		+= log((rho * 0.00005)/2.0);
			else
				lrec_weights.at(i) = -1E308;
		}
	}
		
	if(DEBUG>3)
	{
		cout << "\nThe weightings before normalisation but including priors are: \n";
		Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
	}

	// This function also factors in the prior rate of each event
	Normalise_Weightings(lcoal_weights, lmut_weights, lrec_weights);
	if(DEBUG>=2)
	{
		cout << "The Back-tracking weightings are: \n";
		Display_Weights(lcoal_weights, lmut_weights, lrec_weights);
	}

	Restore_alpha_index(index_daughter2);
	/*live_lineages.at(index_daughter2) = 1;			// Live lineages has the same size and indexing as hap_list.
	actv_lngs ++; dead_lngs --;	
	Calc_index_live();	*/
	lsymmetric_likelihood = lcoal_weights.at(unindexed_alpha - passed_daught2);
}











void CLookD_Copy::Coal_Create_aUb(const haplotype& alpha, const haplotype& beta, haplotype& aUb)
{
	int i;
	aUb.alleles.resize(alpha.alleles.size(), -1);
	for(i=0; i<(int)alpha.alleles.size(); i++)
	{
		if(alpha.alleles.at(i)>0 || beta.alleles.at(i)>0)
		{
			if(alpha.alleles.at(i))
				aUb.alleles.at(i) = alpha.alleles.at(i);
			if(beta.alleles.at(i))
				aUb.alleles.at(i) = beta.alleles.at(i);
		}
		else
			aUb.alleles.at(i) = 0;		
	}
	if(alpha.left_rec_edge < beta.left_rec_edge) // aUb is the union, so only care about sites where either are ancestral. So take left_most left edge.
		aUb.left_rec_edge = alpha.left_rec_edge;
	else
		aUb.left_rec_edge = beta.left_rec_edge;

	if(alpha.right_rec_edge > beta.right_rec_edge)
		aUb.right_rec_edge = alpha.right_rec_edge;
	else
		aUb.right_rec_edge = beta.right_rec_edge;
}
void CLookD_Copy::Coal_Create_aTTb(const haplotype& alpha, const haplotype& beta, haplotype& aTTb)
{
	int i;
	aTTb.alleles.resize(alpha.alleles.size(), -1);
	//cout << "alpha in aTTb = "; Display_a_hap(alpha);
	for(i=0; i<(int)alpha.alleles.size(); i++)
	{
		if(alpha.alleles.at(i)>0 && beta.alleles.at(i)>0)
			aTTb.alleles.at(i) = alpha.alleles.at(i);
		else
			aTTb.alleles.at(i) = 0;
	}
	if(alpha.left_rec_edge > beta.left_rec_edge) // aTTb is the intersection, so only care about sites where they are BOTH ancestral. So take right_most left edge.
		aTTb.left_rec_edge = alpha.left_rec_edge;
	else
		aTTb.left_rec_edge = beta.left_rec_edge;

	if(alpha.right_rec_edge < beta.right_rec_edge)
		aTTb.right_rec_edge = alpha.right_rec_edge;
	else
		aTTb.right_rec_edge = beta.right_rec_edge;
}


/*For possible problems with this method see end*/
void CLookD_Copy::calc_upweightings(int& daughter1, int& daughter2, int& rec_lineage, int& rec_site, int& mut_lineage, int& mut_site, int& new_type)
{
	cout << "You have called the Look down and Copy class version of this function. Only the Simple_IS class version should be called.\n";
	getchar();
}





// This scheme should, given a set of sequences, S, and another seperate sequence, X,  calculate the L&S P(X | S).
// The Li&Stephens scheme is effectively like the 'weather' example Gil described. Site by site.
void CLookD_Copy::Look_Down_and_Copy(const haplotype& alpha, const bool& reverse, vector<double>& lProb_up_to, double& loutput_likelihood)
{
	cout << "Error, the base version of Look Down and Copy has been called when only the specific dervied types should be. \n";
	getchar();
}

// This scheme should, given a set of sequences, S, and another seperate sequence, X,  calculate the L&S P(X | S).
// The Li&Stephens scheme is effectively like the 'weather' example Gil described. Site by site.
void CLookD_Copy::Look_Down_and_Copy_nolog(const haplotype& alpha, const bool& reverse, vector<double>& lProb_up_to, double& loutput_likelihood)
{
	cout << "Error, the base version of Look Down and Copy has been called when only the specific dervied types should be. \n";
	getchar();
}





void CLookD_Copy::Display_Weights(const vector<double>& lcoal_weights, const vector<double>& lmut_weights, const vector<double>& lrec_weights)
{
	int i;
	cout << "\nThe coalesecence weightings now are: \n";
	for(i=0; i<(int)lcoal_weights.size(); i++)
	{
		cout << lcoal_weights.at(i) << " ";
		if((int)lcoal_weights.size()!=actv_lngs) cout << "\n";
	}
	cout << "\nThe Mutation weightings now are: \n";
	for(i=0; i<(int)lmut_weights.size(); i++)
	{
		cout << lmut_weights.at(i) << " ";
		if((int)lmut_weights.size()!=no_segsts) cout << "Wrong mutation vector size?\n";
	}
	cout << "\nThe Recombination weightings now are: \n";
	for(i=0; i<(int)lrec_weights.size(); i++)
	{
		cout << lrec_weights.at(i) << " ";
		if((int)lrec_weights.size()!=(no_segsts-1)) cout << "Wrong recombination vector size?\n";
	}
	cout << "\n\n";
}   	

void CLookD_Copy::Display_Complex_Weights(const int& unindexed_alpha, const vector<vector<double> >& matrix_lcoal_weights, const vector<vector<double> >& matrix_lmut_weights, const vector<vector<double> >& matrix_lrec_weights)
{
	int i;
	int passed_unindexed_alpha;
	cout << "\nThe finally normalised coalesecence weightings on " << unindexed_alpha << " are: \n";
		
	passed_unindexed_alpha = 0;						
	for(i=0; i<actv_lngs; i++)
	{
   		if (i!= unindexed_alpha)
			cout << exp(matrix_lcoal_weights.at(unindexed_alpha).at(i-passed_unindexed_alpha)) << " ";
		else
			passed_unindexed_alpha = 1;
	}
	cout << "\nThe finally normalised Mutation weightings on " << unindexed_alpha << " are: \n";
	for(i=0; i<no_segsts; i++)
		cout << exp(matrix_lmut_weights.at(unindexed_alpha).at(i)) << " ";
	/*cout << "\nThe mutation types on " << unindexed_alpha << " are: \n";   		
	for(i=0; i<no_segsts; i++)
	{
		cout << matrix_lmut_types.at(unindexed_alpha).at(i) << " ";
	}*/
	cout << "\nThe finally normalised Recombination weightings on " << unindexed_alpha << " are: \n";
	for(i=0; i<no_segsts-1; i++)
		cout << exp(matrix_lrec_weights.at(unindexed_alpha).at(i))  << " ";
	cout << "\n\n";
}



void CLookD_Copy::Remove_alpha_index(const int& index_alpha)
{
	live_lineages.at(index_alpha) = 0;		// Must fix all of this once the event is chosen.
	actv_lngs --; dead_lngs ++;	
	Calc_index_live();		// Careful of indices past this point!
}

void CLookD_Copy::Restore_alpha_index(const int& index_alpha)
{
	live_lineages.at(index_alpha) = 1;		// Must fix all of this once the event is chosen.
	actv_lngs ++; dead_lngs --;	
	Calc_index_live();		// Careful of indices past this point!
}
