/*
 * argumentsOfProgram.h
 *
 *  Created on: Jun 16, 2010
 *      Author: niall
 */

#ifndef ARGUMENTSOFPROGRAM_H_
#define ARGUMENTSOFPROGRAM_H_

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <limits>
#include <string>
#include <vector>
#include <set>

#include "hapmapParams.h"
//#include "random.h"

class argumentsOfProgram
{
public:
  //argumentsOfProgram(Random* Inst_of_random);
  argumentsOfProgram();
  ~argumentsOfProgram() {
  }
  ChapmapParams  hapmapParams ; // Stores parameters specific to reading in the hapmap data (filenames etc.)
  double theta            ; // mutation rate
  double rho              ; // recombination rate
  bool   SMC              ; // should the program use the SMC model (if not, then the full coalescent)
  bool   importance_sample; // should the program perform importance sampling? Set to 1 even for intelligent tree simulation.
  bool   sim_data         ; // should the program simulate data (probably not implemented yet)?
  bool   using_rec_map    ; // should the program use an explicit rec_map? Set to FALSE as this is done a different way
  bool   all_nodes        ; // should the program output the leaf nodes in the genealogies (redundant info, but maybe convenient)
  double DEBUG            ; // affects behaviour for minor exceptions and control level of output to the console
  double verbose          ; // Controls level of output to the console - though very new and minimally implemented
  int    sampler          ; // what approximation to P(H1 | H2, ..., Hn) be used. 2 is Li & Stephens. Only 4 is also implemented. (set it to 1)
  int    no_runs          ; // how many genealogies to simulate per location?
  int    change_rho       ; // Simulate genealogies for multiple value of rho (useful for likelihood calculations) - if >0 then value is the number of different values tried
  int    bridge_resolution; // used in likelihood calculation. Sets density of bridge sampling based likelihood values (just: how many do you want?)
  int    change_theta     ; // Simulate genealogies for multiple value of theta (useful for likelihood calculations) - if >0 then value is the number of diifferent values tried
  bool   greedy_mle       ; // This tries to find a single 'best' ARG. When true program is NOT stochastic, should only have use run.
  bool   choose_hap_first ; // Chooses a faster method of generating genealogies - next even hap chosen from prior rates. See FearnHead & Donnelly (2000) or the thesis of cardin (2006)
  bool   coal_fast        ; // A 'fast' method of generating genealogies, not great for likelihood calculation
  double threshold        ; // (0 < Threshold < 1, sets how approximate the genealogies are with coal fast)
  bool   choose_nonrec    ; // Supposed to be a very fast method of generating genealogies - but makes little difference)
  bool   first_nonrec     ; // A very fast method of generating genealogies)
  bool   use_logs         ; // Should calculations within the Li&Stephens function use logs (faster if they don't, but underflow problems)?
  bool   cheap_coal_upwt  ; // always set to false
  bool   display_events   ; // This displays every event in each genealogy when No Runs is small.
  bool   display_haps     ; // This displays the haplotypes in each epoch
  bool   display_weights  ; // This displays the weights for each event in each epoch
  int    set_seed         ; // Put in a positive integer, n, here and the seed will be set to -n (else uses system clock to set seed)
  bool   using_wattersons ; // Should the Watterson estimate of 'theta' be used? If not then theta defined above is used
  bool   write_likelihoods; // should the average likelihoods should be written to file
  bool   write_trees      ; // should the trees should be written to file
  bool   write_full_data  ; // should every likelihood should be written to file
  string writeType        ; // What type of tree format to use at the end (if writing trees)
  string outputDir        ; // What directory should the output files be placed in?
  bool   output_MLE       ; // Must remain at false - not fully implemented yet.
  bool   collect_times    ; // should certain patterns of tMRCA across sequences should be stored
  bool   record_every_time; // should tMRCA at every site should be recorded on every iteration
  bool   collect_densities; // This indicates whether to report the density of each ARG.
  string file_tag         ; // a string to be appended to output files. Can (and probably should) be set at the command line for multiple runs of the program
  string tree_file_prefix ; // prefix to put on the front of treeOutput files
  string dataFormat       ; // names of the data input format to be used - use 'hapmap' unless you know whhat you're doing.
  bool   useHapmap        ; // somewhat redundant boolean to record whether hapmap input data being used.
  // 38 members above^

  int debugMaxNumHaps     ; // for debugging only. Limits the number of haplotypes for ease of debugging

  string data_Fname       ; // probably obsolete. filename for make-sample style data
  int    which_set        ; // probably obsolete. which data set to be used with make-sample style data with multiple data sets in one file...
  string snp_names_file   ; // probably obsolete. filename the specifies a file which gives SNP names - with make-sample style data

  bool   gzipped;
  string output_prefix;
  string snps_to_exclude_file;
  string snps_to_keep_file;
};

#endif /* ARGUMENTSOFPROGRAM_H_ */
