/*
 * The following constructor for Ancestry based on code by Adam Auton, from Nov 11 2009.
 *
// Class for reading in, checking and storing user parameters
*/

#include "parameters.h"

//parameters::parameters(, Random* Inst_of_randomint& argc, char* argv[])
parameters::parameters(const int& argc, char* argv[])
{
  string tmp;
  for (int i=0; i<argc; i++)
  {
    tmp = argv[i];
    argVector.push_back(tmp);
  }
  print_help();
  read_parameters();
  // Need: dataFormat="hapmap"; // or can be "makeSample" which calls: Input_data(data_Fname, which_set);
  //       using_data=1;
  //       

  sim_data = false;
}

void parameters::read_parameters()
{
  unsigned int i=1;
  string in_str;
  while (i<argVector.size())
  {
    in_str = argVector.at(i);
    if      (in_str ==  "-legend"       ) { hapmapParams.legend_file     = get_arg(i+1); i++; gzipped=false; } // legend file to process
    else if (in_str ==  "-haplotypes"   ) { hapmapParams.haps_file.fName = get_arg(i+1); i++; }                // hapoltypes file to process
    else if (in_str ==  "-genMap"       ) { hapmapParams.map_file        = get_arg(i+1); i++; }                // genetic map file to process
    else if (in_str ==  "-upper"        ) { hapmapParams.upper           = atoi(get_arg(i+1).c_str()); i++; }  // upper end of interval to analyse
    else if (in_str ==  "-lower"        ) { hapmapParams.lower           = atoi(get_arg(i+1).c_str()); i++; }  // lower end of interval to analyse
    else if (in_str ==  "-buffer"       ) { hapmapParams.buffer          = atoi(get_arg(i+1).c_str()); i++; }  // extra SNPs (in kb) to avoid edge effects
    else if (in_str ==  "-display_width") { hapmapParams.display_width   = atoi(get_arg(i+1).c_str()); i++; }  // irrelevant display parameter
    else if (in_str ==  "-Ne"           ) { hapmapParams.Ne              = atoi(get_arg(i+1).c_str()); i++; }  // effective population size
    else if (in_str ==  "-rhoScale"     ) {hapmapParams.rhoScale         = get_arg(i+1)              ; i++;}   // scale on which to measure rec rate (rho) - "cMperMb" or "coalescentScale"

    else if (in_str ==  "-theta"            ) {theta            = atof(get_arg(i+1).c_str())       ; i++;}  // mutation rate
    else if (in_str ==  "-rho"              ) {rho              = atof(get_arg(i+1).c_str())       ; i++;}  // recombination rate
    else if (in_str ==  "-SMC"              ) {SMC              = convertStringToBool(get_arg(i+1)); i++;}  // use SMC? (if not coalescent)
    else if (in_str ==  "-choose_hap_first" ) {choose_hap_first = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-coal_fast"        ) {coal_fast        = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-threshold"        ) {threshold        = atof(get_arg(i+1).c_str())       ; i++;}  //
    else if (in_str ==  "-choose_nonrec"    ) {choose_nonrec    = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-first_nonrec"     ) {first_nonrec     = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-use_logs"         ) {use_logs         = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-choose_nonrec"    ) {choose_nonrec    = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-cheap_coal_upwt"  ) {cheap_coal_upwt  = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-display_events"   ) {display_events   = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-display_haps"     ) {display_haps     = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-display_weights"  ) {display_weights  = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-set_seed"         ) {set_seed         = atoi(get_arg(i+1).c_str())       ; i++;}  //
    else if (in_str ==  "-using_wattersons" ) {using_wattersons = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-no_runs"          ) {no_runs          = atoi(get_arg(i+1).c_str())       ; i++;}  // The number of genealogies to simulate
    else if (in_str ==  "-change_rho"       ) {change_rho       = atoi(get_arg(i+1).c_str())       ; i++;}  // The number of values of rho to try (only useful for likelihoods I think)
    else if (in_str ==  "-greedy_mle"       ) {greedy_mle       = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-write_likelihoods") {write_likelihoods= convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-write_trees"      ) {
      write_trees      = convertStringToBool(get_arg(i+1)); i++;
    }  //
    else if (in_str ==  "-write_full_data"  ) {write_full_data  = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-writeType"        ) {writeType        = get_arg(i+1)                     ; i++;}  //
    else if (in_str ==  "-outputDir"        ) {outputDir        = get_arg(i+1)                     ; i++;}  //
    else if (in_str ==  "-collect_times"    ) {collect_times    = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-record_every_time") {record_every_time= convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-collect_densities") {collect_densities= convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-file_tag"         ) {file_tag         = get_arg(i+1)                     ; i++;}  //
  //else if (in_str ==  "-DEBUG"            ) {DEBUG            = convertStringToBool(get_arg(i+1)); i++;}  //
    else if (in_str ==  "-DEBUG"            ) {DEBUG            = atof(get_arg(i+1).c_str())       ; i++;}  // control level of screen output and when program should stop/pause
    else if (in_str ==  "-verbose"          ) {verbose          = atof(get_arg(i+1).c_str())       ; i++;}  // control level of screen output, ideally will take over from DEBUG
    else if (in_str ==  "-debugMaxNumHaps"  ) {debugMaxNumHaps  = atoi(get_arg(i+1).c_str())       ; i++;}  //
    else {
      cout << "Problem reading in parameters:\n";
      cout << "argVector.size() = " << argVector.size() << "\n";
      for(int index=1; index < (int)argVector.size(); index++) {
        cout << "argVector.at(index) = " << argVector.at(index) << "\n";
      }
      cout << "Unrecognised option: " << string(in_str) << "\n";
      cout << "Hmm, press a key to continue:";
      getchar();
      error("Unknown option: " + string(in_str), 0);
    }
    i++;
    check_parameters();
  }
  //(!) Why does this always get rounded down tot he nearest integer!?
  cout << "After checking command line: DEBUG = " << DEBUG << "\n";
  return;
}

bool convertStringToBool(const string& inString)
{
  bool truthValue;
  if(
      inString=="true"  ||  inString=="TRUE"  ||  inString=="True"  ||
      inString=="T"     ||  inString=="t"     ||
      inString=="YES"   ||  inString=="Yes"   ||  inString=="yes"   ||
      inString=="Y"     ||  inString=="y"     ||
      inString=="1"
    )
    truthValue = true;
  else if
  (
      inString=="false" ||  inString=="FALSE" ||  inString=="False" ||
      inString=="F"     ||  inString=="f"     ||
      inString=="NO"    ||  inString=="No"    ||  inString=="no"    ||
      inString=="N"     ||  inString=="n"     ||
      inString=="0"
  )
      truthValue = false;
  else  {
    cout << "  Warning! Problem with program arguments from the command line\n";
    cout << "  Could not recognise value: " << inString << "\n";
    cout << "  Press any key to continue:";
    getchar();
    cout << "\n";
    exit(1);
  }
  return truthValue;
}

string parameters::get_arg(unsigned int i)
{
  if (i>=argVector.size())
    error("Requested Missing Argument",76);
  return argVector[i];
}

void parameters::print_params()
{
  // Line below removed because I don't have a constructor which doens't take in a Random*
  //parameters defaults(0, 0);

  //printLOG("Parameters as interpreted:\n");
  //if (output_prefix != defaults.output_prefix) printLOG("\t--out " + output_prefix + "\n");
  //if (plink_output) printLOG("\t--plink\n");
  //if (min_quality != defaults.min_quality) printLOG("\t--minQ " + dbl2str(min_quality, 3) + "\n");

  //if (snps_to_keep.size() > 0)
  // for (set<string>::iterator it=snps_to_keep.begin(); it != snps_to_keep.end(); ++it)
  //  {
  //    string tmp = *it;
  //    printLOG("\t--snp " + tmp + "\n");
  //  }
  //printLOG("\n");
  return;
}

void parameters::print_help(bool quit)
{
  unsigned int i;
  string in_str;

  if (argVector.size() <= 1)
  {  // If there are no user parameters, display help.
    cout << "  Warning! have executed the program with no arguments. Continue to analyse or get help?\n";
    cout << "  Type analyse to continue or anything else to get help and exit.\n";
    string getHelp;
    cin >> getHelp;
    if (getHelp=="analyse") {
      return;
    } else {
      argVector.push_back("--?");
    }
    //print_help();
  }

  for(i = 0; i < argVector.size(); i++)
  {
    in_str = argVector[i];
    if ((in_str == "-h") || (in_str == "-?") || (in_str == "-help") || (in_str == "--?") || (in_str == "--help") || (in_str == "--h"))
    {
      cout << endl << "treesim (" << "no version numbering yet" << ")" << endl;
      cout << "\u00A9 Niall Cardin 2010" << endl;
      cout << "Generate Genealogies conditional on data, and calculate likelihoods" << endl;
      cout << "For a list of options, please examine parameters.cpp:" << endl;
      //cout << "\thttp://vcftools.sourceforge.net/options.html" << endl;
      cout << endl;

      if(quit)
        exit(0);
      else
        break;
    }
  }
  return;
}

void parameters::check_parameters()
{
  //if (vcf_filename == "") error("VCF required.", 0);
  //if (end_pos < start_pos) error("End position must be greater than Start position.", 1);
  return;
}

void parameters::error(string err_msg, int code)
{
  //printLOG("\n\nError: " + err_msg + "\n\n");
  cout << "\n\nError: " << err_msg << "\n\n";
  exit(code);
}

