#include "read_control.h"

void Read_Parameters(
    double& theta         , double& rho           , bool& SMC              , bool&   importance_sample,
    bool& sim_data        , bool& using_rec_map   , bool& all_nodes        , double& DEBUG            ,
    int& sampler,
    int& no_runs          , int& change_rho       , int& bridge_resolution , bool& change_theta       ,
    bool& output_MLE      , bool& choose_hap_first, bool& coal_fast        , double& threshold        ,
    bool& choose_nonrec   , bool& first_nonrec    , bool& use_logs         , bool& cheap_coal_upwt    ,
    bool& display_events  , bool& display_haps    , bool& display_weights  , int& set_seed            ,
    bool& using_wattersons, bool& greedy_mle      , bool& write_likelihoods, int& write_trees         ,
    bool& write_full_data , bool& collect_times   , bool& record_every_time, bool& collect_densities  ,
    int& which_set        , int& file_tag         ,
    Con& con              , string con_Fname)
{
	int stop;
	int goodCount;
	ifstream ConF;
	char inchar;

	ConF.open(con_Fname.c_str(),ios::in);
  if(!ConF.is_open()){
    cout << "Fatal Error - Control File not opened Properly (is.open()==false).\n";
    cout << "Control filename (Read_Control.cpp)= " << con_Fname.c_str() << "\n";
    exit(1);
  }
	if(!ConF.good()) {
		cout << "WARNING - Problem reading control file (.good()==false).\n";
		cout << "Control filename (Read_Control.cpp)= " << con_Fname.c_str() << "\n";
		exit(1);
	}
	
	stop = 0; goodCount=0; bool readConDebug = false;
	//string checkString; char checkChar[5000]; int checkInt;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(readConDebug)
			cout << inchar;
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> theta;
			//ConF.read(checkChar, 10);
			//cout << "checkChar = " << checkChar << "\n";
			//ConF >> checkInt;
			//cout << "checkInt = " << checkInt << "\n";
			stop = 1;
		}
		goodCount++;
	}
	if(readConDebug)
	   cout << "\n";
	if(!ConF.good()) {
		cout << "WARNING - Control File (probably) not opened Properly.\n";
		cout << "Control filename (Read_Control.cpp)= " << con_Fname.c_str() << "\n";
		cout << "  goodCount    = " << goodCount      << "\n";
		cout << "  read state   = " << ConF.rdstate() << "\n";
		cout << "  end of file? = " << ConF.eof()     << "\n";
		cout << "  bad?         = " << ConF.bad()     << "\n";
		cout << "  fail?        = " << ConF.fail()    << "\n";
		
		ifstream debugF;
		debugF.open(con_Fname.c_str(),ios::in);
		char* debugString;
		debugString = new char [5000];

		unsigned int maxReadLine = 2000;
		const char delim = ':';
		//int returnVal;
		//cout << "  First " << maxReadLine << " chars = " << "\n";
		cout << "  First few lines of Control File = " << "\n";
		for(int lineNum=0; lineNum<12; lineNum++)
		{
			debugF.getline(debugString, maxReadLine, delim );
			cout << debugString << "\n";
		}
		delete debugString;

	}
    con.theta=theta;

	stop = 0; goodCount=0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> rho;
			stop = 1;
		}
		goodCount++;
	}
    con.rho=rho;
	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = "    << ConF.good() 
					  << ", and ConF.eof() = " << ConF.eof() 
					  << ", and goodCount = "  << goodCount
					  << "\n";
		cout << errorMessage2.str() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer (at rho) !\n",
        		errorMessage2.str(),
		        con_Fname.c_str());
        exit(1);
	}

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> SMC;
			stop = 1;
		}
	}
    con.SMC=SMC;
	

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> importance_sample;
			stop = 1;
		}
	}
    con.importance_sample=importance_sample;

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> choose_hap_first;
			stop = 1;
		}
	}     
    con.choose_hap_first=choose_hap_first;

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> coal_fast;
			stop = 1;
		}
	}     
    con.coal_fast=coal_fast;

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> threshold;
			stop = 1;
		}
	}     
    con.threshold=threshold;

	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer (at threshold) !\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> choose_nonrec;
			stop = 1;
		}
	}     
    con.choose_nonrec=choose_nonrec;
	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer (at choose_nonrec) !\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}

	stop = 0;  goodCount=0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		goodCount++;
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> first_nonrec;
			stop = 1;
		}
	}     
    con.first_nonrec=first_nonrec;
	if(!ConF.good() || ConF.eof()) {
        cout << " goodCount = " << goodCount << "\n";
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer (at rho) !\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}


	stop = 0; goodCount=0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		goodCount++;
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> cheap_coal_upwt;
			stop = 1;
		}
	}     
    con.cheap_coal_upwt=cheap_coal_upwt;
    con.using_wattersons=using_wattersons;
	if(!ConF.good() || ConF.eof()) {
        cout << " goodCount = " << goodCount << "\n";
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer! (at cheap_coal_upwt)\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> use_logs;
			stop = 1;
		}
	}     
    con.use_logs=use_logs;
    
    
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> sampler;
			stop = 1;
		}
	}
    con.sampler=sampler;



	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> using_wattersons;
			stop = 1;
		}
	}      
    con.using_wattersons=using_wattersons;
	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer! (at using_wattersons)\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}
	

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> greedy_mle;
			stop = 1;
		}
	}      
    con.greedy_mle=greedy_mle;

	
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> no_runs;
			stop = 1;
		}
	}
    con.no_runs=no_runs;
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> change_rho;
			stop = 1;
		}
	}	
    con.change_rho=change_rho;


	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> bridge_resolution;
			stop = 1;
		}
	}	
    con.bridge_resolution=bridge_resolution;
	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer! (at bridge_resolution)\n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> DEBUG;
			stop = 1;
		}
	}
    con.DEBUG=DEBUG;


	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> write_likelihoods;
			stop = 1;
		}
	}      
    con.write_likelihoods=write_likelihoods;	stop = 0;
	
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> write_trees;
			stop = 1;
		}
	}      
    con.write_trees=write_trees;

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			// DEBUGGING LINES TOO HELP ME UNDERSTAND PROBLEMS
			//char temp ;
			//int  temp2;
			//ConF >> temp;
			//temp2           = atoi("a");
			//write_full_data = temp2     ;
			//cout << "temp2 (char-write_full_data) = " << temp            << "\n";
			//cout << "temp  ( int-write_full_data) = " << temp2           << "\n";
			//cout << "write_full_data              = " << write_full_data << "\n";

			//Original Line:
			ConF >> write_full_data;
			stop = 1;
		}
	}
	if(!ConF.good() || ConF.eof()) {
	    stringstream errorMessage2;
	    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
        Read_Parameters_Error(
        		"  Warning!! something wrong with the file pointer! (at write_full_data) \n",
        		errorMessage2.str(), con_Fname.c_str());
        exit(1);
	}
    con.write_full_data=write_full_data;
	if (write_full_data!=false && write_full_data!=true){
		cout << "Unexpected value for write_full_data(" << write_full_data << ")\n";
		cout << "This is likely a mistake/bug. Debugging info:\n";
		cout << "Control filename = " << con_Fname.c_str() << "\n";
		cout << "Please enter a key to continue, program paused: \n";
        exit(1);
	}
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> collect_times;
			stop = 1;
		}
	}      
    con.collect_times=collect_times;
	

		
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> record_every_time;
			stop = 1;
		}
	}      
    con.record_every_time=record_every_time;
	
		
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> collect_densities;
			stop = 1;
		}
	}      
    con.collect_densities=collect_densities;
	


	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> using_rec_map;
			stop = 1;
		}
	}
    con.using_rec_map=using_rec_map;
	if ((int)using_rec_map!=0 && (int)using_rec_map!=1) {
		cout << "Unexpected value for using_rec_map(" << using_rec_map << ")\n";
		cout << "This is likely a mistake/bug. Debugging info:\n";
		cout << "Control filename = " << con_Fname.c_str() << "\n";
		cout << "Please enter a key to continue, program paused: \n";
        exit(1);
	}



	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> all_nodes;
			stop = 1;
		}
	}
    con.all_nodes=all_nodes;


	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> sim_data;
			stop = 1;
		}
	}
	if (sim_data || !ConF.good() || ConF.eof()){
		cout << "Simulating Data (value = " << con.sim_data << ")\n"; 
		cout << "As this has not been implemented this is likely a mistake. Debugging info:\n";
		cout << "All nodes       = " << con.all_nodes   << "\n";
		cout << "write_full_data = " << write_full_data << "\n";
		cout << "using_rec_map   = " << using_rec_map   << "\n";
		cout << "Control filename (Read_Control.cpp 2)= " << con_Fname.c_str() << "\n";
		cout << "Theta (first read in parameter): "       << con.theta << "\n";
		cout << "Please enter a key to continue, program paused: \n";
		if(!ConF.good() || ConF.eof()) {
			stringstream errorMessage2;
		    errorMessage2 << "  ConF.good() = " << ConF.good() << ", and ConF.eof() = " << ConF.eof() << "\n";
	        Read_Parameters_Error(
	        		"  Warning!! something wrong with the file pointer! (at sim_data) \n",
	        		errorMessage2.str(), con_Fname.c_str());
		}
        exit(1);
	}
    con.sim_data=sim_data;


	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> display_events;
			stop = 1;
		}
	} 
    con.display_events=display_events;


	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> display_haps;
			stop = 1;
		}
	} 
    con.display_haps=display_haps;

	

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> display_weights;
			stop = 1;
		}
	} 
    con.display_weights=display_weights;

	
	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> set_seed;
			stop = 1;
		}
	} 
    con.set_seed=set_seed;
	if(con.greedy_mle && con.set_seed==0)
	{
		cout << "Warning - MLE finder setting time based seed to 1.\n";
		con.set_seed=1;
		set_seed    =1;
	}
	


	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> change_theta;
			stop = 1;
		}
	}	
    con.change_theta=change_theta;

	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> output_MLE;
			stop = 1;
		}
	}      
    con.output_MLE=output_MLE;


	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> which_set;
			stop = 1;
		}
	}      
    con.which_set=which_set;
	

	
	stop = 0;
	while (ConF.good() && !ConF.eof() && stop == 0)
	{
		inchar = ConF.get();
		if(inchar == ':') //Note single quotes refer to characters, double quotes refer to char*
		{
			ConF >> file_tag;
			stop = 1;
		}
	}      
    con.file_tag=file_tag;


	ConF.close();
}

void Read_Parameters_Error(string errorMessage1, string errorMessage2, string errorMessage3)
{
	cout << errorMessage1 << "\n";
	cout << errorMessage2 << "\n";
	cout << "  The Control filename is: " << errorMessage3 << "\n";
	cout << "A common reason for errors here is in incorrectly formatted Control File.\n";
	cout << "Perhaps you are using an obsolete format? Check agreement of variables with read_data.cpp?\n";
}


