#ifndef READ_DATA_H
#define READ_DATA_H

//Memory leak detection:
//#define CRTDBG_MAP_ALLOC
#include <stdlib.h>
//#include <crtdbg.h>

#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <vector>
#include <string>

#include <sstream>
#include <fstream>
#include <map>

#include "control.h"
#include "snp.h"
#include "BadFileNameException.h"

using namespace std;


#ifndef min
  #define min(a,b) ((a) <= (b) ? (a) : (b))
#endif
#ifndef max
  #define max(a,b) ((a) >= (b) ? (a) : (b))
#endif

void Read_Data(int& nseq, int& all_sites, int& no_segsts,  bool& using_rec_map,
    bool& print_data, vector<double>& site_posns, const int& which_set, Con& con, string data_Fname,
    vector< vector<int> >& data,
    double DEBUG = 1);
    // double DEBUG = 1 - gave a problem with calls to function, despite the default
//void Read_Data(int& nseq, int& all_sites, int& no_segsts,  bool& using_rec_map, bool& print_data, vector<double>& site_posns, const int& which_set, string data_Fname, vector< vector<int> >& data);

void setMyDatFromHapmap(
    const map<double, snp>& hapmapDat   ,
    vector < vector<int> >& myFormatHaps,
    vector<double>&         site_posns  ,
    Con&                    con         ,
    double DEBUG = 1
  );

void read_genetic_map(const ChapmapParams params,
      vector<double>& genmap , // genetic map for region
      vector<double>& physmap, // physical map for region
      double DEBUG=1
    );
void read_haplotype_data(
      const ChapmapParams        params ,
      std::map<double, snp>& hapDat ,
      const vector<double>&  genmap , // genetic map for region
      const vector<double>&  physmap, // physical map for region
      int n_gen, int n_hap,
      int DEBUG = 2
    );
void read_sample_file(const ChapmapParams params, int n_gen,
      map<string, int> sample_id_to_index_map
    );
void read_sample_exclusion_file(const ChapmapParams params,
    vector<string> exclusion_list_samples);

double phys2gen(
      const double physpos,
      const vector<double> genmap, // genetic map for region
      const vector<double> physmap // physical map for region
    );

void print_files(
    const ChapmapParams params,
    const int n_hap       , // # haplotypes
    const int n_gen         // # genotypes
) ;
void print_params(
      const ChapmapParams params,
      const double theta,
      const int n_hap, // # haplotypes
      const int n_gen // # genotypes
    ) ;

void printData(
    const int& no_segsts,
    const int& smpl_size,
    const vector< vector<int> >& data,
    const vector<double       >& site_posns,
    const map<double, snp     >& hapmapDat,
    int        max_data_size = 20000,
    int        which_set     = 1    ,
    bool       print_data    = true ,
    double     DEBUG         = 1
  );
//int countLines (string fl);


#endif

//      class Foo {
//      int a;
//      int b;
//      Foo() : a(10) {
//       b = 3;
//      }
//      Foo(int input_b) : a(10) {
//       b = input_b;
//      }
//      ~Foo() {};
//      };

//      Advice and help from Nick Krempel:
//      Default arguments can be specified with: void foo(int a = 10)
//      For default arguments, where we want to set member data of structs: A arg = {blah, blah}
//      but the above example i would do with a single constructor anyway, and make int input_b a default argument with value 3
//      also, for style reasons you should prob initialize both a and b with the ": a(10), b(input_b)" syntax, or assign them both with "a = 10; b = input_b;" - mixing styles is strange
//      for PODs, it doesnt matter which you do, but in general the ": a(10), b(input_b)" syntax is better
//      as it replaces "default construction + assignment" with just "construction"
