#include "write_trees.h"
//if ( i >smpl_size-1 ) {tree_file.width(2); tree_file << marg_gens.at(i).at(0).c_time    << "\t";}

void Write_Tree(
    const vector< vector<node> >&  marg_gens   ,  const map<double, snp>& hapmapDat   ,
    const vector<recomb_event>&    rec_records ,  const vector<CEvent>&   event_vector,
    const vector<double>&      choice_quotients,  const int&              rec_sites   ,
    const double&                  rho         ,  const int&              smpl_size   ,
    const bool&                    SMC         ,  const bool&             all_nodes   ,
    const vector<double>&          positions   ,  const Con&              con         ,
    const string& outputDir                    ,
    const string& writeType                    ,  int writeVerbose
  )
{
  int start_node=0;
  string suffix;
  string dirStatus="unknown";

  checkDirectory(outputDir, writeVerbose, dirStatus);

  if (!all_nodes)
    start_node = smpl_size;

  string treeFilename;
  makeSingleTreesFname(con, treeFilename, outputDir, writeType);
  if(writeVerbose>=1) {
    cout << "  Writing " << writeType << " trees to: " << treeFilename << "\n";
  }

  ofstream tree_file;
  if(writeType=="oldStyle") {
    tree_file.open(treeFilename.c_str(), ios::out);
    writeTreePreamble(tree_file, treeFilename, rec_sites, rho, smpl_size, SMC);
    // The following writes all trees tto a single file (I think)
    oldWriteMainTree(rec_sites, start_node, smpl_size, tree_file, marg_gens);
    tree_file.close();
  }
  if (writeType == "genecluster") {
    if(writeVerbose>2) 
      cout << "  Trying to open the file: " << treeFilename << "\n";

    if (!tree_file.good()) {
      cout << "  Warning! problem strange problem with tree_file (in Write_Tree).\n";
      cout << "  the attribute .good() is false even before opening!\n";
      cout << "  Press any key to continue:";
      getchar();
      cout << "\n";
      writeVerbose = 3;
    } else {
      if(writeVerbose >2)
        cout << "Tree file has declared and .good()==TRUE\n";
    }

    tree_file.open(treeFilename.c_str(), ios::out);

    if (!tree_file.good()) {
      cout << "  Warning! problem opening the treeFile (in Write_Tree).\n";
      cout << "  Press any key to continue:";
      getchar();
      cout << "\n";
      writeVerbose = 3;
    } else
        if(writeVerbose>=2)
          cout << "Tree file has been successfully opened and .good()==TRUE\n";

    writeGeneclusterTree(hapmapDat, smpl_size, tree_file, marg_gens, choice_quotients, event_vector, writeVerbose);
    tree_file.close();
  }
  vector<string> file_suffices;

  if(writeType=="ZhanMark1" || writeType=="allFiles") {
    makeSNPnamedTreeFilenames(con, file_suffices, rec_sites, suffix);
  }
  if(writeType=="ZhanMark1") {
    int j=0;
    while (j < rec_sites) {
      oneTreePerFileWriteMainTree(rec_sites, con, outputDir, file_suffices, smpl_size,
          start_node, marg_gens, j);
      j++;
    }
  }
  if(writeType=="allEvents") {
    tree_file.open(treeFilename.c_str(), ios::out);
    bool check2 = tree_file.good();
    if (!check2) {
      cout << "  Error! Problem opening the file recording all the events (in Write_Tree).\n";
      cout << "  Filename attempting to write to is:" << treeFilename << "\n";
      getchar();
    }
    writeAllEventsData(tree_file, choice_quotients, event_vector, marg_gens);
    tree_file.close();
  }
  return;
}

void writeGeneclusterTree(
    const map<double, snp>& hapmapDat       , const int&                    smpl_size   ,
    ofstream &              treeFile        , const vector< vector<node> >& marg_gens   ,
    const vector<double>&   choice_quotients, const vector<CEvent>&         event_vector,
    int gcVerbose                           )
{
  //position probability; numLeaves; lineage11 lineage12, lineage21 lineage22, ..., lineage((n+k)1) lineage((n+k)2);
  //for(int site=0; site<(int)hapmapDat.size(); site++) {
  // Loop to write each tree on a single line of the file. No pre or post amble.
  snp tempSNP;
  int tempPos;
  int site=0;
  for(map<double, snp>::const_iterator hapDatIt  = hapmapDat.begin();
                                       hapDatIt != hapmapDat.end()  ;
                                     ++hapDatIt)
  {
    stringstream output;
    if(!treeFile.good()) {
      cout << "  Warning! problem opening the treeFile (in writeGeneclusterTree).\n";
      cout << "  Press any key to continue:";
      getchar();
      cout << "\n";
      gcVerbose =3;
    } else
      if(gcVerbose>2)
        cout << "  treeFile still good in writeGeneclusterTree\n";
    // Pull out the physical position of the SNP:
    tempSNP = hapDatIt->second;
    tempPos = hapDatIt->first ;

    output << tempPos   << " "
           << "1"       << "; "
           << smpl_size << "; ";
    writeOutputUnit(gcVerbose, output, treeFile);
    vector< vector<int> > coalescences;
    generateOrderedCoalescences(coalescences, marg_gens, smpl_size, site, gcVerbose);
    for(int epoch=0; epoch<smpl_size-1; epoch++) {
      output << coalescences.at(0).at(epoch)+1 <<  " "; //This should be daughter1 at the coalescence in this epoch
      output << coalescences.at(1).at(epoch)+1; //This should be daughter2 at the coalescence in this epoch
      if(epoch<smpl_size-2)
        output << ", ";
    }
    output << ";\n";
    if(gcVerbose>2)
      cout << output.str();
    writeOutputUnit(gcVerbose, output, treeFile);
    site++;
  }
  if(gcVerbose>2) {
    cout << "  The 'good' status of the file is: " << treeFile.good();
  }
  return;
}

// Function to use marginal genealogies to produce coalescence events inthe order that Zhan wants.
void generateOrderedCoalescences(
    vector< vector<int> >& coalescences,
    const vector< vector<node> >& marg_gens,
    const int& smpl_size,
    const int& site,
    int& verbose
  )
{
  coalescences.resize(2); // A vector of daughter1's and of daughter2's
  coalescences.at(0).resize(smpl_size-1); // This vector records the sequence (N-1) daughter 1's
  coalescences.at(1).resize(smpl_size-1);
//  int epoch=-1;
  for(int nodeIndex=smpl_size; nodeIndex<smpl_size+(smpl_size-1); nodeIndex++) {
    //    epoch = marg_gens.at(nodeIndex).at(site).index_epoch-smpl_size;
    //    coalescences.at(0).at(epoch) = marg_gens.at(nodeIndex).at(site).getN1()->index_epoch;
    //    coalescences.at(1).at(epoch) = marg_gens.at(nodeIndex).at(site).getN2()->index_epoch;
    coalescences.at(0).at(nodeIndex-smpl_size) = marg_gens.at(nodeIndex).at(site).getN1()->index_epoch;
    coalescences.at(1).at(nodeIndex-smpl_size) = marg_gens.at(nodeIndex).at(site).getN2()->index_epoch;
    if(coalescences.at(0).at(nodeIndex-smpl_size)==coalescences.at(1).at(nodeIndex-smpl_size))
    {
      cout << "  Warning! coalescences.at(0).at(nodeIndex-smpl_size)=coalescences.at(1).at(nodeIndex-smpl_size)\n";
      cout << "nodeIndex=" << nodeIndex << ", smpl_size=" << smpl_size << ", site=" << site <<"\n";
      cout << "coalescences.at(1).at(nodeIndex-smpl_size) = " << coalescences.at(1).at(nodeIndex-smpl_size) << "\n";
      cout << "marg_gens.at(nodeIndex).at(site).getN1()->index_epoch = " << marg_gens.at(nodeIndex).at(site).getN1()->index_epoch << "\n";
      cout << "marg_gens.at(nodeIndex).at(site).getN2()->index_epoch = " << marg_gens.at(nodeIndex).at(site).getN2()->index_epoch << "\n";
      verbose = 3;
      getchar();
    }
    if(verbose>=2) {

    }
  }
  return;
}

void writeOutputUnit(int& verbose, stringstream& outputUnit, ofstream& fileStream)
{
    if(verbose >= 3) {
      //cout << "Writing this to file (in writeOutputUnit): " << outputUnit << "\n";
      cout << "Writing this to file (in writeOutputUnit): " << outputUnit.str() << "\n";
    }
    fileStream << outputUnit.str();
    outputUnit.str("");
    if(!fileStream.good()) {
      cout << "  Problem writing to file (in writeOutputUnit): " << fileStream.good()  << "\n";
      cout << "  The 'good' status of the file is: " << fileStream.good() << "\n";
      verbose =3;
    } else
      if(verbose >= 3)
          cout << "No problems detected in writeOutputUnit having written to file: " << outputUnit.str() << "\n";
}

void writeAllEventsData(
    ofstream&               tree_file       ,
    const vector<double>&   choice_quotients,
    const vector<CEvent>&   event_vector    ,
    const vector<vector<node> >& marg_gens
  )
{
  bool check1 = choice_quotients.size()==event_vector.size();
  if(!check1) {
    cout << "  Warning! you don't remember how your own program works, Niall (in writeAllEventsData).\n";
    getchar();
  }
  int numLines = choice_quotients.size();
  bool check2 = tree_file.good();
  if(!check2) {
    cout << "  Error! Problem opening the file recording all the events (in writeAllEventsData).\n";
    getchar();
  }

  tree_file
    << "event_type\t"    << "evnt_seq\t"    << "daughter1\t"
    << "daughter2\t"     << "rec_lineage\t" << "mut_lineage\t"
    << "rec_site\t"      << "mut_site\t"    << "new_type\t"
    << "levent_weight\t" << "c_time\t"      << "Site0_time_below\t"
    << "cq\n";
  check2 = tree_file.good();
  if (!check2) {
    cout << "  Error! Problem write to file recording all the events (in writeAllEventsData).\n";
    getchar();
  }
  for(int line=0; line<numLines; line++)
  {
    tree_file << event_vector.at(line).event_type    << "\t";
    tree_file << event_vector.at(line).evnt_seq      << "\t";
    tree_file << event_vector.at(line).daughter1     << "\t";
    tree_file << event_vector.at(line).daughter2     << "\t";
    tree_file << event_vector.at(line).rec_lineage   << "\t";
    tree_file << event_vector.at(line).mut_lineage   << "\t";
    tree_file << event_vector.at(line).rec_site      << "\t";
    tree_file << event_vector.at(line).mut_site      << "\t";
    tree_file << event_vector.at(line).new_type      << "\t";
    tree_file << event_vector.at(line).levent_weight << "\t";
    tree_file << marg_gens.at(line).at(0).c_time     << "\t";
    tree_file << marg_gens.at(line).at(0).time_below << "\t";
    tree_file << choice_quotients.at(line)           << "\n";
    check2 = tree_file.good();
    if (!check2) {
      cout << "  Error! Could not write some data to events log (in writeAllEventsData).\n";
      getchar();
    }
  }
  return;
}


void makeSNPnamedTreeFilenames(const Con& con, vector<string>& file_suffices, const int& rec_sites, string& suffix)
{
    if(con.snp_names_file != "no_file"){
        string tmp_string;
        ifstream snp_namesF;
        snp_namesF.open(con.snp_names_file.c_str(), ios::in);
        // Error reporting when there's a problem opening the SNP names file. (Eclipse refactored function
        snpNamesFileError(snp_namesF, con);
        snp_namesF >> tmp_string;
        while(snp_namesF.good() && !snp_namesF.eof()){
            //   cout << "\n tmp_string = " << tmp_string << "\n";
            file_suffices.push_back(tmp_string);
            snp_namesF >> tmp_string;
        }
        if(!snp_namesF.good())
            cout << "Problem reading SNP names file (encountered after first byte read in).\n";
        automaticallyNameFilesWhenSNPsFileNotGiven(file_suffices, con);
    }else{
        for(int index = 0;index < rec_sites;index++){
            stringstream tmp_name;
            tmp_name << index;
            suffix = tmp_name.str();
            file_suffices.push_back(suffix);
        }
    }

}

void automaticallyNameFilesWhenSNPsFileNotGiven(vector<string>& file_suffices, const Con& con)
{
    if(file_suffices.size() != (unsigned int)((con.no_segsts))){
        cout << "\nWarning! Number of SNP names (in file) (" << file_suffices.size() << ") given doesn't equal number of sites(" << con.no_segsts << ") in main data.\n";
        cout << "The file being used for SNP names is :" << con.snp_names_file.c_str() << "\n\n";
        cout << "The names are:\n\n";
        for(int SNP_name = 0;(unsigned int)((SNP_name)) < file_suffices.size();SNP_name++)
            cout << "Name number " << SNP_name + 1 << " = " << file_suffices.at(SNP_name) << "\n";

        cout << "  SETTING remaining names to general labels!:\n";
        stringstream snpLabel;
        for(int snpIndex = file_suffices.size(); snpIndex < con.no_segsts; snpIndex++){
            snpLabel.str("");
            snpLabel << "SNP" << snpIndex + 1;
            //cout << "snpLabel is currently: " << snpLabel.str() << "\n";
            file_suffices.push_back(snpLabel.str());
        }
        //cout << "Aborting.\n\n";
        //exit(1);
    }

}



void writeTreePreamble(ofstream& tree_file, string& tree_fileN,
    const int& rec_sites, const double& rho, const int& smpl_size,
    const int& SMC) {
  tree_file << "No sites:                     " << rec_sites << endl
      << "Rho from 1st to last site:    " << rho << endl
      << "Number of sequences:          " << smpl_size << endl
      << "Using SMC? (0=Coal):          " << SMC << endl << endl;
  tree_file << "site\t" << "Dscndt1\t" << "Dscndt2\t" << "Time\t" << "Total\t"
      << "time\t" << "below\t" << "Active?" << endl << endl;
}

void snpNamesFileError(ifstream& snp_namesF, const Con& con) {
  if (!snp_namesF.good()) {
    cout << "  Warning! problem opening SNP names file, does it exist?\n";
    cout << "  The file I tried to open was: con.snp_names_file = "
        << con.snp_names_file << "";
    cout << ", " << con.snp_names_file.c_str() << "\n";
    cout
        << "  The size of the string being used was: con.snp_names_file.size() = "
        << con.snp_names_file.size() << "\n";
    cout << "  Press any key to continue:";
    getchar();
  }
}

void oldWriteMainTree(const int& rec_sites, const int& start_node,
    const int& smpl_size, ofstream & tree_file,
    const vector<vector<node> >& marg_gens)
{
  int i, j = 0;
  while (j < rec_sites) {
    for (i = start_node; i < 2 * smpl_size - 1; i++) {
      tree_file << j
                << "\t"
                << "   ";
      if (marg_gens.at(i).at(j).n1)
        tree_file << marg_gens.at(i).at(j).n1->index_epoch + 1 << "\t   ";
      else
        tree_file << "-1\t   ";

      if (marg_gens.at(i).at(j).n2)
        tree_file << marg_gens.at(i).at(j).n2->index_epoch + 1 << "\t";
      else
        tree_file << "-1\t";

      if (i > smpl_size - 1) {
        tree_file.precision(3);
        tree_file << marg_gens.at(i).at(j).c_time << "\t    ";
      } else
        tree_file << marg_gens.at(i).at(j).c_time << "\t\t    ";

      if (i > smpl_size - 1)
        tree_file << marg_gens.at(i).at(j).time_below << "\t\t   ";
      else
        tree_file << marg_gens.at(i).at(j).time_below << "\t\t";

      tree_file << marg_gens.at(i).at(j).active << endl;
    }
    j++;
  }
  tree_file << endl << endl;
}

void makeTreeFileName(
    const Con& con,
    const string& outputDir,
    const vector<string>& file_suffices,
    const int& j,
    const int& smpl_size,
    string& treeFilename)
{
    stringstream tmp1;
    string tmp2;
    tmp1 << outputDir;
    if(con.tree_file_prefix != "use_default")  {
        tmp1 << con.tree_file_prefix;
        if(con.snp_names_file != "no_file")
            tmp1 << file_suffices.at(j) << ".txt";
        else
            tmp1 << j << ".txt";
    }
    else  {
        tmp1 << "test_output/Test_Trees_n=" << smpl_size << "_site=" << j << ".txt";
        //tmp1 << "zhan_trees/output/Trees_n=" << smpl_size << "_site=" << positions.at(j) << ".txt" << set.precision(8);
    }
    tmp2 = tmp1.str();
    treeFilename = tmp2.c_str();
}


void makeSingleTreesFname(const Con& con, string& tree_fileN, const string& outputDir, const string& writeType)
{
    stringstream create_Tree_fileN;
    string treeModel;
    if(con.SMC)
        treeModel = "SMC_";
    else
        treeModel = "Coal_";

    create_Tree_fileN << outputDir
                      << con.file_tag
                      << "_Trees_model=" << treeModel << "n=" << con.smpl_size
                      << "_" << writeType
                      << ".txt";
                      //<< "_rho=" << con.rho << "_perSeq.txt";
    tree_fileN = create_Tree_fileN.str().c_str();
}

void oneTreePerFileWriteMainTree(
    const int& rec_sites,
    const Con& con,
    const string& outputDir,
    const vector<string>& file_suffices,
    const int& smpl_size,
    const int& start_node,
    const vector<vector<node> >& marg_gens,
    const int& j
  )
{
  // Extra +1's are because Zhan wants the nodes to start at 1.
  int i;
  string treeFileName;
  ofstream treeFile;
  makeTreeFileName(con, outputDir, file_suffices, j, smpl_size, treeFileName);
  treeFile.open(treeFileName.c_str(), ios::out);
  if (!treeFile.good()) {
    cout << "  Could not open tree output file - perhaps directory does not exist?\n";
    cout << "  Filename attempting to write to is:" << treeFileName << "\n";
    cout << "  Press any key to continue, or Ctrl+C to quit.\n";
    getchar();
  }
  if (j == 0) {
    cout << "The first tree file being written is:" << treeFileName << "\n";
  }
  if (!treeFile.good()) {
    cout << "  Error! Could not write to tree output file.\n";
    cout << "  Filename attempting to write to is:" << treeFileName << "\n";
    cout << "  Press any key to continue, or Ctrl+C to quit.\n";
    getchar();
  }
  for (i = start_node; i < 2 * smpl_size - 1; i++) {
    treeFile << i + 1 << " ";
    if (i >= smpl_size)
      treeFile << (i - smpl_size + 1) << " ";
    else
      treeFile << "0 ";

    if (marg_gens.at(i).at(j).n1)
      treeFile << marg_gens.at(i).at(j).n1->index_epoch + 1 << " ";
    else
      treeFile << "0 ";

    if (marg_gens.at(i).at(j).n2)
      treeFile << marg_gens.at(i).at(j).n2->index_epoch + 1 << " ";
    else
      treeFile << "0 ";

    treeFile << "0 " << "\n";
  }
  treeFile << endl << endl;
  treeFile.close();
}

void checkDirectory(const string& strPath, int verbose, string& dirStatus, bool stopOnProblem)
{
  if(verbose>2)
    cout << "Checking Directory: ";

  struct stat status;
  stat( strPath.c_str(), &status );

  if ( access( strPath.c_str(), 0 ) == 0 )
  {
      if ( status.st_mode & S_IFDIR )
        dirStatus = "validExtantDirectory";
      else
        dirStatus = "inputWasLocationOfAfile";
  }
  else
    dirStatus = "directoryDoesNotExist";

  if(stopOnProblem && ((dirStatus != "validExtantDirectory") || verbose>2)) {
    if((dirStatus != "validExtantDirectory"))
      cout << "  Error, the directory that tree files are to be written to doesn't exist:\n";
    cout << "  The (supposed) directory given was: " << strPath << "\n";
    cout << "  The status of this is: " << dirStatus << "\n";
    if((dirStatus != "validExtantDirectory")) {
      cout << "  Press any key to continue, or Ctrl+C to quit.";
      getchar();
      cout << "\n";
    }
  }
  if(verbose>2) cout << "  returning from checkDirectory\n";
  return;
}








// Removed code from write genecluster trees
//        CEvent ev; int evType;
//        for(int event=0; event<(int)event_vector.size(); event++)
//        {
//          ev     = event_vector.at(event);
//          evType = ev.event_type;         //shorthand for the event type at the end of this epoch
//          if(evType==1 ||evType==2 ||evType==3) {
//            if(evType==1) {
//              output << ev.daughter1 << " " ;
//              output << ev.daughter2 << ", ";
//            } else {}
//          } else {
//            cout << "  Warning! Unrecognised event type in writeGeneclusterTree";
//            getchar();
//          }
//        }
//        output << ";";
//        if(gcVerbose>2)
//          cout << output.str();
//        writeOutputUnit(gcVerbose, output, treeFile);
//      }
//
//    if(gcVerbose>2) cout << "Entering main trees output loop:\n";
//    for(int nodeIndex=0; nodeIndex<(smpl_size-1); nodeIndex++)
//    {
//      if(gcVerbose>2) {
//        cout << "nodeIndex                                   =" << nodeIndex                                        << "\n";
//        cout << "site                                        =" << site                                             << "\n";
//        cout << "marg_gens.at(nodeIndex).at(site).index_epoch=" << marg_gens.at(nodeIndex).at(site).index_epoch     << "\n";
//        if(marg_gens.at(nodeIndex).at(site).n1==NULL) {
//          cout << "marg_gens.at(..).at(..)n1 is NULL - probably a leaf node. And cannot be accessed\n";
//          cout << "press any key to continue. Or Ctrl+C to exit.\n";
//          if(gcVerbose>3) getchar();
//        } else
//          cout << "marg_gens.at(..).at(..)n1->index_epoch      =" << marg_gens.at(nodeIndex).at(site).getN1()->index_epoch << "\n";
//          //cout << "marg_gens.at(..).at(..)n1->index_epoch      =" << marg_gens.at(nodeIndex).at(site).n1->index_epoch << "\n";
//      }
//      if(marg_gens.at(nodeIndex).at(site).n1==NULL) {
//        output << " " << -1;
//      } else
//        output << " " << marg_gens.at(nodeIndex).at(site).getN1()->index_epoch + 1;
//
//      writeOutputUnit(gcVerbose, output, treeFile);
//
//      if(marg_gens.at(nodeIndex).at(site).n1==NULL) {
//        output << " " << -1                                                   << ",";
//      } else
//        output << " " << marg_gens.at(nodeIndex).at(site).getN2()->index_epoch + 1 << ",";
//      writeOutputUnit(gcVerbose, output, treeFile);
//    }
//    treeFile << "\n";
// Zhans's preferred output format
// position probability; numLeaves; lineage11 lineage12, lineage21 lineage22, ..., lineage((n+k)1) lineage((n+k)2);
// Coalescence k creates lineage n+k
//104000 1.0; 120; 31 82, 113 12, 121 9, 95 65, 32 109, 99 100, 51 13, 69 81, 49 70, 57 4, 106 123, 89 117, 21 93, 112 3, 52 66, 102 72, 41 25, 6 2, 22 62, 90 77, 134 5, 71 80, 131 110, 79 126, 133 140, 98 130, 40 78, 28 39, 142 87, 94 120, 91 42, 122 53, 46 105, 139 149, 83 8, 36 14, 151 144, 15 127, 155 103, 38 145, 160 114, 84 68, 60 86, 141 7, 147 152, 136 162, 137 119, 29 24, 129 167, 125 54, 150 165, 75 159, 44 43, 143 111, 50 88, 11 64, 59 161, 61 92, 45 168, 166 132, 58 138, 37 47, 35 1, 164 153, 96 85, 184 180, 97 10, 33 156, 124 104, 170 146, 48 19, 181 187, 176 177, 26 107, 192 190, 67 195, 188 148, 172 158, 182 101, 128 196, 163 171, 183 154, 115 30, 169 185, 203 116, 205 56, 186 198, 108 200, 34 17, 207 202, 208 206, 211 135, 210 194, 55 191, 214 175, 63 213, 157 179, 217 199, 204 20, 16 193, 201 23, 197 173, 220 76, 73 219, 74 222, 218 174, 216 18, 118 178, 227 228, 27 223, 229 226, 209 189, 212 225, 232 224, 221 233, 234 231, 236 230, 235 237, 238 215;

// my preferred format (all fields separated by , and no terminal delimiter) - easy to read in to eg. R:
// position, probability, numLeaves, lineage11 lineage12, lineage21 lineage22, ..., lineage((n+k)1) lineage((n+k)2)
